# Copyright 2017 inspur Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import math
import random
import re
import time
import unicodedata
from eventlet import greenthread
import paramiko
import six

from cinder import context
from cinder import exception
from cinder.i18n import _
from cinder import ssh_utils
from cinder import utils as cinder_utils
from cinder.volume import volume_types
from cinder.volume import driver
from cinder.volume.drivers.inspur.instorage import (
    replication as instorage_rep)
from cinder.volume.drivers.inspur.instorage.instorage_inband import (
     RestExecutor, InStorageRestAssistant)
from cinder.volume.drivers.inspur.instorage import instorage_const
from cinder.volume.drivers.inspur.instorage.utils import RsaUtil
from cinder.volume.drivers.san import san
from cinder.volume import qos_specs

# !!!ATTENTION
# the openstack common package is move out from project
# as some independent library from Kilo
try:
    # the config is work as oslo more early
    from oslo.config import cfg
    from cinder.openstack.common import excutils
    from cinder.openstack.common import processutils
    from cinder.openstack.common import log as logging
    from cinder.openstack.common import jsonutils as json
    from cinder.openstack.common import strutils
    from cinder.openstack.common import units
except ImportError:
    from oslo_config import cfg
    from oslo_concurrency import processutils
    from oslo_log import log as logging
    from oslo_serialization import jsonutils as json
    from oslo_utils import excutils
    from oslo_utils import strutils
    from oslo_utils import units

try:
    from cinder.openstack.common import loopingcall
except ImportError:
    from oslo_service import loopingcall


# !!!ATTENTION
# from Train, openstack rename utils in cinder/volume as volume_utils
try:
    from cinder.volume import utils as volume_utils
except ImportError:
    from cinder.volume import volume_utils

INTERVAL_1_SEC = 1
DEFAULT_TIMEOUT = 20
LOG = logging.getLogger(__name__)


instorage_mcs_opts = [
    cfg.BoolOpt('instorage_mcs_vol_autoexpand',
                default=True,
                help='Storage system autoexpand parameter for volumes '
                     '(True/False)'),
    cfg.BoolOpt('instorage_mcs_vol_compression',
                default=False,
                help='Storage system compression option for volumes'),
    cfg.BoolOpt('instorage_mcs_vol_deduplicate',
                default=False,
                help='Storage system deduplicate option for volumes'),
    cfg.BoolOpt('instorage_mcs_vol_intier',
                default=True,
                help='Enable InTier for volumes'),
    cfg.BoolOpt('instorage_mcs_allow_tenant_qos',
                default=False,
                help='Allow tenants to specify QOS on create'),
    cfg.BoolOpt('instorage_mcs_vol_nofmtdisk',
                default=False,
                help='Specifies that the volume not be formatted during '
                     'creation.'),
    cfg.IntOpt('instorage_mcs_vol_grainsize',
               default=256,
               help='Storage system grain size parameter for volumes '
                    '(32/64/128/256)'),
    cfg.IntOpt('instorage_mcs_vol_rsize',
               default=-1,
               # min=-1, max=100,
               help='Storage system space-efficiency parameter for volumes '
                    '[-1 - 100](percentage)'),
    cfg.IntOpt('instorage_mcs_vol_warning',
               default=80,
               # min=-1, max=100,
               help='Storage system threshold for volume capacity warnings '
                    '[-1 - 100] (percentage)'),
    cfg.IntOpt('instorage_mcs_localcopy_timeout',
               default=120,
               # min=1, max=600,
               help='Maximum number of seconds to wait for LocalCopy to be '
                    'prepared. [1 - 600]'),
    cfg.IntOpt('instorage_mcs_localcopy_rate',
               default=100,
               # min=1, max=150,
               help='Specifies the InStorage LocalCopy copy rate to be used '
               'when creating a full volume copy. The default is rate '
               'is 100, and the valid rates are 1-150.'),
    cfg.IntOpt('instorage_mcs_localcopy_clean_rate',
               default=100,
               help='Specifies the InStorage LocalCopy clean rate to be used '
               'when creating a full volume copy. The default is rate '
               'is 100, and the valid rates are 1-150.'),
    cfg.StrOpt('instorage_mcs_vol_iogrp',
               default='0',
               help='The I/O group in which to allocate volumes. It can be a '
               'comma-separated list in which case the driver will select an '
               'io_group based on least number of volumes associated with the '
               'io_group.'),
    cfg.StrOpt('instorage_mcs_vol_access_iogrp',
               default=None,
               help='Specifies the members of the volume I/O group access set.'),
    cfg.StrOpt('instorage_mcs_stretched_cluster_partner',
               default=None,
               help='If operating in stretched cluster mode, specify the '
                    'name of the pool in which mirrored copies are stored.'
                    'Example: "pool2"'),
    cfg.ListOpt('instorage_mcs_volpool_name',
                default=['volpool'],
                help='Comma separated list of storage system storage '
                     'pools for volumes.'),
    cfg.StrOpt('instorage_san_secondary_ip',
               default=None,
               help='Specifies secondary management IP or hostname to be '
                    'used if san_ip is invalid or becomes inaccessible.'),
    cfg.BoolOpt('instorage_mcs_enable_aa',
                default=False,
                help='Create active-active volume when do volume create'),
    cfg.ListOpt('instorage_mcs_aa_volpool_name_map',
                default=None,
                help='Comma separated list of storage system storage '
                     'pools map for active-active volumes. Each item is '
                     'a base pool to second pool map, seperated by :, '
                     'like Pool0:Pool1, and Pool0 should be a argument to '
                     'instorage_mcs_volpool_name'),
    cfg.ListOpt('instorage_mcs_aa_iogrp_map',
                default=None,
                help='Comma separated list of storage system I/O group '
                     'map for active-active volumes. Each item is '
                     'a base iogrp to second iogrp map, seperated by :, '
                     'like iogrp0:iogrp1, and iogrp0 should be a argument to '
                     'instorage_mcs_vol_iogrp'),
    cfg.BoolOpt('instorage_mcs_auto_delete_host',
                default=True,
                help='Auto delete the host when no volume mapped to it.'),
    cfg.IntOpt('instorage_mcs_syncrate',
               default=80,
               # min=1, max=100,
               help='Specifies the InStorage sync rate to be used '
               'when creating a volume for formatting. The default is rate '
               'is 80, and the valid rates are 1-100.'),
    cfg.IntOpt('instorage_mcs_loop_check_interval',
               default=60,
               # min=1, max=600,
               help='Maximum number of seconds to wait between each check.'),
    cfg.StrOpt('instorage_mcs_supervisor_mode',
               default="SSH",
               help='the mode of managing storage, support SSH, In-band, Rest'),
    cfg.BoolOpt('instorage_mcs_check_root_cert',
                default=False,
                help='Whether do the root cert check.'),
    cfg.StrOpt('instorage_mcs_ca_root',
                help='The ca root path of the as18000.'),
    cfg.BoolOpt('instorage_mcs_check_client_cert',
                default=False,
                help='Whether do the client cert check.'),
    cfg.StrOpt('instorage_mcs_ca_key',
                help='The path of the client cert key.'),
    cfg.StrOpt('instorage_mcs_ca_cert',
                help='The path of the client cert file.'),
    cfg.IntOpt('instorage_mcs_host_identifier',
               # min=0, max=1023, should be unique
               default=-1,
               help='controller node identifier.'),
    cfg.StrOpt('instorage_mcs_cluster_id',
               default=None,
               help='cluster identifier.'),
    cfg.StrOpt('instorage_rsa_private_key',
               default='',
               help='RSA private key for decrypt.'),
    cfg.IntOpt('instorage_mcs_thread_num',
               default=10,
               help='thread num for rest'),
    cfg.IntOpt('ssh_exec_timeout',
               default=60,
               help="SSH execute timeout in seconds"),
]

CONF = cfg.CONF
CONF.register_opts(instorage_mcs_opts)


# !!! ATTENTION
# in Juno san.SanDriver inherited from driver.VolumeDriver, however, dure to
# the inheriate problem, we need first inheriate from san.SanDriver, and then
# from driver.VolumeDriver
# from K to P, CommonDriver need inherit from both VolumeDriver and SanDriver
class InStorageMCSCommonDriver(san.SanDriver, driver.VolumeDriver):
    """inspur InStorage MCS abstract base class for iSCSI/FC volume drivers."""

    VERSION = "4.2.7"
    VDISKCOPYOPS_INTERVAL = 120
    DEFAULT_GR_SLEEP = random.randint(20, 500) / 100.0

    def __init__(self, *args, **kwargs):
        super(InStorageMCSCommonDriver, self).__init__(*args, **kwargs)
        self.configuration.append_config_values(instorage_mcs_opts)
        self._backend_name = self.configuration.safe_get('volume_backend_name')
        self.active_ip = self.configuration.san_ip
        self.inactive_ip = self.configuration.instorage_san_secondary_ip

        self.active_active_pool_map = {}
        pool_maps = self.configuration.instorage_mcs_aa_volpool_name_map
        if pool_maps:
            for pool_map in pool_maps:
                primary, second = tuple(pool_map.split(':'))
                self.active_active_pool_map[primary] = second
                self.active_active_pool_map[second] = primary
            LOG.debug("aa pool map: %s", self.active_active_pool_map)

        self.active_active_iogrp_map = {}
        iogrp_maps = self.configuration.instorage_mcs_aa_iogrp_map
        if iogrp_maps:
            for iogrp_map in iogrp_maps:
                primary, second = tuple(iogrp_map.split(':'))
                self.active_active_iogrp_map[primary] = second
                self.active_active_iogrp_map[second] = primary
            LOG.debug("aa iogrp map: %s", self.active_active_iogrp_map)

        self.loop_check_interval = self.configuration.instorage_mcs_loop_check_interval
        self.sshpool = None
        self.sshpool_available = True
        self.supervisor_mode = self.configuration.instorage_mcs_supervisor_mode
        self.port = self.configuration.safe_get('san_api_port')
        self.ca_root = self.configuration.instorage_mcs_ca_root
        self.ca_key = self.configuration.instorage_mcs_ca_key
        self.ca_cert = self.configuration.instorage_mcs_ca_cert
        self.use_ssl = self.configuration.safe_get('driver_use_ssl') or False
        self.check_root_cert = self.configuration.instorage_mcs_check_root_cert
        self.check_client_cert = self.configuration.instorage_mcs_check_client_cert
        self.host_identifier = self.configuration.safe_get('instorage_mcs_host_identifier')
        self.cluster_id = self.configuration.safe_get('instorage_mcs_cluster_id')
        self._local_backend_rest = None
        self._rest = self._local_backend_rest
        self.rsa_util = None

        if self.supervisor_mode == 'In-band':
            required_flags = ['instorage_mcs_host_identifier', 'instorage_mcs_cluster_id']
            for flag in required_flags:
                if self.configuration.safe_get(flag) == -1:
                    raise exception.InvalidInput(reason=_('%s is not set.') % flag)
                if self.configuration.safe_get(flag) is None:
                    raise exception.InvalidInput(reason=_('%s is not set.') % flag)

        if not self.port:
            self.port = '8446' if self.use_ssl else '8443'

        if self.supervisor_mode == 'SSH':
            assistant = InStorageAssistant(self._run_ssh,
                                       self.active_active_pool_map,
                                       self.active_active_iogrp_map,
                                       self.loop_check_interval)
        elif self.supervisor_mode == 'Rest' or self.supervisor_mode == 'In-band':
            assistant = InStorageRestAssistant(self._run_rest,
                                       self.active_active_pool_map,
                                       self.active_active_iogrp_map,
                                       self.loop_check_interval)
        if self.supervisor_mode == 'In-band':
            assistant.clear_scsi_cache()

        self._local_backend_assistant = assistant
        self._aux_backend_assistant = None
        self._assistant = self._local_backend_assistant
        self._vdiskcopyops = {}
        self._vdiskcopyops_loop = None
        self.protocol = None
        self.replication = None
        self._state = {'storage_nodes': {},
                       'enabled_protocols': set(),
                       'compression_enabled': False,
                       'deduped_enabled': False,
                       'available_iogrps': [],
                       'system_name': None,
                       'system_id': None,
                       'code_level': None,
                       'cloned_extend_support':False,
                       'cascade_cloned_auto_delete':False
                       }
        self._active_backend_id = kwargs.get('active_backend_id')

        # This dictionary is used to map each replication target to certain
        # replication manager object.
        self.replica_manager = {}

        # One driver can be configured with only one replication target
        # to failover.
        self._replica_target = {}

        # This boolean is used to indicate whether replication is supported
        # by this storage.
        self._replica_enabled = False

        # This list is used to save the supported replication modes.
        self._supported_replica_types = []

        # This is used to save the available pools in failed-over status
        self._secondary_pools = None

        # InStorage has the limitation that can not burst more than 3 new ssh
        # connections within 1 second. So slow down the initialization.
        time.sleep(1)

    def do_setup(self, ctxt):
        """Check that we have all configuration details from the storage."""
        LOG.debug('enter: do_setup')

        # Update the instorage state
        self._update_instorage_state()

        # v2.1 replication setup
        self._get_instorage_config()

        # Validate that the pool exists
        self._validate_pools_exist()

        LOG.debug('leave: do_setup')

    def _update_instorage_state(self):
        # Get storage system name, id, and code level
        self._state.update(self._assistant.get_system_info())

        capabilities_value = self._assistant.get_capabilities_value()

        # Check if compression is supported
        self._state['compression_enabled'] = (self._assistant.
                                              compression_enabled())
        # Check if deduped is supported
        self._state['deduped_enabled'] = (self._assistant.
                                          deduped_enabled())

        # Get the available I/O groups
        self._state['available_iogrps'] = (self._assistant.
                                           get_available_io_groups())

        # Get the capability of extending a cloning volume
        self._state['cloned_extend_support'] = (
            self._assistant.extend_cloning_support(capabilities_value))

        # Get the capability of cascade cloned auto delete
        self._state['cascade_cloned_auto_delete'] = (
            self._assistant.cascade_cloned_autodelete_support(capabilities_value))

        # Get the iSCSI and FC names of the InStorage/MCS nodes
        self._state['storage_nodes'] = self._init_node_info()

    def _init_node_info(self):
        """get the controller node infomation."""
        nodes = {}
        nodes = self._assistant.get_node_info()
        # Add the IP addresses and WWPNs to the storage node info
        self._assistant.add_ip_addrs(nodes)
        self._assistant.add_fc_wwpns(nodes)
        # For each node, check what connection modes it supports.  Delete any
        # nodes that do not support any types (may be partially configured).
        to_delete = []
        for k, node in nodes.items():
            if ((len(node['ipv4']) or len(node['ipv6'])) and
                    len(node['iscsi_name'])):
                node['enabled_protocols'].append('iSCSI')
                self._state['enabled_protocols'].add('iSCSI')
            if len(node['WWPN']):
                node['enabled_protocols'].append('FC')
                self._state['enabled_protocols'].add('FC')
            if len(node['ipv4']) or len(node['ipv6']):
                node['enabled_protocols'].append('RDMA')
                self._state['enabled_protocols'].add('RDMA')
            if not node['enabled_protocols']:
                to_delete.append(k)
        for delkey in to_delete:
            del nodes[delkey]
        return nodes

    def _get_backend_pools(self):
        if not self._active_backend_id:
            return self.configuration.instorage_mcs_volpool_name
        elif not self._secondary_pools:
            self._secondary_pools = [self._replica_target.get('pool_name')]
        return self._secondary_pools

    def _validate_pools_exist(self):
        # Validate that the pool exists
        pools = self._get_backend_pools()
        for pool in pools:
            try:
                self._assistant.get_pool_attrs(pool)
            except exception.VolumeBackendAPIException:
                msg = _('Failed getting details for pool %s.') % pool
                raise exception.InvalidInput(reason=msg)

    def check_for_setup_error(self):
        """Ensure that the flags are set properly."""
        LOG.debug('enter: check_for_setup_error')

        # Check that we have the system ID information
        if self._state['system_name'] is None:
            exception_msg = (_('Unable to determine system name.'))
            raise exception.VolumeBackendAPIException(data=exception_msg)
        if self._state['system_id'] is None:
            exception_msg = (_('Unable to determine system id.'))
            raise exception.VolumeBackendAPIException(data=exception_msg)

        # Make sure we have at least one node configured
        if not self._state['storage_nodes']:
            msg = _('do_setup: No configured nodes.')
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)

        if self.protocol not in self._state['enabled_protocols']:
            raise exception.InvalidInput(
                reason=_('The storage device does not support %(prot)s. '
                         'Please configure the device to support %(prot)s or '
                         'switch to a driver using a different protocol.')
                % {'prot': self.protocol})

        required_flags = ['san_ip', 'san_ssh_port', 'san_login',
                          'instorage_mcs_volpool_name']
        for flag in required_flags:
            if not self.configuration.safe_get(flag):
                raise exception.InvalidInput(reason=_('%s is not set.') % flag)

        if not (self.configuration.san_password or
                self.configuration.san_private_key):
            raise exception.InvalidInput(
                reason=_('Password or SSH private key is required for '
                         'authentication: set either san_password or '
                         'san_private_key option.'))

        opts = self._assistant.build_default_opts(self.configuration)
        self._assistant.check_vdisk_opts(self._state, opts)

        LOG.debug('leave: check_for_setup_error')

    def _run_ssh(self, cmd_list, check_exit_code=True, attempts=3):
        """SSH tool"""
        cinder_utils.check_ssh_injection(cmd_list)
        command = ' '.join(cmd_list)
        if not self.sshpool:
            for retry_num in range(attempts):
                temp_active_ip = self.active_ip
                try:
                    self.sshpool = self._set_up_sshpool(self.active_ip)
                    self.sshpool_available = True
                    break
                except paramiko.SSHException:
                    if retry_num == attempts - 1:
                        LOG.error('_run_ssh, Retry was exceeded. Unable '
                                  'to create SSHPool using san_ip')
                        if self._switch_ip(temp_active_ip):
                            try:
                                self.sshpool = self._set_up_sshpool(self.active_ip)
                                self.sshpool_available = True
                                break
                            except paramiko.SSHException:
                                with excutils.save_and_reraise_exception():
                                    LOG.error(_('Unable to create SSHPool using'
                                                'instorage_san_secondary_ip, '
                                                'Please check the network.'))
                        else:
                            with excutils.save_and_reraise_exception():
                                LOG.error('Unable to create SSHPool using san_ip '
                                            'and instorage_san_secondary_ip is '
                                            'not configured. '
                                            'Please check the network.')
                    LOG.warning('_run_ssh, retry, '
                                'ip: %(ip)s, '
                                'TryNum: %(rn)s.',
                                {'ip': self.active_ip,
                                 'rn': retry_num + 1})
                    greenthread.sleep(self.DEFAULT_GR_SLEEP)

        temp_active_ip = self.active_ip
        try:
            while not self.sshpool_available:
                LOG.warning('wait for sshpool available')
                greenthread.sleep(3)
            return self._ssh_execute(self.sshpool, command, check_exit_code, attempts)
        except processutils.ProcessExecutionError as ex:
            if ('mcsinq' in ex.cmd or 'mcsop' in ex.cmd) and 'CMMVC' in ex.stderr:
                with excutils.save_and_reraise_exception():
                    self.sshpool_available = True
                    LOG.error(_("Error running SSH command: %s"),
                              command)
            else:
                with excutils.save_and_reraise_exception():
                    LOG.error("No connection found in ssh pool. Please check the network")
        except Exception:
            # Need to check if creating an SSHPool instorage_san_secondary_ip
            # before raising an error.
            try:
                self.sshpool_available = False
                if self.sshpool:
                    del self.sshpool
                    self.sshpool = None
                if self._switch_ip(temp_active_ip):
                    self.sshpool = self._set_up_sshpool(self.active_ip)
                    self.sshpool_available = True
                    return self._ssh_execute(self.sshpool, command,
                                check_exit_code, attempts)
                else:
                    self.sshpool_available = True
                    LOG.warning(_('unable to use '
                                  'instorage_san_secondary_ip since it is '
                                  'not configured.'))
                    raise
            except processutils.ProcessExecutionError as ex:
                if ('mcsinq' in ex.cmd or 'mcsop' in ex.cmd) and 'CMMVC' in ex.stderr:
                    with excutils.save_and_reraise_exception():
                        self.sshpool_available = True
                        LOG.error(_("Error running SSH command: %s"),
                                command)
            except Exception:
                with excutils.save_and_reraise_exception():
                    self.sshpool_available = False
                    LOG.error("No connection found in ssh pool. Please check the network")

    def _set_up_sshpool(self, san_ip):
        password = self.configuration.san_password
        privatekey = self.configuration.san_private_key
        if ((self.configuration.instorage_rsa_private_key is not None) and
               (self.configuration.instorage_rsa_private_key.strip() != '')):
            self.rsa_util = RsaUtil(self.configuration.instorage_rsa_private_key)
            password = self.rsa_util.decrypt_by_private_key(password)
        min_size = self.configuration.ssh_min_pool_conn
        max_size = self.configuration.ssh_max_pool_conn
        sshpool = ssh_utils.SSHPool(
            san_ip,
            self.configuration.san_ssh_port,
            self.configuration.ssh_conn_timeout,
            self.configuration.san_login,
            password=password,
            privatekey=privatekey,
            min_size=min_size,
            max_size=max_size)
        return sshpool

    def _ssh_execute(self, sshpool, command,
                     check_exit_code=True, attempts=1):
        try:
            with sshpool.item() as ssh:
                last_exception = None
                while attempts > 0:
                    attempts -= 1
                    try:
                        out, err = processutils.ssh_execute(
                            ssh,
                            command,
                            timeout=self.configuration.ssh_exec_timeout,
                            check_exit_code=check_exit_code)
                        if err and 'CMMVC' not in err:
                            LOG.error('ignore storage error: %s' % err)
                            err = ""
                        return out, err
                    except paramiko.ssh_exception.SSHException:
                        with excutils.save_and_reraise_exception():
                            LOG.error('execute a ssh command failed'
                                      'due to the SSHException')
                    except Exception as exp:
                        last_exception = exp
                        greenthread.sleep(self.DEFAULT_GR_SLEEP)
                try:
                    raise processutils.ProcessExecutionError(
                        exit_code=last_exception.exit_code,
                        stdout=last_exception.stdout,
                        stderr=last_exception.stderr,
                        cmd=last_exception.cmd)
                except AttributeError:
                    raise processutils.ProcessExecutionError(
                        exit_code=-1,
                        stdout="",
                        stderr="Error running SSH command",
                        cmd=command)
        except processutils.ProcessExecutionError:
            with excutils.save_and_reraise_exception():
                LOG.error(_("Error running SSH command: %s"), command)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.error("No connection found in ssh pool. Please check the network")

    def _switch_ip(self,temp_active_ip):
        # Change active_ip if instorage_san_secondary_ip is set.
        if self.configuration.instorage_san_secondary_ip is None:
            return False
        if temp_active_ip == self.active_ip:
            self.inactive_ip, self.active_ip = self.active_ip, self.inactive_ip
            LOG.info(_('Switch active_ip from %(old)s to '
                       '%(new)s.'),
                     {'old': self.inactive_ip,
                      'new': self.active_ip})
        return True

    def _run_rest(self, method, params=None,
                        request_type='get',
                        fail_handler=None):
        if not self._rest:
            self._rest = RestExecutor(self.active_ip,
                                      self.port,
                                      self.configuration.san_login,
                                      self.configuration.san_password,
                                      ca_root=self.ca_root,
                                      ca_key=self.ca_key,
                                      ca_cert=self.ca_cert,
                                      use_ssl=self.use_ssl,
                                      check_root_cert=self.check_root_cert,
                                      check_client_cert=self.check_client_cert,
                                      host_identifier=self.host_identifier,
                                      thread_num=self.configuration.instorage_mcs_thread_num,
                                      cluster_id = self.cluster_id)

        return self._rest_execute(self._rest, method, params, request_type, fail_handler)

    def _rest_execute(self, rest_client, method,
                     params=None, request_type='get',
                     fail_handler=None):
        return rest_client.send_rest_api(method, params, request_type, fail_handler)

    def generate_volume_name(self, volume):
        if not hasattr(volume,'_name_id'):
            return "snapshot-%s" % volume['id']
        else:
            return "volume-%s" % (volume['id'] if not volume['_name_id'] else volume['_name_id'])

    def generate_snapshot_name(self, snapshot):
        return "snapshot-%s" % snapshot['id']

    def generate_rccg_name(self, group):
        return  ("rccg-%s" % group.id)[0:15]

    def ensure_export(self, ctxt, volume):
        """Check that the volume exists on the storage."""
        vol_name = self._get_target_vol(volume)
        volume_defined = self._assistant.is_vdisk_defined(vol_name)

        if not volume_defined:
            LOG.error(_('ensure_export: Volume %s not found on storage.'),
                      vol_name)

    # ATTENTION the signature of create_export change from L, which add
    # the connector argument, we set it to None as default so that J, K
    # can also use it
    def create_export(self, ctxt, volume, connector=None):
        pass

    def remove_export(self, ctxt, volume):
        pass

    def create_export_snapshot(self, context, snapshot, connector=None):
        pass

    def remove_export_snapshot(self, context, snapshot):
        pass

    def _get_vdisk_params(self, type_id, volume_type=None,
                          volume_metadata=None):
        return self._assistant.get_vdisk_params(
            self.configuration,
            self._state,
            type_id,
            volume_type=volume_type,
            volume_metadata=volume_metadata)

    def create_volume(self, volume):
        vol_name = self.generate_volume_name(volume)
        LOG.debug('enter: create_volume: volume %s', vol_name)
        opts = self._get_vdisk_params(
            volume['volume_type_id'],
            volume_metadata=volume.get('volume_metadata'))
        pool = volume_utils.extract_host(volume['host'], 'pool')

        opts['iogrp'] = self._assistant.select_io_group(self._state, opts)

        if opts['aa']:
            self._assistant.create_volume(
                vol_name, str(volume['size']), 'gb', pool, opts
            )
        else:
            self._assistant.create_vdisk(
                vol_name, str(volume['size']), 'gb', pool, opts,
                self.configuration.instorage_mcs_vol_nofmtdisk
            )

        if opts['qos']:
            self._assistant.add_vdisk_qos(vol_name, opts['qos'])

        ctxt = context.get_admin_context()
        rep_type = self._get_volume_replicated_type(ctxt, volume)

        if rep_type:
            group_name = None
            if volume.group_id:
                if not volume_utils.is_group_a_type(
                        volume.group, 'consistent_group_replication_enabled'):
                    msg = _('Create volume with a replication '
                    'group_id is not supported. Please add volume to '
                    'group after volume creation.')
                    LOG.error(msg)
                    raise exception.VolumeDriverException(reason=msg)
                group_name = self.generate_rccg_name(volume.group)

            replica_obj = self._get_replica_obj(rep_type)
            replica_obj.volume_replication_setup(ctxt, volume, group_name)
            model_update = {'replication_status': 'enabled'}
        else:
            model_update = {'replication_status': 'not-capable'}

        model_update['is_active_active'] = opts['aa']
        LOG.debug('leave: create_volume:\n volume: %(vol)s\n '
                  'model_update %(model_update)s',
                  {'vol': vol_name,
                   'model_update': model_update})
        return model_update

    def create_volume_from_snapshot(self, volume, snapshot):
        vol_name = self.generate_volume_name(volume)
        snap_name = self.generate_snapshot_name(snapshot)

        if snapshot['volume_size'] > volume['size']:
            msg = (_("create_volume_from_snapshot: snapshot %(snapshot_name)s "
                     "size is %(snapshot_size)dGB and doesn't fit in target "
                     "volume %(volume_name)s of size %(volume_size)dGB.") %
                   {'snapshot_name': snap_name,
                    'snapshot_size': snapshot['volume_size'],
                    'volume_name': vol_name,
                    'volume_size': volume['size']})
            LOG.error(msg)
            raise exception.InvalidInput(message=msg)

        opts = self._get_vdisk_params(
            volume['volume_type_id'],
            volume_metadata=volume.get('volume_metadata'))
        pool = volume_utils.extract_host(volume['host'], 'pool')
        self._assistant.create_copy(snap_name, vol_name,
                                    snapshot['id'], self.configuration,
                                    opts, True, pool=pool, nofmtdisk=True)

        if volume['size'] > snapshot['volume_size']:
            # extend the new created target volume to expected size.
            self._extend_volume_op(volume, volume['size'], snapshot['volume_size'])

        if opts['qos']:
            self._assistant.add_vdisk_qos(vol_name, opts['qos'])

        if opts['aa']:
            # what we need is an active-active volume
            # we need wait clone finish and then
            # transfter that vdisk to a volume
            self._assistant.ensure_vdisk_no_lc_mappings(vol_name, allow_snaps=False)
            self._assistant.transfer_vdisk_to_volume(vol_name, pool, opts)

        ctxt = context.get_admin_context()
        rep_type = self._get_volume_replicated_type(ctxt, volume)

        model_update = dict()
        if rep_type:
            self._validate_replication_enabled()
            replica_obj = self._get_replica_obj(rep_type)
            replica_obj.volume_replication_setup(ctxt, volume)
            model_update['replication_status'] = 'enabled'
        else:
            model_update = {'replication_status': 'not-capable'}

        model_update['is_active_active'] = opts['aa']
        return model_update

    def create_cloned_volume(self, tgt_volume, src_volume):
        """Creates a clone of the specified volume."""
        src_vol_name = self.generate_volume_name(src_volume)
        tgt_vol_name = self.generate_volume_name(tgt_volume)

        if src_volume['size'] > tgt_volume['size']:
            msg = (_("create_cloned_volume: source volume %(src_vol)s "
                     "size is %(src_size)dGB and doesn't fit in target "
                     "volume %(tgt_vol)s of size %(tgt_size)dGB.") %
                   {'src_vol': src_vol_name,
                    'src_size': src_volume['size'],
                    'tgt_vol': tgt_vol_name,
                    'tgt_size': tgt_volume['size']})
            LOG.error(msg)
            raise exception.InvalidInput(message=msg)

        opts = self._get_vdisk_params(
            tgt_volume['volume_type_id'],
            volume_metadata=tgt_volume.get('volume_metadata'))
        pool = volume_utils.extract_host(tgt_volume['host'], 'pool')
        self._assistant.create_copy(src_vol_name, tgt_vol_name,
                                    src_volume['id'], self.configuration,
                                    opts, True, pool=pool, nofmtdisk=True)

        # The source volume size is equal to target volume size
        # in most of the cases. But in some scenarios, the target
        # volume size may be bigger than the source volume size.
        # InStorage does not support localcopy between two volumes
        # with two different sizes. So InStorage will copy volume
        # from source volume first and then extend target
        # volume to original size.
        if tgt_volume['size'] > src_volume['size']:
            # extend the new created target volume to expected size.
            self._extend_volume_op(tgt_volume, tgt_volume['size'], src_volume['size'])

        if opts['qos']:
            self._assistant.add_vdisk_qos(tgt_vol_name, opts['qos'])

        if opts['aa']:
            # what we need is an active-active volume
            # we need wait clone finish and then
            # transfter that vdisk to a volume
            self._assistant.ensure_vdisk_no_lc_mappings(tgt_vol_name, allow_snaps=False)
            self._assistant.transfer_vdisk_to_volume(tgt_vol_name, pool, opts)

        ctxt = context.get_admin_context()
        rep_type = self._get_volume_replicated_type(ctxt, tgt_volume)

        model_update = dict()
        if rep_type:
            self._validate_replication_enabled()
            replica_obj = self._get_replica_obj(rep_type)
            replica_obj.volume_replication_setup(ctxt, tgt_volume)
            model_update['replication_status'] = 'enabled'
        else:
            model_update = {'replication_status': 'not-capable'}

        model_update['is_active_active'] = opts['aa']
        return model_update

    def extend_volume(self, volume, new_size):
        volume_name = self._get_target_vol(volume)
        is_aa_vdisk = self._assistant.is_aa_vdisk(volume_name)
        if is_aa_vdisk:
            opts = self._get_vdisk_params(
                volume['volume_type_id'],
                volume_metadata=volume.get('volume_metadata'))
            pool = volume_utils.extract_host(volume['host'], 'pool')
            # first transfer it to a general vdisk then do extend
            self._assistant.transfer_volume_to_vdisk(volume_name, pool)

        self._extend_volume_op(volume, new_size)

        if is_aa_vdisk:
            # transfer back to active-active volume
            self._assistant.transfer_vdisk_to_volume(volume_name, pool, opts)

    def _extend_volume_op(self, volume, new_size, old_size=None):
        LOG.debug('enter: _extend_volume_op: volume %s', volume['id'])
        volume_name = self._get_target_vol(volume)
        if not self._state['cloned_extend_support']:
            ret = self._assistant.ensure_vdisk_no_lc_mappings(volume_name,
                                                          allow_snaps=False)
            if not ret:
                msg = (_('_extend_volume_op: Extending a volume with snapshots is '
                     'not supported.'))
                LOG.error(msg)
                raise exception.VolumeDriverException(message=msg)

        ret = self._assistant.ensure_vdisk_not_formatting(volume_name)
        if not ret:
            msg = (_('_extend_volume_op: Extending a formatting volume is '
                     'not supported.'))
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)

        if old_size is None:
            old_size = volume['size']
        extend_amt = int(new_size) - old_size

        rel_info = self._assistant.get_relationship_info(volume_name)
        if rel_info:
            LOG.warning(_('_extend_volume_op: Extending a volume with '
                          'remote copy is not recommended.'))
            try:
                vol_name = self.generate_volume_name(volume)

                tgt_vol = instorage_const.REPLICA_AUX_VOL_PREFIX + vol_name
                rep_type = rel_info['copy_type']
                self._local_backend_assistant.delete_relationship(vol_name)
                self._local_backend_assistant.extend_vdisk(vol_name,
                                                           extend_amt)
                self._aux_backend_assistant.extend_vdisk(tgt_vol, extend_amt)
                tgt_sys = self._aux_backend_assistant.get_system_info()
                self._local_backend_assistant.create_relationship(
                    vol_name, tgt_vol, tgt_sys.get('system_name'),
                    True if instorage_const.ASYNC == rep_type else False)
            except Exception as exp:
                msg = (_('Failed to extend a volume with remote copy '
                         '%(volume)s. Exception: '
                         '%(err)s.') % {'volume': volume['id'],
                                        'err': six.text_type(exp)})
                LOG.error(msg)
                raise exception.VolumeDriverException(message=msg)
        else:
            self._assistant.extend_vdisk(volume_name, extend_amt)
        LOG.debug('leave: _extend_volume_op: volume %s', volume['id'])

    def delete_volume(self, volume):
        vol_name = self.generate_volume_name(volume)

        LOG.debug('enter: delete_volume: volume %s', vol_name)
        ctxt = context.get_admin_context()

        rep_type = self._get_volume_replicated_type(ctxt, volume)
        if rep_type:
            self._aux_backend_assistant.delete_rc_volume(vol_name,
                                                         target_vol=True)
            if not self._active_backend_id:
                self._local_backend_assistant.delete_rc_volume(vol_name)
            else:
                # If it's in fail over state, also try to delete the volume
                # in master backend
                try:
                    self._local_backend_assistant.delete_rc_volume(vol_name)
                except Exception as ex:
                    LOG.error(_('Failed to get delete volume %(volume)s in '
                                'master backend. Exception: %(err)s.'),
                              {'volume': vol_name,
                               'err': six.text_type(ex)})
        else:
            if self._active_backend_id:
                msg = (_('Error: delete non-replicate volume in failover mode'
                         ' is not allowed.'))
                LOG.error(msg)
                raise exception.VolumeDriverException(message=msg)
            else:
                self._assistant.delete_vdisk(vol_name, False)

        if volume['id'] in self._vdiskcopyops:
            del self._vdiskcopyops[volume['id']]

            if not self._vdiskcopyops:
                self._vdiskcopyops_loop.stop()
                self._vdiskcopyops_loop = None
        LOG.debug('leave: delete_volume: volume %s', vol_name)

    def format_volume(self, volume, fstype='ext4'):
        uuid = None
        attach_info = self._attach_device(volume)
        backup_device_path = attach_info['device']['path']
        try:
            # format device
            root_helper = cinder_utils.get_root_helper()
            if fstype in ['ext2','ext3','ext4']:
                cmd = ['mkfs', '-t', fstype, '-F','-m0',backup_device_path]
            elif fstype in ['xfs']:
                cmd = ['mkfs', '-t', fstype, '-f',backup_device_path]
            processutils.execute(*cmd, root_helper=root_helper, run_as_root=True)
            # get uuid of partition
            cmd = ['blkid', backup_device_path]
            stdout, _ = processutils.execute(*cmd, root_helper=root_helper, run_as_root=True)
            if 'PTUUID' in stdout:
                uuid_str = stdout.split(' ')[1]
                uuid = uuid_str[8:len(uuid_str)-1]
            elif 'UUID' in stdout:
                uuid_str = stdout.split(' ')[1]
                uuid = uuid_str[6:len(uuid_str)-1]
        except Exception as err:
            with excutils.save_and_reraise_exception():
                LOG.error(
                    'Format device %s failed for %s.',
                    volume['id'],
                    err
                )
        self._detach_device(volume, attach_info)
        return uuid

    def _attach_device(self, volume):
        properties = cinder_utils.brick_get_connector_properties()
        conn = self.initialize_connection(volume, properties)
        connector = cinder_utils.brick_get_connector(
            conn['driver_volume_type'],
            use_multipath=properties.get('multipath', False),
            device_scan_attempts=5,
            conn=conn,
            expect_raw_disk=True
        )
        vol_handle = connector.connect_volume(conn['data'])

        attach_info = {
            'conn': conn,
            'device': vol_handle,
            'connector': connector,
            'properties': properties
        }
        return attach_info

    def _detach_device(self, volume, attach_info):
        connector = attach_info['connector']
        connector.disconnect_volume(
            attach_info['conn']['data'],
            attach_info['device'],
        )
        self.terminate_connection(volume, attach_info['properties'])

    def create_snapshot(self, snapshot):
        ctxt = context.get_admin_context()
        try:
            source_vol = self.db.volume_get(ctxt, snapshot['volume_id'])
        except Exception:
            msg = (_('create_snapshot: get source volume failed.'))
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)
        pool = volume_utils.extract_host(source_vol['host'], 'pool')
        opts = self._get_vdisk_params(source_vol['volume_type_id'])
        snap_name = self.generate_snapshot_name(snapshot)
        snap_vol_name = self.generate_volume_name(snapshot['volume'])
        if opts['aa']:
            rep_info = self._assistant.get_relationship_info(snap_vol_name)
            if rep_info:
                if rep_info['primary'] == 'master':
                    if 'master_vdisk_name' in rep_info:
                        snap_vol_name = rep_info['master_vdisk_name']
                    else:
                        snap_vol_name = rep_info['masterVdiskName']
                elif rep_info['primary'] == 'aux':
                    if 'aux_vdisk_name' in rep_info:
                        snap_vol_name = rep_info['aux_vdisk_name']
                    else:
                        snap_vol_name = rep_info['auxVdiskName']
            else:
                msg = ('failed to create a snapshot for A-A volume due to missing rcrelationship')
                LOG.error(msg)
                raise exception.VolumeDriverException(message=msg)

        opts['rsize'] = 0
        self._assistant.create_copy(snap_vol_name, snap_name,
                                    snapshot['volume_id'], self.configuration,
                                    opts, False, pool=pool)

    def delete_snapshot(self, snapshot):
        snap_name = self.generate_snapshot_name(snapshot)
        self._assistant.delete_vdisk(snap_name, False)

    def rollback_snapshot(self, snapshot, volume):
        """rollback_snapshot is a tiexinCloud specific interface."""
        ctxt = context.get_admin_context()
        self.revert_to_snapshot(ctxt,volume,snapshot)

    def rollback_snapshot(self, context, snapshot, volume=None):
        """rollback_snapshot is a Huawei FusionSphere specific interface."""
        src_volume = volume
        if not src_volume:
            src_volume = snapshot['volume']
        self.revert_to_snapshot(context, src_volume, snapshot)

    def rollback_snapshot(self, volume, snapshot):
        ctxt = context.get_admin_context()
        self.revert_to_snapshot(ctxt, volume, snapshot)

    def revert_to_snapshot(self, context, volume, snapshot):
        if snapshot['volume_size'] != volume['size']:
            msg = ('Revert volume size must equal with the snapshot size')
            raise exception.InvalidInput(reason=msg)

        rep_type = self._get_volume_replicated_type(context, volume)
        if rep_type:
            msg = ('Reverting replication volume is not supported.')
            raise exception.InvalidInput(reason=msg)

        vol_name = self.generate_volume_name(volume)
        snap_name = self.generate_snapshot_name(snapshot)
        is_aa_vdisk = self._assistant.is_aa_vdisk(vol_name)

        primary = None
        if is_aa_vdisk:
            rep_info = self._assistant.get_relationship_info(vol_name)
            if rep_info:
                if rep_info['state'] not in['consistent_synchronized',
                                            'consistent_stopped', 'idling']:
                    msg = (('the state of the rcrelationship of %(vol_name)s is %(state)s') %
                           {'vol_name': vol_name,
                            'state': rep_info['state']})
                    raise exception.VolumeDriverException(message=msg)
                _, tgt_mappings = self._assistant.get_vdisk_lc_mappings(
                                                  snap_name, type='target')
                if not tgt_mappings or len(tgt_mappings) > 1:
                    msg = ('unexpect error: snapshot volume is not'
                            'in a lcmap or in more than one lcmap')
                    LOG.error(msg)
                    raise exception.VolumeDriverException(message=msg)
                lcmap = tgt_mappings[0]
                snap_vol_name = (lcmap['source_vdisk_name'] if 'source_vdisk_name' in lcmap
                                                            else lcmap['sourceVdiskName'])
                if snap_vol_name.startswith('volume-'):
                    primary = 'master'
                else:
                    primary = 'aux'
                self._assistant.stop_relationship(vol_name, True)
            else:
                msg = ('failed to restore a snapshot for A-A volume'
                       'due to missing rcrelationship')
                LOG.error(msg)
                raise exception.VolumeDriverException(message=msg)
        else:
            src_mappings, tgt_mappings = self._assistant.get_vdisk_lc_mappings(
                                                        vol_name, type='target')
            for mappings in [src_mappings, tgt_mappings]:
                for attrs in mappings:
                    map_id = attrs['id']
                    status = attrs['status']
                    progress = attrs['progress']
                    source = (attrs['source_vdisk_name'] if 'source_vdisk_name' in attrs
                                                         else attrs['sourceVdiskName'])
                    target = (attrs['target_vdisk_name'] if 'target_vdisk_name' in attrs
                                                         else attrs['targetVdiskName'])
                    if source == snap_name:
                        continue
                    if progress not in [100,'100']:
                        msg = ('the source volume of the snapshot is a target'
                                'in a clone rcrelationship'
                                'and the clone progress is not reach 100%,'
                                'please try again after a few minites')
                        raise exception.VolumeDriverException(message=msg)
                    elif (status == 'copying' and progress in [100,'100']):
                        # stop and delete lcmap
                        temp_src_mappings, _ = self._assistant.get_vdisk_lc_mappings(
                                                   target, type='source')
                        if temp_src_mappings:
                            self._assistant.stop_lcmap(map_id, split=True)
                        else:
                            self._assistant.stop_lcmap(map_id, split=False)

                        attrs = self._assistant.get_localcopy_mapping_attributes(map_id)
                        if attrs and attrs['status'] == 'idle_or_copied':
                            try:
                                self._assistant.rm_lcmap(map_id)
                            except exception.VolumeBackendAPIException as ex:
                                if 'CMMVC6318E' not in ex.msg:
                                    raise exception.VolumeDriverException(
                                        message=ex.msg)
                        else:
                            msg = ('the clean progress is not reach 100%,'
                                   'please try again after a few minites')
                            raise exception.VolumeDriverException(message=msg)
                    elif status =='idle_or_copied':
                        try:
                            self._assistant.rm_lcmap(map_id)
                        except exception.VolumeBackendAPIException as ex:
                            if 'CMMVC6318E' not in ex.msg:
                                raise exception.VolumeDriverException(
                                    message=ex.msg)

        try:
            self._assistant.run_localcopy(
                snap_name,
                vol_name,
                self.configuration.instorage_mcs_localcopy_timeout,
                self.configuration.instorage_mcs_localcopy_rate,
                self.configuration.instorage_mcs_localcopy_clean_rate,
                full_copy=True,
                restore=True
            )
            self._assistant.ensure_rollback_snapshot_ok(vol_name, snap_name)
            if is_aa_vdisk:
                self._assistant.start_relationship(vol_name, primary)
        except Exception as err:
            with excutils.save_and_reraise_exception():
                LOG.error(
                    'Revert to snapshot from %s to %s failed for %s.',
                    snap_name,
                    vol_name,
                    err
                )

    def add_vdisk_copy(self, volume, dest_pool, vol_type):
        return self._assistant.add_vdisk_copy(volume, dest_pool,
                                              vol_type, self._state,
                                              self.configuration)

    def _add_vdisk_copy_op(self, ctxt, volume, new_op):
        if volume['id'] in self._vdiskcopyops:
            self._vdiskcopyops[volume['id']].append(new_op)
        else:
            self._vdiskcopyops[volume['id']] = [new_op]

        # We added the first copy operation, so start the looping call
        if len(self._vdiskcopyops) == 1:
            self._vdiskcopyops_loop = loopingcall.FixedIntervalLoopingCall(
                self._check_volume_copy_ops)
            self._vdiskcopyops_loop.start(interval=self.VDISKCOPYOPS_INTERVAL)

    def _rm_vdisk_copy_op(self, ctxt, volume, orig_copy_id, new_copy_id):
        try:
            self._vdiskcopyops[volume['id']].remove((orig_copy_id,
                                                     new_copy_id))
            if not self._vdiskcopyops[volume['id']]:
                del self._vdiskcopyops[volume['id']]
            if not self._vdiskcopyops:
                self._vdiskcopyops_loop.stop()
                self._vdiskcopyops_loop = None
        except KeyError:
            LOG.error(_('_rm_vdisk_copy_op: Volume %s does not have any '
                        'registered vdisk copy operations.'), volume['id'])
            return
        except ValueError:
            LOG.error(_('_rm_vdisk_copy_op: Volume %(vol)s does not have '
                        'the specified vdisk copy operation: orig=%(orig)s '
                        'new=%(new)s.'),
                      {'vol': volume['id'], 'orig': orig_copy_id,
                       'new': new_copy_id})
            return

    def _check_volume_copy_ops(self):
        LOG.debug("Enter: update volume copy status.")
        ctxt = context.get_admin_context()
        copy_items = list(self._vdiskcopyops.items())
        for vol_id, copy_ops in copy_items:
            try:
                volume = self.db.volume_get(ctxt, vol_id)
            except Exception:
                LOG.warning(_('Volume %s does not exist.'), vol_id)
                del self._vdiskcopyops[vol_id]
                if not self._vdiskcopyops:
                    self._vdiskcopyops_loop.stop()
                    self._vdiskcopyops_loop = None
                continue

            vol_name = self.generate_volume_name(volume)
            for copy_op in copy_ops:
                try:
                    synced = self._assistant.check_vdisk_copy_synced(
                        vol_name, copy_op[1])
                except Exception:
                    LOG.info(_('_check_volume_copy_ops: Volume %(vol)s does '
                               'not have the specified vdisk copy '
                               'operation: orig=%(orig)s new=%(new)s.'),
                             {'vol': volume['id'], 'orig': copy_op[0],
                              'new': copy_op[1]})
                else:
                    if synced:
                        self._assistant.rm_vdisk_copy(vol_name, copy_op[0])
                        self._rm_vdisk_copy_op(ctxt, volume, copy_op[0],
                                               copy_op[1])
        LOG.debug("Exit: update volume copy status.")

    def migrate_volume(self, ctxt, volume, host):
        """Migrate directly if source and dest are managed by same storage.

        We create a new vdisk copy in the desired pool, and add the original
        vdisk copy to the admin_metadata of the volume to be deleted. The
        deletion will occur using a periodic task once the new copy is synced.

        :param ctxt: Context
        :param volume: A dictionary describing the volume to migrate
        :param host: A dictionary describing the host to migrate to, where
                     host['host'] is its name, and host['capabilities'] is a
                     dictionary of its reported capabilities.
        """
        LOG.debug('enter: migrate_volume: id=%(id)s, host=%(host)s',
                  {'id': volume['id'], 'host': host['host']})
        false_ret = (False, None)
        volume_name = self.generate_volume_name(volume)
        # migrate a InMetro volume must be through the host side
        if self._assistant.is_aa_vdisk(volume_name):
            msg = _('migrate: migrate a InMetro volume must be through the host side')
            LOG.warning(msg)
            return false_ret

        dest_pool = self._assistant.can_migrate_to_host(host, self._state)
        if dest_pool is None:
            return false_ret

        ctxt = context.get_admin_context()
        volume_type_id = volume['volume_type_id']
        if volume_type_id is not None:
            vol_type = volume_types.get_volume_type(ctxt, volume_type_id)
        else:
            vol_type = None

        # migrate volume with vdisk copy and without vdisk copy
        resp = self._assistant.executor.lsvdiskcopy(volume_name)
        if len(resp) > 1:
            copies = self._assistant.get_vdisk_copies(volume_name)
            src_pool = copies['primary']['mdisk_grp_name']
            if self._assistant.is_ssa_pool(src_pool):
                msg = _('Unable to migrate: the SSA '
                        'volume with volume copy can not be migrated. ')
                raise exception.VolumeDriverException(message=msg)
            self._assistant.migratevdisk(volume_name, dest_pool,
                                         copies['primary']['copy_id'])
        else:
            self._check_volume_copy_ops()
            new_op = self.add_vdisk_copy(volume_name, dest_pool, vol_type)
            self._add_vdisk_copy_op(ctxt, volume, new_op)

        LOG.debug('leave: migrate_volume: id=%(id)s, host=%(host)s',
                  {'id': volume['id'], 'host': host['host']})
        return (True, None)

    def retype(self, ctxt, volume, new_type, diff, host):
        """Convert the volume to be of the new type.

        Returns a boolean indicating whether the retype occurred.

        :param ctxt: Context
        :param volume: A dictionary describing the volume to migrate
        :param new_type: A dictionary describing the volume type to convert to
        :param diff: A dictionary with the difference between the two types
        :param host: A dictionary describing the host to migrate to, where
                     host['host'] is its name, and host['capabilities'] is a
                     dictionary of its reported capabilities.
        """
        def retype_iogrp_property(volume, new, old):
            if new != old:
                vol_name = self.generate_volume_name(volume)
                self._assistant.change_vdisk_iogrp(vol_name,
                                                   self._state, (new, old))

        LOG.debug('enter: retype: id=%(id)s, new_type=%(new_type)s,'
                  'diff=%(diff)s, host=%(host)s', {'id': volume['id'],
                                                   'new_type': new_type,
                                                   'diff': diff,
                                                   'host': host})
        volume_name = self.generate_volume_name(volume)
        # InMetro volume does not support migrate
        if self._assistant.is_aa_vdisk(volume_name):
            msg = _('retype: retyping a InMetro volume must be through the host side')
            LOG.warning(msg)
            return False

        need_copy = False
        # check whether the dest_pool is on the same storage with this driver
        dest_pool = self._assistant.can_migrate_to_host(host, self._state)
        if dest_pool is None:
            return False

        # check whether pool changes
        src_pool = volume_utils.extract_host(volume['host'], 'pool')
        dest_pool = volume_utils.extract_host(host['host'], 'pool')
        if src_pool != dest_pool:
            need_copy = True

        old_opts = self._get_vdisk_params(
            volume['volume_type_id'],
            volume_metadata=volume.get('volume_matadata')
        )
        new_opts = self._get_vdisk_params(
            new_type['id'], volume_type=new_type
        )

        # cache change attribute when no need disk copy
        vdisk_changes = []
        no_copy_keys = []
        copy_keys = []
        if not need_copy:
            is_ssa_pool = self._assistant.is_ssa_pool(dest_pool)
            if is_ssa_pool:
                no_copy_keys = ['warning', 'autoexpand','deduped', 'compressed']
                copy_keys = []
            else:
                no_copy_keys = ['warning', 'autoexpand', 'intier']
                copy_keys = ['rsize', 'grainsize', 'compressed']

            all_keys = no_copy_keys + copy_keys

            for key in all_keys:
                if old_opts[key] != new_opts[key]:
                    if key in copy_keys:
                        need_copy = True
                        break
                    elif key in no_copy_keys:
                        vdisk_changes.append(key)
        vol_name = self.generate_volume_name(volume)
        # Check if retype affects volume replication
        model_update = None
        new_rep_type = self._get_specs_replicated_type(new_type)
        old_rep_type = self._get_volume_replicated_type(ctxt, volume)
        old_io_grp = self._assistant.get_volume_io_group(vol_name)

        # There are three options for rep_type: None, sync, async
        if new_rep_type != old_rep_type:
            if (old_io_grp not in
                    InStorageAssistant.get_valid_requested_io_groups(
                        self._state, new_opts)):
                msg = (_('Unable to retype: it is not allowed to change '
                         'replication type and io group at the same time.'))
                LOG.error(msg)
                raise exception.VolumeDriverException(message=msg)
            if new_rep_type and old_rep_type:
                msg = (_('Unable to retype: it is not allowed to change '
                         '%(old_rep_type)s volume to %(new_rep_type)s '
                         'volume.') %
                       {'old_rep_type': old_rep_type,
                        'new_rep_type': new_rep_type})
                LOG.error(msg)
                raise exception.VolumeDriverException(message=msg)
            # If volume is replicated, can't copy
            if need_copy:
                msg = (_('Unable to retype: Current action needs volume-copy,'
                         ' it is not allowed when new type is replication.'
                         ' Volume = %s') % volume['id'])
                raise exception.VolumeDriverException(message=msg)

        new_io_grp = self._assistant.select_io_group(self._state, new_opts)

        if need_copy:
            self._check_volume_copy_ops()
            retype_iogrp_property(volume, new_io_grp, old_io_grp)

            try:
                new_op = self.add_vdisk_copy(vol_name,
                                             dest_pool,
                                             new_type)
                self._add_vdisk_copy_op(ctxt, volume, new_op)
            except exception.VolumeDriverException:
                # roll back changing iogrp property
                retype_iogrp_property(volume, old_io_grp, new_io_grp)
                msg = (_('Unable to retype:  A copy of volume %s exists. '
                         'Retyping would exceed the limit of 2 copies.'),
                       volume['id'])
                raise exception.VolumeDriverException(message=msg)
        else:
            retype_iogrp_property(volume, new_io_grp, old_io_grp)

            self._assistant.change_vdisk_options(
                vol_name, vdisk_changes, new_opts
            )

        if new_opts['qos']:
            # Add the new QoS setting to the volume. If the volume has an
            # old QoS setting, it will be overwritten.
            self._assistant.update_vdisk_qos(vol_name, new_opts['qos'])
        elif old_opts['qos']:
            # If the old_opts contain QoS keys, disable them.
            self._assistant.disable_vdisk_qos(vol_name, old_opts['qos'])

        # Delete replica if needed
        if old_rep_type and not new_rep_type:
            self._aux_backend_assistant.delete_rc_volume(vol_name,
                                                         target_vol=True)
            model_update = {'replication_status': 'disabled',
                            'replication_driver_data': None,
                            'replication_extended_status': None}
        # Add replica if needed
        if not old_rep_type and new_rep_type:
            replica_obj = self._get_replica_obj(new_rep_type)
            replica_obj.volume_replication_setup(ctxt, volume)
            model_update = {'replication_status': 'enabled'}

        LOG.debug('exit: retype: ild=%(id)s, new_type=%(new_type)s,'
                  'diff=%(diff)s, host=%(host)s', {'id': volume['id'],
                                                   'new_type': new_type,
                                                   'diff': diff,
                                                   'host': host['host']})
        return True, model_update

    def update_migrated_volume(self, ctxt, volume, new_volume,
                               original_volume_status):
        """Return model update from InStorage for migrated volume.

        This method should rename the back-end volume name(id) on the
        destination host back to its original name(id) on the source host.

        :param ctxt: The context used to run the method update_migrated_volume
        :param volume: The original volume that was migrated to this backend
        :param new_volume: The migration volume object that was created on
                           this backend as part of the migration process
        :param original_volume_status: The status of the original volume
        :returns: model_update to update DB with any needed changes
        """
        name_id = None
        provider_location = None
        if original_volume_status == 'in-use':
            name_id = new_volume['_name_id'] or new_volume['id']
            provider_location = new_volume['provider_location']
            return {'_name_id': name_id,
                    'provider_location': provider_location}

        existing_name = self.generate_volume_name(new_volume)
        # wanted_name = self.generate_volume_name(volume)
        wanted_name = "volume-%s" % volume['id']
        vdisk_exist_flag = self._assistant.is_vdisk_defined(wanted_name)
        try:
            if not vdisk_exist_flag:
                self._assistant.rename_vdisk(str(existing_name),
                                             str(wanted_name))
            else:
                name_id = new_volume['_name_id'] or new_volume['id']
                provider_location = new_volume['provider_location']
        except exception.VolumeBackendAPIException:
            LOG.error(_('Unable to rename the logical volume '
                        'for volume: %s'), volume['id'])
            name_id = new_volume._name_id or new_volume.id
            provider_location = new_volume['provider_location']
        # If the back-end name(id) for the volume has been renamed,
        # it is OK for the volume to keep the original name(id) and there is
        # no need to use the column "_name_id" to establish the mapping
        # relationship between the volume id and the back-end volume
        # name(id).
        # Set the key "_name_id" to None for a successful rename.
        return {'_name_id': name_id, 'provider_location': provider_location}

    def manage_existing_snapshot_get_size(self, snapshot, existing_ref):
        """Return size of an existing image for manage_existing.

        :param snapshot:
            snapshot ref info to be set
        :param existing_ref:
            existing_ref is a dictionary of the form:
            {'source-name': <name of snapshot>}
        """
        # Check that the reference is valid
        if not isinstance(existing_ref, dict):
            existing_ref = {"source-name": existing_ref}
        if 'source-name' not in existing_ref:
            reason = _('Reference must contain source-name element.')
            raise exception.ManageExistingInvalidReference(
                existing_ref=existing_ref, reason=reason)
        volume_name = self.generate_volume_name(snapshot.volume)
        snapshot_name = existing_ref['source-name']
        # get all snapshots of the volume
        snapshots_list, _ = self._assistant.get_vdisk_lc_mappings(
                                            volume_name, 'source')

        for snapshot_obj in snapshots_list:
            if snapshot_name == snapshot_obj['target_vdisk_name']:
                snap_attr = self._assistant.get_vdisk_attributes(snapshot_name)
                return int(math.ceil(float(snap_attr['capacity']) / units.Gi))

        kwargs = {'existing_ref': snapshot_name,
                      'reason': 'Specified snapshot does not exist.'}
        raise exception.ManageExistingInvalidReference(**kwargs)

    def manage_existing_snapshot(self, snapshot, existing_ref):
        """Manages an existing snapshot.

        Renames the snapshot name to match the expected name for the snapshot.
        Error checking done by manage_existing_get_size is not repeated.

        :param snapshot:
            snapshot ref info to be set
        :param existing_ref:
            existing_ref is a dictionary of the form:
            {'source-name': <name of snapshot>}
        """
        if not isinstance(existing_ref, dict):
            existing_ref = {"source-name": existing_ref}
        volume_name = self.generate_volume_name(snapshot.volume)
        src_snapshot_name = existing_ref['source-name']
        dest_snapshot_name = self.generate_snapshot_name(snapshot)
        is_existed = False

        src, target = self._assistant.get_vdisk_lc_mappings(
                    src_snapshot_name, type='source')
        if src:
            # Exclude clone snapshot.
            msg = (_('Failed to manage existing snapshot %s,'
                    'due to being used for clone snap') % src_snapshot_name)
            raise exception.InvalidSnapshot(reason=msg)


        # get all snapshots of the volume
        snapshots_list, target = self._assistant.get_vdisk_lc_mappings(
                    volume_name, type='source')

        for snapshot_obj in snapshots_list:
            if src_snapshot_name == snapshot_obj['target_vdisk_name']:
                is_existed = True
                if self._assistant.is_vdisk_in_use(src_snapshot_name):
                    msg = (_('Failed to manage existing snapshot %s,'
                            'due to being in use') % src_snapshot_name)
                    raise exception.InvalidSnapshot(reason=msg)
                else:
                    break
        if not is_existed:
            msg = (_('Failed to manage existing snapshot %s,'
                     'due to snapshot is not existed') % src_snapshot_name)
            raise exception.InvalidSnapshot(reason=msg)
        if dest_snapshot_name != src_snapshot_name:
            self._assistant.rename_vdisk(str(src_snapshot_name),
                                             str(dest_snapshot_name))

    def get_manageable_snapshots(self, cinder_snapshots, marker, limit, offset,
                                 sort_keys, sort_dirs):
        """List manageable snapshots on AS18000 backend."""
        LOG.debug("Listing manageable snapshots")
        results = []
        cinder_snapshot_ids = [resource['id'] for resource in cinder_snapshots]
        new_volume_list = []

        for pool in self.configuration.instorage_mcs_volpool_name:
            volume_list = self._assistant.get_volumes(pool)
            for lun in volume_list:
                volume_tmp = {}
                vol_name = lun['name']
                # get all snapshots of the volume
                snapshots_list, _ = self._assistant.get_vdisk_lc_mappings(
                    vol_name, type='source')
                if not snapshots_list:
                    continue

                volume_tmp = {
                    'name': vol_name,
                    'pool': pool,
                    'snapshotlist': snapshots_list
                }
                new_volume_list.append(volume_tmp)

        for volume in new_volume_list:
            for snapshot in volume['snapshotlist']:
                snapshot_name = snapshot['target_vdisk_name']
                snap_attr = self._assistant.get_vdisk_attributes(snapshot_name)
                size = int(math.ceil(float(snap_attr['capacity']) / units.Gi))
                snapshot_info = {
                    'reference': {'source-name': snapshot_name},
                    'size': size,
                    'cinder_id': None,
                    'extra_info': {'pool': volume['pool']},
                    'source_reference': {'source-name': volume['name']}
                }
                src, _ = self._assistant.get_vdisk_lc_mappings(snapshot_name, 'source')
                if self._assistant.is_vdisk_in_use(snapshot_name):
                    # Exclude snapshots in use.
                    snapshot_info['safe_to_manage'] = False
                    snapshot_info['reason_not_safe'] = 'snapshot in use'
                elif (snapshot_name.startswith('snapshot-')
                        and (snapshot_name.replace('snapshot-', '') in cinder_snapshot_ids)):
                    # Exclude snapshots already managed.
                    snapshot_info['safe_to_manage'] = False
                    snapshot_info['reason_not_safe'] = 'already managed'
                    snapshot_info['cinder_id'] = snapshot_name.replace('snapshot-', '')
                elif src:
                    snapshot_info['safe_to_manage'] = False
                    snapshot_info['reason_not_safe'] = 'it is a cascade snapshot'
                else:
                    snapshot_info['safe_to_manage'] = True
                    snapshot_info['reason_not_safe'] = None

                results.append(snapshot_info)

        return volume_utils.paginate_entries_list(
            results, marker, limit, offset, sort_keys, sort_dirs)

    def unmanage_snapshot(self, snapshot):
        """Removes the specified snapshot from Cinder management."""
        pass

    def manage_existing(self, volume, ref):
        """Manages an existing vdisk.

        Renames the vdisk to match the expected name for the volume.
        Error checking done by manage_existing_get_size is not repeated -
        if we got here then we have a vdisk that isn't in use (or we don't
        care if it is in use.
        """
        # Check that the reference is valid
        vdisk = self._manage_input_check(ref)
        vdisk_io_grp = self._assistant.get_volume_io_group(vdisk['name'])
        if vdisk_io_grp not in self._state['available_iogrps']:
            msg = (_("Failed to manage existing volume due to "
                     "the volume to be managed is not in a valid "
                     "I/O group."))
            raise exception.ManageExistingVolumeTypeMismatch(reason=msg)

        is_aa_vdisk = self._assistant.is_aa_vdisk(vdisk['name'])
        if not is_aa_vdisk and self._replica_enabled:
            # Add replication check
            ctxt = context.get_admin_context()
            rep_type = self._get_volume_replicated_type(ctxt, volume)
            vol_rep_type = None
            rel_info = self._assistant.get_relationship_info(vdisk['name'])
            if rel_info:
                vol_rep_type = rel_info['copy_type']
                aux_info = self._aux_backend_assistant.get_system_info()
                if rel_info['aux_cluster_id'] != aux_info['system_id']:
                    msg = (_("Failed to manage existing volume due to the aux "
                             "cluster for volume %(volume)s is %(aux_id)s. The "
                             "configured cluster id is %(cfg_id)s") %
                           {'volume': vdisk['name'],
                            'aux_id': rel_info['aux_cluster_id'],
                            'cfg_id': aux_info['system_id']})
                    raise exception.ManageExistingVolumeTypeMismatch(reason=msg)

            if vol_rep_type != rep_type:
                msg = (_("Failed to manage existing volume due to "
                         "the replication type of the volume to be managed is "
                         "mismatch with the provided replication type."))
                raise exception.ManageExistingVolumeTypeMismatch(reason=msg)

        # check the pool
        pool = volume_utils.extract_host(volume['host'], 'pool')
        if (('mdisk_grp_name' in vdisk and vdisk['mdisk_grp_name'] != pool) or
                ('mdiskGrpName' in vdisk and vdisk['mdiskGrpName'] != pool)):
            msg = (_("Failed to manage existing volume due to the "
                     "pool of the volume to be managed does not "
                     "match the backend pool. Pool of the "
                     "volume to be managed is %(vdisk_pool)s. Pool "
                     "of the backend is %(backend_pool)s.") %
                   {'vdisk_pool': vdisk['mdisk_grp_name'],
                    'backend_pool':
                        self._get_backend_pools()})
            raise exception.ManageExistingVolumeTypeMismatch(reason=msg)

        is_ssa_pool = self._assistant.is_ssa_pool(pool)

        if volume['volume_type_id']:
            opts = self._get_vdisk_params(
                volume['volume_type_id'],
                volume_metadata=volume.get('volume_metadata')
            )

            vdisk_copy = (self._assistant.get_vdisk_copy_attrs(vdisk['name'], '0') or
                          self._assistant.get_vdisk_copy_attrs(vdisk['name'], '1'))
            if not vdisk_copy:
                msg = (_("Failed to manage existing volume due to the "
                             "vdisk copy of %s can't to be found") % vdisk['name'] )
                raise exception.ManageExistingVolumeTypeMismatch(reason=msg)

            if opts['aa'] and not is_aa_vdisk:
                msg = (_("Failed to manage existing volume due to "
                         "the volume to be managed is general, but "
                         "the volume type chosen is active-active."))
                raise exception.ManageExistingVolumeTypeMismatch(reason=msg)

            if not opts['aa'] and is_aa_vdisk:
                msg = (_("Failed to manage existing volume due to "
                         "the volume to be managed is active-active, but "
                         "the volume type chosen is general."))
                raise exception.ManageExistingVolumeTypeMismatch(reason=msg)

            if opts['aa'] and is_aa_vdisk:
                master, aux = self._assistant.get_volume_pool_iogrp_map(vdisk['name'])
                if self.active_active_pool_map[master['mdisk_grp_name']] != aux['mdisk_grp_name']:
                    msg = (_("Failed to manage existing volume due to "
                             "active-active master aux pool map not match configure set."))
                    raise exception.ManageExistingVolumeTypeMismatch(reason=msg)

                if self.active_active_iogrp_map[master['IO_group_id']] != aux['IO_group_id']:
                    msg = (_("Failed to manage existing volume due to "
                             "active-active master aux iogrp map not match configure set."))
                    raise exception.ManageExistingVolumeTypeMismatch(reason=msg)

            if not is_ssa_pool:
                if (vdisk_copy['autoexpand'] == 'on' and
                        (opts['rsize'] == -1 or opts['rsize'] == 100)):
                    msg = (_("Failed to manage existing volume due to "
                             "the volume to be managed is thin, but "
                             "the volume type chosen is thick."))
                    raise exception.ManageExistingVolumeTypeMismatch(reason=msg)

                if not vdisk_copy['autoexpand'] and (opts['rsize'] != -1 and opts['rsize'] != 100):
                    msg = (_("Failed to manage existing volume due to "
                             "the volume to be managed is thick, but "
                             "the volume type chosen is thin."))
                    raise exception.ManageExistingVolumeTypeMismatch(reason=msg)

            if is_ssa_pool:
                if ((('deduped_copy' in vdisk_copy and vdisk_copy['deduped_copy'] == 'no') or
                     ('dedupedCopy' in vdisk_copy and vdisk_copy['dedupedCopy'] == 'no')) and
                      opts['deduped']):
                    msg = (_("Failed to manage existing volume due to the "
                             "volume to be managed is not deduped, but "
                             "the volume type chosen is deduped."))
                    raise exception.ManageExistingVolumeTypeMismatch(reason=msg)

                if ((('deduped_copy' in vdisk_copy and vdisk_copy['deduped_copy'] == 'yes') or
                     ('dedupedCopy' in vdisk_copy and vdisk_copy['dedupedCopy'] == 'yes')) and
                       not opts['deduped']):
                    msg = (_("Failed to manage existing volume due to the "
                             "volume to be managed is deduped, but "
                             "the volume type chosen is not deduped."))
                    raise exception.ManageExistingVolumeTypeMismatch(reason=msg)

            if ((('compressed_copy' in vdisk_copy and vdisk_copy['compressed_copy'] == 'no') or
                ('compressedCopy' in vdisk_copy and vdisk_copy['compressedCopy'] == 'no')) and
                    opts['compressed']):
                msg = (_("Failed to manage existing volume due to the "
                         "volume to be managed is not compress, but "
                         "the volume type chosen is compress."))
                raise exception.ManageExistingVolumeTypeMismatch(reason=msg)

            if ((('compressed_copy' in vdisk_copy and vdisk_copy['compressed_copy'] == 'yes') or
                ('compressedCopy' in vdisk_copy and vdisk_copy['compressedCopy'] == 'yes')) and
                    not opts['compressed']):
                msg = (_("Failed to manage existing volume due to the "
                         "volume to be managed is compress, but "
                         "the volume type chosen is not compress."))
                raise exception.ManageExistingVolumeTypeMismatch(reason=msg)

            if (vdisk_io_grp not in
                    InStorageAssistant.get_valid_requested_io_groups(
                        self._state, opts)):
                msg = (_("Failed to manage existing volume due to "
                         "I/O group mismatch. The I/O group of the "
                         "volume to be managed is %(vdisk_iogrp)s. I/O group"
                         "of the chosen type is %(opt_iogrp)s.") %
                       {'vdisk_iogrp': vdisk['IO_group_name'],
                        'opt_iogrp': opts['iogrp']})
                raise exception.ManageExistingVolumeTypeMismatch(reason=msg)

        vol_name = self.generate_volume_name(volume)
        model_update = {}
        if vdisk['name'] != vol_name:
            self._assistant.rename_vdisk(vdisk['name'], vol_name)

        if not is_aa_vdisk and self._replica_enabled and vol_rep_type:
            aux_vol = instorage_const.REPLICA_AUX_VOL_PREFIX + vol_name
            self._aux_backend_assistant.rename_vdisk(
                rel_info['aux_vdisk_name'], aux_vol)
            model_update = {'replication_status': 'enabled', ' is_active_active': opts['aa']}
        else:
            model_update = {'is_active_active': opts['aa']}
        return model_update

    def manage_existing_get_size(self, volume, ref):
        """Return size of an existing Vdisk for manage_existing.

        existing_ref is a dictionary of the form:
        {'source-id': <uid of disk>} or
        {'source-name': <name of the disk>}

        Optional elements are:
          'manage_if_in_use':  True/False (default is False)
            If set to True, a volume will be managed even if it is currently
            attached to a host system.
        """

        # Check that the reference is valid
        vdisk = self._manage_input_check(ref)

        # Check if the disk is in use, if we need to.
        manage_if_in_use = ref.get('manage_if_in_use', False)
        if (not manage_if_in_use and
                self._assistant.is_vdisk_in_use(vdisk['name'])):
            reason = _('The specified vdisk is mapped to a host.')
            raise exception.ManageExistingInvalidReference(existing_ref=ref,
                                                           reason=reason)

        return int(math.ceil(float(vdisk['capacity']) / units.Gi))

    def get_manageable_volumes(self, cinder_volumes, marker, limit, offset,
                               sort_keys, sort_dirs):
        """List volumes on the backend available for management by Cinder.
        """
        LOG.debug("Listing manageable volumes")
        results = []
        PREFIX = "volume-"
        cinder_volume_ids = [vol['id'] for vol in cinder_volumes]

        for pool in self.configuration.instorage_mcs_volpool_name:
            volume_list = self._assistant.get_volumes(pool)
            for volume in volume_list:
                vol_name = volume['name']
                src_vol_info = self._assistant.get_vdisk_attributes(vol_name)
                if src_vol_info is None:
                    kwargs = {'reason': 'volume does not exist.'}
                    continue
                size = int(math.ceil(float(src_vol_info['capacity']) / units.Gi))
                rcrel = self._assistant.get_relationship_info(vol_name)
                volume_info = {
                    'reference': {'source-name': vol_name},
                    'size': size,
                    'cinder_id': None,
                    'extra_info': {'pool': pool}
                }
                if (vol_name.startswith(PREFIX) 
                       and (vol_name.replace(PREFIX, '') in cinder_volume_ids)):
                    # Exclude already managed by cinder
                    volume_info['safe_to_manage'] = False
                    volume_info['reason_not_safe'] = 'already managed'
                    volume_info['cinder_id'] = vol_name.replace(PREFIX, '')
                elif self._assistant.is_vdisk_in_use(vol_name):
                    # Exclude volume in use
                    volume_info['safe_to_manage'] = False
                    volume_info['reason_not_safe'] = 'volume in use'
                elif (rcrel and (rcrel['consistency_group_name'] 
                                 or rcrel['consistencyGroupName'])):
                    # Exclude already consistencygroup
                    volume_info['safe_to_manage'] = False
                    volume_info['reason_not_safe'] = 'belong consistencygroup'
                else:
                    volume_info['safe_to_manage'] = True
                    volume_info['reason_not_safe'] = None

                results.append(volume_info)

        page_results = volume_utils.paginate_entries_list(
            results, marker, limit, offset, sort_keys, sort_dirs)

        return page_results

    def unmanage(self, volume):
        """Remove the specified volume from Cinder management."""
        pass

    def get_volume_stats(self, refresh=False):
        """Get volume stats.

        If we haven't gotten stats yet or 'refresh' is True,
        run update the stats first.
        """
        if not self._stats or refresh:
            self._update_volume_stats()

        return self._stats

    def get_volumes_used_size(self, volume=None):
        """get the volumes used size(ICOS private interface)"""
        pools = self.configuration.instorage_mcs_volpool_name
        volume_name = None
        result = {}

        if volume:
            volume_name = self.generate_volume_name(volume)
            resp = self._assistant.get_volumes_used_size(volume_name=volume_name)
            result.update(resp)
        else:
            for pool in pools:
                resp = self._assistant.get_volumes_used_size(pool=pool)
                result.update(resp)
        return result

    # ## Group method ## #
    def create_group(self, context, group):
        """Create a group."""

        support_grps = ['consistent_group_snapshot_enabled',
                        'consistent_group_replication_enabled']
        supported_grp = False
        for grp_spec in support_grps:
            if volume_utils.is_group_a_type(group, grp_spec):
                supported_grp = True
                break
        if not supported_grp:
            LOG.error('Unable to create group: %s is not a supported group '
                      'type.', group.group_type_id)
            raise NotImplementedError()

        if (volume_utils.is_group_a_type(group, 'consistent_group_snapshot_enabled')
                and volume_utils.is_group_a_type(group, 'consistent_group_replication_enabled')):
            LOG.error('Unable to create group: create group with mixed specs is'
                      'not supported.')
            model_update = {'status': 'error'}
            return model_update

        if volume_utils.is_group_a_type(group, 'consistent_group_snapshot_enabled'):
            return self.create_consistencygroup(context, group)

        if volume_utils.is_group_a_type(group, 'consistent_group_replication_enabled'):
            return self._create_replicationgroup(context, group)

    def _create_replicationgroup(self, cxt, group):
        """Create a replication group."""
        LOG.debug("Creating replication group.")
        model_update = {'status': 'available'}
        self._validate_replication_enabled()
        rccg_type = None
        for vol_type_id in group.volume_type_ids:
            replication_type = self._get_volume_replicated_type(
                cxt, None, vol_type_id)
            if not replication_type:
                # An unsupported configuration
                LOG.error('Unable to create group: create consistent '
                            'replication group with non-replication '
                            'volume type is not supported.')
                model_update = {'status': 'error'}
                return model_update
            if not rccg_type:
                rccg_type = replication_type
            elif rccg_type != replication_type:
                # An unsupported configuration
                LOG.error('Unable to create group: create consistent '
                            'replication group with different replication '
                            'type is not supported.')
                model_update = {'status': 'error'}
                return model_update
        group_name = self.generate_rccg_name(group)
        try:
            tgt_sys = self._aux_backend_assistant.get_system_info()
            self._assistant.create_rc_consistgrp(
                group_name, tgt_sys.get('system_id'))
            model_update.update({'replication_status': 'enabled'})
        except exception.VolumeBackendAPIException as ex:
            LOG.error("Failed to create rccg %(rccg)s. "
                        "Exception: %(exception)s.",
                        {'rccg': group_name, 'exception': ex})
            model_update = {'status': 'error'}
        return model_update


    def create_consistencygroup(self, context, group):
        """Create a consistencygroup.

        inspur InStorage will create group until group-snapshot creation,
        db will maintain the volumes and group relationship.
        """
        LOG.debug("Creating consistencygroup.")
        for vol_type_id in group.volume_type_ids:
            replication_type = self._get_volume_replicated_type(
                context, None, vol_type_id)
            if replication_type:
                # An unsupported configuration
                LOG.error('Unable to create group: create consistent snapshot '
                          'group with replication volume type is not supported.')
                model_update = {'status': 'error'}
                return model_update
        model_update = {'status': 'available'}
        return model_update

    def create_group_from_src(self, context, group, volumes,
                              group_snapshot=None, snapshots=None,
                              source_group=None, source_vols=None):
        """Creates a group from source.

        :param context: the context of the caller.
        :param group: the dictionary of the group to be created.
        :param volumes: a list of volume dictionaries in the group.
        :param group_snapshot: the dictionary of the group_snapshot as source.
        :param snapshots: a list of snapshot dictionaries
                in the group_snapshot.
        :param source_group: the dictionary of a group as source.
        :param source_vols: a list of volume dictionaries in the source_group.
        :returns: model_update, volumes_model_update
        """
        LOG.debug('Enter: create_group_from_src.')
        if (not volume_utils.is_group_a_type(
                group, 'consistent_group_snapshot_enabled')
            and not volume_utils.is_group_a_type(
                group, 'consistent_group_replication_enabled')):
            raise NotImplementedError()

        return self.create_consistencygroup_from_src(
                   context, group, volumes, group_snapshot, snapshots,
                   source_group, source_vols)

    def create_consistencygroup_from_src(self, context, group, volumes,
                                         group_snapshot=None, snapshots=None,
                                         source_group=None, source_vols=None):
        if group_snapshot and snapshots:
            group_name = 'lcgroup-' + group_snapshot.id
            sources = snapshots
            src_name_generator = self.generate_snapshot_name
        elif source_group and source_vols:
            group_name = 'lcgroup-' + source_group.id
            sources = source_vols
            src_name_generator = self.generate_volume_name
        else:
            error_msg = _("create_group_from_src must be creating from"
                          " a group snapshot, or a source group.")
            raise exception.InvalidInput(reason=error_msg)

        tgt_name_generator = self.generate_volume_name

        LOG.debug('create_group_from_src: group_name %(group_name)s'
                  ' %(sources)s', {'group_name': group_name,
                                   'sources': sources})
        self._assistant.create_lc_consistgrp(group_name)  # create group
        timeout = self.configuration.instorage_mcs_localcopy_timeout
        model_update, volumes_model = (
            self._assistant.create_group_from_source(group, group_name,
                                                     sources, volumes,
                                                     self._state,
                                                     self.configuration,
                                                     timeout,
                                                     src_name_generator,
                                                     tgt_name_generator))
        if volume_utils.is_group_a_type(
                group, 'consistent_group_replication_enabled'):
            self._validate_replication_enabled()
            rccg_name = self.generate_rccg_name(group)
            try:
                tgt_sys = self._aux_backend_assistant.get_system_info()
                self._assistant.create_rc_consistgrp(
                    rccg_name, tgt_sys.get('system_id'))
                model_update.update({'replication_status': 'enabled'})
            except exception.VolumeBackendAPIException as ex:
                LOG.error("Failed to create rccg %(rccg)s. "
                            "Exception: %(exception)s.",
                            {'rccg': rccg_name, 'exception': ex})
                model_update = {'status': 'error'}

        for vol in volumes:
            rep_type = self._get_volume_replicated_type(context, vol)
            volume_model = dict()
            for model in volumes_model:
                if vol.id == model["id"]:
                    volume_model = model
                    break
            if rep_type:
                replica_obj = self._get_replica_obj(rep_type)
                replica_obj.volume_replication_setup(context, vol)
                volume_model['replication_status'] = 'enabled'

            opts = self._get_vdisk_params(
            vol['volume_type_id'],
            volume_metadata=vol.get('volume_metadata'))
            if opts['qos']:
                vol_name = self.generate_volume_name(vol)
                self._assistant.add_vdisk_qos(vol_name, opts['qos'])

        if volume_utils.is_group_a_type(
                group, 'consistent_group_replication_enabled'):
            model_update, added_vols, removed_vols = (
                self._update_replication_grp(context, group, volumes, []))
            model_update.update({'replication_status': 'enabled'})

        LOG.debug("Leave: create_group_from_src.")
        return model_update, volumes_model

    def delete_group(self, context, group, volumes):
        """Deletes a group.

        inspur InStorage will delete the volumes of the group.
        """
        if (not volume_utils.is_group_a_type(
                group, 'consistent_group_snapshot_enabled')
            and not volume_utils.is_group_a_type(
                group, 'consistent_group_replication_enabled')):
            raise NotImplementedError()

        if volume_utils.is_group_a_type(
                group, 'consistent_group_replication_enabled'):
            return self._delete_replication_grp(group, volumes)

        if volume_utils.is_group_a_type(
                group, 'consistent_group_snapshot_enabled'):
            return self.delete_consistencygroup(context, group, volumes)

    def _delete_replication_grp(self, group, volumes):
        LOG.debug("Deleting replication group.")
        model_update = {'status': 'deleted'}
        volumes_model_update = []
        group_name = self.generate_rccg_name(group)
        try:
            self._assistant.delete_rc_consistgrp(group_name)
        except exception.VolumeBackendAPIException as ex:
            LOG.error("Failed to delete rccg %(rccg)s. "
                      "Exception: %(exception)s.",
                      {'rccg': group_name, 'exception': ex})
            model_update = {'status': 'error_deleting'}

        for volume in volumes:
            try:
                vol_name = self.generate_volume_name(volume)
                self._aux_backend_assistant.delete_rc_volume(vol_name,
                                                            target_vol=True)

                self._local_backend_assistant.delete_rc_volume(vol_name)
                volumes_model_update.append(
                    {'id':volume.id, 'status': 'deleted'})
            except exception.VolumeBackendAPIException as ex:
                model_update['status'] = 'error_deleting'
                LOG.error(_('Failed to get delete volume %(volume)s '
                            'Exception: %(err)s.'),
                            {'volume': vol_name,
                            'err': six.text_type(ex)})
                volumes_model_update.append(
                    {'id':volume.id, 'status': 'error_deleting'})

        return model_update, volumes_model_update


    def delete_consistencygroup(self, context, group, volumes):
        LOG.debug("Deleting consistencygroup.")
        model_update = {'status': 'deleted'}
        volumes_model_update = []

        for volume in volumes:
            vol_name = self.generate_volume_name(volume)
            try:
                self._assistant.delete_vdisk(vol_name, True)
                volumes_model_update.append(
                    {'id': volume['id'], 'status': 'deleted'})
            except exception.VolumeBackendAPIException as err:
                model_update['status'] = 'error_deleting'
                LOG.error(_("Failed to delete the volume %(vol)s of group. "
                            "Exception: %(exception)s."),
                          {'vol': vol_name, 'exception': err})
                volumes_model_update.append(
                    {'id': volume['id'], 'status': 'error_deleting'})

        return model_update, volumes_model_update

    def update_group(self, ctxt, group, add_volumes=None,
                     remove_volumes=None):
        """Adds or removes volume(s) to/from an existing group."""
        if (not volume_utils.is_group_a_type(
                group, 'consistent_group_snapshot_enabled')
            and not volume_utils.is_group_a_type(
                group, 'consistent_group_replication_enabled')):
            raise NotImplementedError()

        if volume_utils.is_group_a_type(
                group, 'consistent_group_replication_enabled'):
            return self._update_replication_grp(
                   ctxt, group, add_volumes, remove_volumes)

        if volume_utils.is_group_a_type(
                group, 'consistent_group_snapshot_enabled'):
            return self.update_consistencygroup(
                   ctxt, group, add_volumes, remove_volumes)

    def _update_replication_grp(self, context, group,
                                add_volumes=None, remove_volumes=None):
        LOG.debug("Updating replication group.")
        model_update = {'status': 'available'}
        group_name = self.generate_rccg_name(group)
        try:
            rccg = self._assistant.get_remotecopy_consistgrp_attr(group_name)
        except Exception as exp:
            if add_volumes:
                LOG.exeception("Unable to retrieve replication group infomation."
                               "Failed with exception %(exp)s", exp)
        if not rccg and len(add_volumes) > 0:
            LOG.error("Failed to update group: %(grp)s does not exist in backend.",
                      {'grp':group.id})
            model_update['status'] = 'error'
            return model_update, None, None

        # add remote copy relationship to rccg
        added_vols = []
        for volume in add_volumes:
            try:
                vol_name = None
                if not self._active_backend_id:
                    vol_name = self.generate_volume_name(volume)
                else:
                    vol_name = (instorage_const.REPLICA_AUX_VOL_PREFIX
                                + self.generate_volume_name(volume))
                rcrel = self._assistant.get_relationship_info(vol_name)
                if not rcrel:
                    LOG.error("Failed to update group: remote copy relationship"
                            " of %(vol)s does not exist in backend",
                            {'vol':vol_name})
                    model_update['status'] = 'error'
                    break
                else:
                    if(rccg['copy_type'] != 'empty_group' and
                    any(k for k in ('copy_type', 'state', 'primary')
                        if rccg[k] != rcrel[k])):
                        LOG.error("Failed to update rccg %(rccg)s: remote copy "
                                "type of %(vol)s is %(vol_rc_type)s, the rccg "
                                "type is %(rccg_type)s. rcrel state is ",
                                "%(rcrel_state)s rccg state is %(rccg_state)s, "
                                "the rcrel primary is %(rcrel_primary)s, "
                                "the rccg primary is %(rccg_primary)s.",
                                {'rccg':group_name,
                                'vol':volume.id,
                                'vol_rc_type':rcrel['copy_type'],
                                'rccg_type':rccg['copy_type'],
                                'rcrel_state':rcrel['state'],
                                'rccg_state':rccg['state'],
                                'rcrel_primary':rcrel['primary'],
                                'rccg_primary':rccg['primary']})
                        #self._assistant.start_relationship(vol_name)
                        model_update['status'] = 'error'
                    else:
                        self._assistant.chrcrelationship(
                            'name', rcrel['name'], group_name)
                        if rccg['copy_type'] == 'empty_group':
                            rccg = self._assistant.get_remotecopy_consistgrp_attr(
                                group_name)
                        added_vols.append({'id': volume.id, 'group_id': group.id})
                        self._update_rccg_properties(context, volume, group)

            except exception.VolumeBackendAPIException as ex:
                model_update['status'] = 'error'
                LOG.error("Failed to add the remote copy of volume %(vol)s to"
                          " group.Exception: %(exeception)s.",
                          {'vol':vol_name, 'exeception': ex})
        # Remove remote copy relationship from rccg
        removed_vols = []
        for volume in remove_volumes:
            try:
                vol_name = None
                if not self._active_backend_id:
                    vol_name = self.generate_volume_name(volume)
                else:
                    vol_name = (instorage_const.REPLICA_AUX_VOL_PREFIX
                                + self.generate_volume_name(volume))
                rcrel = self._assistant.get_relationship_info(vol_name)
                if not rcrel:
                    LOG.error("Failed to update group: remote copy relationship"
                            " of %(vol)s does not exist in backend",
                            {'vol':vol_name})
                    model_update['status'] = 'error'
                    break
                else:
                    self._assistant.chrcrelationship('name', rcrel['name'])
                    removed_vols.append({'id': volume.id, 'group_id': None})
                    self._update_rccg_properties(context, volume)

            except exception.VolumeBackendAPIException as ex:
                model_update['status'] = 'error'
                LOG.error("Failed to remove the remote copy of volume %(vol)s "
                          "from group.Exception: %(exeception)s.",
                          {'vol':vol_name, 'exeception': ex})

        return model_update, added_vols, removed_vols

    def _update_rccg_properties(self, ctxt, volume, group=None):
        group_name = None
        if group:
            group_name = self.generate_rccg_name(group)
        else:
            group_name = ""
        if not volume.metadata:
            volume.metadata = dict()
        volume.metadata['Consistency Group Name'] = group_name
        volume.save()


    def update_consistencygroup(self, context, group,
                                add_volumes=None, remove_volumes=None):
        LOG.debug("Updating consistencygroup.")
        return None, None, None

    def create_group_snapshot(self, ctxt, group_snapshot, snapshots):
        """Creates a group_snapshot.

        :param context: the context of the caller.
        :param group_snapshot: the GroupSnapshot object to be created.
        :param snapshots: a list of Snapshot objects in the group_snapshot.
        :returns: model_update, snapshots_model_update
        """
        if (not volume_utils.is_group_a_type(
                group_snapshot, 'consistent_group_snapshot_enabled')
            and not volume_utils.is_group_a_type(
                group_snapshot, 'consistent_group_replication_enabled')):
            raise NotImplementedError()
        return self.create_cgsnapshot(ctxt, group_snapshot, snapshots)

    def create_cgsnapshot(self, ctxt, group_snapshot, snapshots):
        # Use cgsnapshot id as cg name
        group_name = 'group_snap-' + group_snapshot.id
        # Create new cg as cg_snapshot
        self._assistant.create_lc_consistgrp(group_name)

        timeout = self.configuration.instorage_mcs_localcopy_timeout
        model_update, snapshots_model = (
            self._assistant.run_group_snapshots(group_name,
                                                snapshots,
                                                self._state,
                                                self.configuration,
                                                timeout,
                                                self.generate_snapshot_name,
                                                self.generate_volume_name))

        return model_update, snapshots_model

    def delete_group_snapshot(self, context, group_snapshot, snapshots):
        """Deletes a group_snapshot.

        :param context: the context of the caller.
        :param group_snapshot: the GroupSnapshot object to be deleted.
        :param snapshots: a list of snapshot objects in the group_snapshot.
        :returns: model_update, snapshots_model_update
        """
        if (not volume_utils.is_group_a_type(
                group_snapshot, 'consistent_group_snapshot_enabled')
            and not volume_utils.is_group_a_type(
                group_snapshot, 'consistent_group_replication_enabled')):
            raise NotImplementedError()
        return self.delete_cgsnapshot(context, group_snapshot, snapshots)

    def delete_cgsnapshot(self, context, group_snapshot, snapshots):
        group_snapshot_id = group_snapshot['id']
        group_name = 'group_snap-' + group_snapshot_id
        model_update, snapshots_model = (
            self._assistant.delete_group_snapshots(group_name,
                                                   snapshots,
                                                   self.generate_snapshot_name))

        return model_update, snapshots_model

    def get_pool(self, volume):
        vol_name = self.generate_volume_name(volume)
        attr = self._assistant.get_vdisk_attributes(vol_name)

        if attr is None:
            msg = (_('get_pool: Failed to get attributes for volume '
                     '%s') % volume['id'])
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)

        return attr['mdisk_grp_name']

    def _check_aa_alive(self):
        online_nodes = self._assistant.get_node_info()
        available_io_groups = set(node['IO_group']
                                  for _, node in online_nodes.items())
        config_io_groups = []
        if self.active_active_iogrp_map:
            for primary, second in self.active_active_iogrp_map.items():
                config_io_groups.append(primary)
                config_io_groups.append(second)
        config_io_groups = set(config_io_groups)
        if not config_io_groups.issubset(available_io_groups):
            msg = _('_check_aa_alive failed, available_io_groups: %s, '
                    'config_io_groups: %s') % (available_io_groups, config_io_groups)
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)

    def _update_volume_stats(self):
        """Retrieve stats info from volume group."""

        LOG.debug("Updating volume stats.")
        if self.configuration.safe_get('instorage_mcs_enable_aa'):
            self._check_aa_alive()

        data = {}

        data['vendor_name'] = 'inspur'
        data['driver_version'] = self.VERSION
        data['storage_protocol'] = self.protocol
        data['pools'] = []

        backend_name = self.configuration.safe_get('volume_backend_name')
        data['volume_backend_name'] = (backend_name or
                                       self._state['system_name'])

        data['pools'] = [self._build_pool_stats(pool)
                         for pool in
                         self._get_backend_pools()]
        if self._replica_enabled:
            data['replication'] = self._replica_enabled
            data['replication_enabled'] = self._replica_enabled
            data['replication_targets'] = self._get_replication_targets()
            data['consistent_group_replication_enabled'] = True
        self._stats = data

    def _build_pool_stats(self, pool):
        """Build pool status"""
        qos_support = True
        pool_stats = {}
        try:
            pool_data = self._assistant.get_pool_attrs(pool)
            if pool_data:
                in_tier = pool_data['in_tier'] in ['on', 'auto']
                total_capacity_gb = float(pool_data['capacity']) / units.Gi
                free_capacity_gb = float(pool_data['free_capacity']) / units.Gi
                allocated_capacity_gb = (float(pool_data['used_capacity']) /
                                         units.Gi)
                provisioned_capacity_gb = float(
                    pool_data['virtual_capacity']) / units.Gi

                deduped_support = (self._state['deduped_enabled'] and
                                   'category' in pool_data and
                                   pool_data['category'] == 'ssa')
                online_extend_support = self.configuration.safe_get(
                    'online_extend_support')
                active_active_enabled = self.configuration.safe_get('instorage_mcs_enable_aa')
                if active_active_enabled:
                    online_extend_support = False
                over_sub_ratio = self.configuration.safe_get(
                    'max_over_subscription_ratio')
                location_info = ('InStorageMCSDriver:%(sys_id)s:%(pool)s' %
                                 {'sys_id': self._state['system_id'],
                                  'pool': pool_data['name']})
                pool_stats = {
                    'pool_name': pool_data['name'],
                    'total_capacity_gb': total_capacity_gb,
                    'free_capacity_gb': free_capacity_gb,
                    'allocated_capacity_gb': allocated_capacity_gb,
                    'provisioned_capacity_gb': provisioned_capacity_gb,
                    'compressed': self._state['compression_enabled'],
                    'deduped': deduped_support,
                    'reserved_percentage':
                        self.configuration.reserved_percentage,
                    'QoS_support': qos_support,
                    'consistencygroup_support': True,
                    'consistent_group_snapshot_enabled': True,
                    'location_info': location_info,
                    'intier_support': in_tier,
                    'multiattach': True,
                    'thin_provisioning_support': True,
                    'thick_provisioning_support': True,
                    'max_over_subscription_ratio': over_sub_ratio,
                    'online_extend_support': online_extend_support,
                    'ip_addr': self.active_ip,
                    'active_active_enabled':active_active_enabled
                }
            if self._replica_enabled:
                pool_stats.update({
                    'replication_enabled': self._replica_enabled,
                    'replication_type': self._supported_replica_types,
                    'replication_targets': self._get_replication_targets(),
                    'replication_count': len(self._get_replication_targets()),
                    'consistent_group_replication_enabled': True
                })

        except exception.VolumeBackendAPIException:
            msg = _('Failed getting details for pool %s.') % pool
            raise exception.VolumeBackendAPIException(data=msg)

        resp_info = self._assistant.get_cluster_capacity_info()
        total_mdisk_capacity_gb = math.floor(float(resp_info['total_mdisk_capacity']) / units.Gi)
        total_allocated_extent_capacity_gb = math.floor(float(
            resp_info['total_allocated_extent_capacity']) / units.Gi)
        total_free_capacity_gb = math.floor((float(total_mdisk_capacity_gb -
                                                   total_allocated_extent_capacity_gb)))

        if total_mdisk_capacity_gb and total_free_capacity_gb:
            pool_stats['inspur_total_capacity'] = total_mdisk_capacity_gb
            pool_stats['inspur_free_capacity'] = total_free_capacity_gb

        return pool_stats

    def _get_replication_targets(self):
        return [self._replica_target['backend_id']]

    def _manage_input_check(self, ref):
        """Verify the input of manage function."""
        # Check that the reference is valid
        if 'source-name' in ref:
            manage_source = ref['source-name']
            vdisk = self._assistant.get_vdisk_attributes(manage_source)
        elif 'source-id' in ref:
            manage_source = ref['source-id']
            vdisk = self._assistant.vdisk_by_uid(manage_source)
        else:
            reason = _('Reference must contain source-id or '
                       'source-name element.')
            raise exception.ManageExistingInvalidReference(existing_ref=ref,
                                                           reason=reason)

        if vdisk is None:
            reason = (_('No vdisk with the UID specified by ref %s.')
                      % manage_source)
            raise exception.ManageExistingInvalidReference(existing_ref=ref,
                                                           reason=reason)
        return vdisk

    # OpenStack Pike introduced replication group support
    def failover_host(self, context, volumes, secondary_id=None, groups=None):
        LOG.debug('enter: failover_host: secondary_id=%(id)s',
                  {'id': secondary_id})
        if not self._replica_enabled:
            msg = _("Replication is not properly enabled on backend.")
            LOG.error(msg)
            raise exception.UnableToFailOver(reason=msg)

        if instorage_const.FAILBACK_VALUE == secondary_id:
            # In this case the administrator would like to fail back.
            secondary_id, volumes_update, groups_update = self._replication_failback(
                context, volumes, groups)
        elif (secondary_id == self._replica_target['backend_id'] or
              secondary_id is None):
            # In this case the administrator would like to fail over.
            secondary_id, volumes_update, groups_update = self._replication_failover(
                context, volumes, groups)
        else:
            msg = (_("Invalid secondary id %s.") % secondary_id)
            LOG.error(msg)
            raise exception.InvalidReplicationTarget(reason=msg)

        LOG.debug('leave: failover_host: secondary_id=%(id)s',
                  {'id': secondary_id})
        if groups is None:
            return secondary_id, volumes_update
        else:
            return secondary_id, volumes_update, groups_update

    def enable_replication(self, context, group, volumes):
        """Enables replication for a group and volumes in the group."""
        model_update = {'replication_status': 'enabled'}
        volumes_update = []
        rccg_name = self.generate_rccg_name(group)
        rccg = self._assistant.get_remotecopy_consistgrp_attr(rccg_name)
        if rccg and rccg['relationship_count'] != '0':
            try:
                if rccg['primary'] == 'aux':
                    self._assistant.start_rccg(rccg_name, primary='aux')
                else:
                    self._assistant.start_rccg(rccg_name, primary='master')
            except exception.VolumeBackendAPIException as exp:
                LOG.error("Failed to enable group replication on %(rccg)s. "
                          "Exception: %(exception)s.",
                          {'rccg': rccg_name, 'exception': exp})
                model_update['replication_status'] = 'error'
        else:
            if rccg:
                LOG.error("Enable replication on empty group %(rccg)s is "
                          "forbidden.", {'rccg': rccg['name']})
            else:
                LOG.error("Failed to enable group replication: %(grp)s does "
                          "not exist in backend.", {'grp': group.id})
            model_update['replication_status'] = 'error'

        for volume in volumes:
            volumes_update.append(
                {'id': volume.id,
                 'replication_status': model_update['replication_status']})
        return model_update, volumes_update

    def disable_replication(self, context, group, volumes):
        """Disables replication for a group and volumes in the group."""
        model_update = {'replication_status': 'disabled'}
        volumes_update = []
        rccg_name = self.generate_rccg_name(group)
        rccg = self._assistant.get_remotecopy_consistgrp_attr(rccg_name)
        if rccg and rccg['relationship_count'] != '0':
            try:
                self._assistant.stop_rccg(rccg_name)
            except exception.VolumeBackendAPIException as exp:
                LOG.error("Failed to disable group replication on %(rccg)s. "
                          "Exception: %(exception)s.",
                          {'rccg': rccg_name, 'exception': exp})
                model_update['replication_status'] ='error'
        else:
            if rccg:
                LOG.error("Disable replication on empty group %(rccg)s is "
                          "forbidden.", {'rccg': rccg['name']})
            else:
                LOG.error("Failed to disable group replication: %(grp)s does "
                          "not exist in backend.", {'grp': group.id})
            model_update['replication_status'] = 'error'

        for volume in volumes:
            volumes_update.append(
                {'id': volume.id,
                 'replication_status': model_update['replication_status']})
        return model_update, volumes_update

    def failover_replication(self, context, group, volumes,
                             secondary_backend_id=None):
        """Fails over replication for a group and volumes in the group."""
        volumes_model_update = []
        model_update = {}
        if not self._replica_enabled:
            msg = _("Replication is not properly enabled on backend.")
            LOG.error(msg)
            raise exception.UnableToFailOver(reason=msg)

        if instorage_const.FAILBACK_VALUE == secondary_backend_id:
            # In this case the administrator would like to group fail back.
            self._sync_replica_groups(context, [group])
            self._wait_replica_groups_ready(context, [group])
            model_update = self._rep_grp_failback(context, group)
        elif (secondary_backend_id == self._replica_target['backend_id'] or
                secondary_backend_id is None):
            # In this case the administrator would like to group fail over.
            model_update = self._rep_grp_failover(context, group)
        else:
            msg = (_("Invalid secondary id %s.") % secondary_backend_id)
            LOG.error(msg)
            raise exception.InvalidReplicationTarget(reason=msg)

        for vol in volumes:
            volume_model_update = {'id': vol.id,
                                   'replication_status':
                                       model_update['replication_status']}
            volumes_model_update.append(volume_model_update)
        return model_update, volumes_model_update

    def _rep_grp_failback(self, ctxt, group, sync_grp=True):
        """Fail back all the volume in the replication group."""
        model_update = {
            'replication_status': 'enabled'}
        rccg_name = self.generate_rccg_name(group)

        try:
            self._aux_backend_assistant.stop_rc_consistgrp(
                rccg_name, access=True)
            self._aux_backend_assistant.start_rc_consistgrp(
                rccg_name, primary='master')
            return model_update
        except exception.VolumeBackendAPIException as exp:
            msg = (('Unable to fail back the group %(rccg)s, error: '
                     '%(error)s') % {"rccg": rccg_name, "error": exp})
            LOG.exception(msg)
            raise exception.UnableToFailOver(reason=msg)

    def _rep_grp_failover(self, ctxt, group):
        """Fail over all the volume in the replication group."""
        model_update = {
            'replication_status': 'failed-over'}
        rccg_name = self.generate_rccg_name(group)

        try:
            self._aux_backend_assistant.stop_rccg(rccg_name, access=True)
            self._assistant.start_rccg(rccg_name, primary='aux')
            return model_update
        except exception.VolumeBackendAPIException as exp:
            msg = (('Unable to fail over the group %(rccg)s to the aux '
                     'back-end, error: %(error)s') %
                   {"rccg": rccg_name, "error": exp})
            LOG.exception(msg)
            raise exception.UnableToFailOver(reason=msg)

    def _replication_failback(self, ctxt, volumes, groups):
        """Fail back all the volume on the secondary backend."""
        volumes_update = []
        groups_update = []
        if not self._active_backend_id:
            LOG.info(_("Host has been failed back. doesn't need "
                       "to fail back again"))
            return None, volumes_update, groups_update

        try:
            self._local_backend_assistant.get_system_info()
        except Exception:
            msg = (_("Unable to failback due to primary is not reachable."))
            LOG.error(msg)
            raise exception.UnableToFailOver(reason=msg)

        normal_volumes, rep_volumes = self._classify_volume(ctxt, volumes)

        # start synchronize from aux volume to master volume
        self._sync_with_aux(ctxt, rep_volumes)
        self._sync_replica_groups(ctxt, groups)
        self._wait_replica_ready(ctxt, rep_volumes)
        self._wait_replica_groups_ready(ctxt, groups)

        rep_volumes_update = self._failback_replica_volumes(ctxt,
                                                            rep_volumes)
        volumes_update.extend(rep_volumes_update)

        rep_vols_in_grp_update, groups_update = self._failback_replica_groups(
            ctxt, groups)
        volumes_update.extend(rep_vols_in_grp_update)

        normal_volumes_update = self._failback_normal_volumes(normal_volumes)
        volumes_update.extend(normal_volumes_update)

        self._assistant = self._local_backend_assistant
        self._active_backend_id = None

        # Update the instorage state
        self._update_instorage_state()
        self._update_volume_stats()
        return instorage_const.FAILBACK_VALUE, volumes_update, groups_update

    def _failback_replica_volumes(self, ctxt, rep_volumes):
        LOG.debug('enter: _failback_replica_volumes')
        volumes_update = []

        for volume in rep_volumes:
            vol_name = self.generate_volume_name(volume)
            rep_type = self._get_volume_replicated_type(ctxt, volume)
            replica_obj = self._get_replica_obj(rep_type)
            tgt_volume = instorage_const.REPLICA_AUX_VOL_PREFIX + vol_name
            rep_info = self._assistant.get_relationship_info(tgt_volume)
            if not rep_info:
                volumes_update.append(
                    {'volume_id': volume['id'],
                     'updates': {'replication_status': 'error',
                                 'status': 'error'}})
                LOG.error(_('_failback_replica_volumes:no rc-releationship '
                            'is established between master: %(master)s and '
                            'aux %(aux)s. Please re-establish the '
                            'relationship and synchronize the volumes on '
                            'backend storage.'),
                          {'master': vol_name, 'aux': tgt_volume})
                continue
            LOG.debug('_failover_replica_volumes: vol=%(vol)s, master_vol='
                      '%(master_vol)s, aux_vol=%(aux_vol)s, state=%(state)s'
                      'primary=%(primary)s',
                      {'vol': vol_name,
                       'master_vol': rep_info['master_vdisk_name'],
                       'aux_vol': rep_info['aux_vdisk_name'],
                       'state': rep_info['state'],
                       'primary': rep_info['primary']})
            try:
                replica_obj.replication_failback(volume)
                model_updates = {'replication_status':'enabled'}
                volumes_update.append(
                    {'volume_id': volume['id'],
                     'updates': model_updates})
            except exception.VolumeDriverException:
                LOG.error(_('Unable to fail back volume %(volume_id)s'),
                          {'volume_id': volume.id})
                volumes_update.append(
                    {'volume_id': volume['id'],
                     'updates': {'replication_status': 'error',
                                 'status': 'error'}})
        LOG.debug('leave: _failback_replica_volumes '
                  'volumes_update=%(volumes_update)s',
                  {'volumes_update': volumes_update})
        return volumes_update

    def _failback_normal_volumes(self, normal_volumes):
        volumes_update = []
        for vol in normal_volumes:
            pre_status = 'available'
            if ('replication_driver_data' in vol and
                    vol['replication_driver_data']):
                rep_data = json.loads(vol['replication_driver_data'])
                pre_status = rep_data['previous_status']
            volumes_update.append(
                {'volume_id': vol['id'],
                 'updates': {'status': pre_status,
                             'replication_driver_data': ''}})
        return volumes_update

    def _sync_with_aux(self, ctxt, volumes):
        LOG.debug('enter: _sync_with_aux ')
        try:
            rep_mgr = self._get_replica_mgr()
            rep_mgr.establish_target_partnership()
        except Exception as ex:
            LOG.warning(_('Fail to establish partnership in backend. '
                          'error=%(ex)s'), {'error': six.text_type(ex)})
        for volume in volumes:
            vol_name = self.generate_volume_name(volume)
            tgt_volume = instorage_const.REPLICA_AUX_VOL_PREFIX + vol_name
            rep_info = self._assistant.get_relationship_info(tgt_volume)
            if not rep_info:
                LOG.error(_('_sync_with_aux: no rc-releationship is '
                            'established between master: %(master)s and aux '
                            '%(aux)s. Please re-establish the relationship '
                            'and synchronize the volumes on backend '
                            'storage.'), {'master': vol_name,
                                          'aux': tgt_volume})
                continue
            LOG.debug('_sync_with_aux: volume: %(volume)s rep_info:master_vol='
                      '%(master_vol)s, aux_vol=%(aux_vol)s, state=%(state)s, '
                      'primary=%(primary)s',
                      {'volume': vol_name,
                       'master_vol': rep_info['master_vdisk_name'],
                       'aux_vol': rep_info['aux_vdisk_name'],
                       'state': rep_info['state'],
                       'primary': rep_info['primary']})
            try:
                if rep_info['state'] != instorage_const.REP_CONSIS_SYNC:
                    if rep_info['primary'] == 'master':
                        self._assistant.start_relationship(tgt_volume)
                    else:
                        self._assistant.start_relationship(tgt_volume,
                                                           primary='aux')
            except Exception as ex:
                LOG.warning(_('Fail to copy data from aux to master. master:'
                              ' %(master)s and aux %(aux)s. Please '
                              're-establish the relationship and synchronize'
                              ' the volumes on backend storage. error='
                              '%(ex)s'), {'master': vol_name,
                                          'aux': tgt_volume,
                                          'error': six.text_type(ex)})
        LOG.debug('leave: _sync_with_aux.')

    def _wait_replica_ready(self, ctxt, volumes):
        for volume in volumes:
            vol_name = self.generate_volume_name(volume)
            tgt_volume = instorage_const.REPLICA_AUX_VOL_PREFIX + vol_name
            try:
                self._wait_replica_vol_ready(ctxt, tgt_volume)
            except Exception as ex:
                LOG.error(_('_wait_replica_ready: wait for volume:%(volume)s'
                            ' remote copy synchronization failed due to '
                            'error:%(err)s.'), {'volume': tgt_volume,
                                                'err': six.text_type(ex)})

    def _wait_replica_vol_ready(self, ctxt, volume):
        LOG.debug('enter: _wait_replica_vol_ready: volume=%(volume)s',
                  {'volume': volume})

        def _replica_vol_ready():
            rep_info = self._assistant.get_relationship_info(volume)
            if not rep_info:
                msg = (_('_wait_replica_vol_ready: no rc-releationship'
                         'is established for volume:%(volume)s. Please '
                         're-establish the rc-relationship and '
                         'synchronize the volumes on backend storage.'),
                       {'volume': volume})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            LOG.debug('_replica_vol_ready:volume: %(volume)s rep_info: '
                      'master_vol=%(master_vol)s, aux_vol=%(aux_vol)s, '
                      'state=%(state)s, primary=%(primary)s',
                      {'volume': volume,
                       'master_vol': rep_info['master_vdisk_name'],
                       'aux_vol': rep_info['aux_vdisk_name'],
                       'state': rep_info['state'],
                       'primary': rep_info['primary']})
            if rep_info['state'] == instorage_const.REP_CONSIS_SYNC:
                return True
            if rep_info['state'] == instorage_const.REP_IDL_DISC:
                msg = (_('Wait synchronize failed. volume: %(volume)s'),
                       {'volume': volume})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            return False

        self._assistant.wait_for_a_condition(
            _replica_vol_ready, timeout=instorage_const.DEFAULT_RC_TIMEOUT,
            interval=instorage_const.DEFAULT_RC_INTERVAL,
            raise_exception=True)
        LOG.debug('leave: _wait_replica_vol_ready: volume=%(volume)s',
                  {'volume': volume})

    def _failback_replica_groups(self, ctxt, groups):
        volumes_update = []
        groups_update = []
        for grp in groups:
            try:
                grp_rep_status = self._rep_grp_failback(
                    ctxt, grp, sync_grp=False)['replication_status']
            except Exception as ex:
                LOG.error('Fail to failback group %(grp)s during host '
                          'failback due to error: %(error)s',
                          {'grp': grp.id, 'error': ex})
                grp_rep_status = 'error'

            # Update all the volumes' status in that group
            for vol in grp.volumes:
                vol_update = {'volume_id': vol.id,
                              'updates':
                                  {'replication_status': grp_rep_status,
                                   'status': (
                                       vol.status if grp_rep_status ==
                                       'enabled' else 'error')}}
                volumes_update.append(vol_update)
            grp_status = ('available' if grp_rep_status == 'enabled'
                           else 'error')
            grp_update = {'group_id': grp.id,
                          'updates': {'replication_status': grp_rep_status,
                                      'status': grp_status}}
            groups_update.append(grp_update)
        return volumes_update, groups_update

    def _sync_replica_groups(self, ctxt, groups):
        for grp in groups:
            rccg_name = self.generate_rccg_name(grp)
            self._sync_with_aux_grp(ctxt, rccg_name)

    def _sync_with_aux_grp(self, ctxt, rccg_name):
        try:
            rccg = self._assistant.get_remotecopy_consistgrp_attr(rccg_name)
            if rccg and rccg['relationship_count'] != '0':
                if (rccg['state'] not in
                        [instorage_const.REP_CONSIS_SYNC,
                        instorage_const.REP_CONSIS_COPYING]):
                    if rccg['primary'] == 'master':
                        self._assistant.start_rccg(rccg_name, primary='master')
                    else:
                        self._assistant.start_rccg(rccg_name, primary='aux')
            else:
                LOG.warning('group %(grp)s is not in sync.',
                            {'grp': rccg_name})
        except exception.VolumeBackendAPIException as ex:
            LOG.warning('Fail to copy data from aux group %(rccg)s to master '
                        'group. Please recheck the relationship and '
                        'synchronize the group on backend storage. error='
                        '%(error)s', {'rccg': rccg['name'], 'error': ex})

    def _wait_replica_groups_ready(self, ctxt, groups):
        for grp in groups:
            rccg_name = self.generate_rccg_name(grp)
            self._wait_replica_grp_ready(ctxt, rccg_name)

    def _wait_replica_grp_ready(self, ctxt, rccg_name):
        LOG.debug('_wait_replica_grp_ready: group=%(rccg)s',
                {'rccg': rccg_name})

        def _replica_grp_ready():
            rccg = self._assistant.get_remotecopy_consistgrp_attr(rccg_name)
            if not rccg:
                msg = (_('_replica_grp_ready: no group %(rccg)s exists on the '
                        'backend. Please re-create the rccg and synchronize '
                        'the volumes on backend storage.'),
                        {'rccg': rccg_name})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            if rccg['relationship_count'] == '0':
                return True
            LOG.debug('_replica_grp_ready: group: %(rccg)s: state=%(state)s, '
                    'primary=%(primary)s',
                    {'rccg': rccg['name'], 'state': rccg['state'],
                    'primary': rccg['primary']})
            if rccg['state'] in [instorage_const.REP_CONSIS_SYNC]:
                return True
            if rccg['state'] == instorage_const.REP_IDL_DISC:
                msg = (_('Wait synchronize failed. group: %(rccg)s') %
                    {'rccg': rccg_name})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            return False
        try:
            self._assistant.wait_for_a_condition(
                _replica_grp_ready,
                timeout=instorage_const.DEFAULT_RC_TIMEOUT,
                interval=instorage_const.DEFAULT_RC_INTERVAL,
                raise_exception=True)
        except Exception as ex:
            msg = ('_wait_replica_grp_ready: wait for group %(rccg)s '
                    'synchronization failed due to '
                    'error: %(err)s.', {'rccg': rccg_name,
                                        'err': ex})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

    def _replication_failover(self, ctxt, volumes, groups):
        volumes_update = []
        groups_update = []
        if self._active_backend_id:
            LOG.info(_("Host has been failed over to %s"),
                     self._active_backend_id)
            return self._active_backend_id, volumes_update, groups_update

        try:
            self._aux_backend_assistant.get_system_info()
        except Exception as ex:
            msg = (_("Unable to failover due to replication target is not "
                     "reachable. error=%(ex)s"), {'error': six.text_type(ex)})
            LOG.error(msg)
            raise exception.UnableToFailOver(reason=msg)

        normal_volumes, rep_volumes = self._classify_volume(ctxt, volumes)

        rep_volumes_update = self._failover_replica_volumes(ctxt, rep_volumes)
        volumes_update.extend(rep_volumes_update)

        rep_vols_in_grp_update, groups_update = self._failover_replica_groups(
            ctxt, groups)
        volumes_update.extend(rep_vols_in_grp_update)

        normal_volumes_update = self._failover_normal_volumes(normal_volumes)
        volumes_update.extend(normal_volumes_update)

        self._assistant = self._aux_backend_assistant
        self._active_backend_id = self._replica_target['backend_id']
        self._secondary_pools = [self._replica_target['pool_name']]

        # Update the instorage state
        self._update_instorage_state()
        self._update_volume_stats()
        return self._active_backend_id, volumes_update, groups_update

    def _failover_replica_volumes(self, ctxt, rep_volumes):
        LOG.debug('enter: _failover_replica_volumes')
        volumes_update = []

        for volume in rep_volumes:
            vol_name = self.generate_volume_name(volume)
            rep_type = self._get_volume_replicated_type(ctxt, volume)
            replica_obj = self._get_replica_obj(rep_type)
            # Try do the fail-over.
            try:
                rep_info = self._aux_backend_assistant.get_relationship_info(
                    instorage_const.REPLICA_AUX_VOL_PREFIX + vol_name)
                if not rep_info:
                    volumes_update.append(
                        {'volume_id': volume['id'],
                         'updates':
                             {'replication_status': 'failover-error',
                              'status': 'error'}})
                    LOG.error(_('_failover_replica_volumes: no rc-'
                                'releationship is established for master:'
                                '%(master)s. Please re-establish the rc-'
                                'relationship and synchronize the volumes on'
                                ' backend storage.'),
                              {'master': vol_name})
                    continue
                LOG.debug('_failover_replica_volumes: vol=%(vol)s, '
                          'master_vol=%(master_vol)s, aux_vol=%(aux_vol)s, '
                          'state=%(state)s, primary=%(primary)s',
                          {'vol': vol_name,
                           'master_vol': rep_info['master_vdisk_name'],
                           'aux_vol': rep_info['aux_vdisk_name'],
                           'state': rep_info['state'],
                           'primary': rep_info['primary']})
                if volume.status == 'in-use':
                    LOG.warning('_failover_replica_volumes: failover in-use '
                                'volume: %(volume)s is not recommended.',
                                {'volume': vol_name})
                replica_obj.failover_volume_host(ctxt, volume)
                model_updates = {
                    'replication_status': 'failed-over'}
                volumes_update.append(
                    {'volume_id': volume['id'],
                     'updates': model_updates})
            except exception.VolumeDriverException:
                LOG.error(_('Unable to failover to aux volume. Please make '
                            'sure that the aux volume is ready.'))
                volumes_update.append(
                    {'volume_id': volume['id'],
                     'updates': {'status': 'error',
                                 'replication_status': 'failed-over'}})
        LOG.debug('leave: _failover_replica_volumes '
                  'volumes_update=%(volumes_update)s',
                  {'volumes_update': volumes_update})
        return volumes_update

    def _failover_normal_volumes(self, normal_volumes):
        volumes_update = []
        for volume in normal_volumes:
            # If the volume is not of replicated type, we need to
            # force the status into error state so a user knows they
            # do not have access to the volume.
            rep_data = json.dumps({'previous_status': volume['status']})
            volumes_update.append(
                {'volume_id': volume['id'],
                 'updates': {'status': 'error',
                             'replication_driver_data': rep_data}})
        return volumes_update

    def _classify_volume(self, ctxt, volumes):
        normal_volumes = []
        replica_volumes = []

        for vol in volumes:
            volume_type = self._get_volume_replicated_type(ctxt, vol)
            grp = vol.group
            if grp and volume_utils.is_group_a_type(
                    grp, "consistent_group_replication_enabled"):
                continue
            elif volume_type and vol['status'] in ['available', 'in-use']:
                replica_volumes.append(vol)
            else:
                normal_volumes.append(vol)

        return normal_volumes, replica_volumes

    def _failover_replica_groups(self, ctxt, groups):
        volumes_update = []
        groups_update = []
        for grp in groups:
            try:
                grp_rep_status = self._rep_grp_failover(
                    ctxt, grp)['replication_status']
            except Exception as ex:
                LOG.error('Fail to failover group %(grp)s during host '
                          'failover due to error: %(error)s',
                          {'grp': grp.id, 'error': ex})
                grp_rep_status = 'error'

            # Update all the volumes' status in that group
            for vol in grp.volumes:
                vol_update = {'volume_id': vol.id,
                              'updates':
                                  {'replication_status': grp_rep_status,
                                   'status': (
                                       vol.status if grp_rep_status ==
                                       'failed-over'
                                       else 'error')}}
                volumes_update.append(vol_update)
            grp_status = ('available' if grp_rep_status == 'failed-over'
                          else 'error')
            grp_update = {'group_id': grp.id,
                          'updates': {'replication_status': grp_rep_status,
                                      'status': grp_status}}
            groups_update.append(grp_update)
        return volumes_update, groups_update

    def _get_replica_obj(self, rep_type):
        replica_manager = self.replica_manager[
            self._replica_target['backend_id']]
        return replica_manager.get_replica_obj(rep_type)

    def _get_replica_mgr(self):
        replica_manager = self.replica_manager[
            self._replica_target['backend_id']]
        return replica_manager

    def _get_target_vol(self, volume):
        tgt_vol = self.generate_volume_name(volume)
        if self._active_backend_id:
            ctxt = context.get_admin_context()
            rep_type = self._get_volume_replicated_type(ctxt, volume)
            if rep_type:
                tgt_vol = instorage_const.REPLICA_AUX_VOL_PREFIX + tgt_vol
        return tgt_vol

    def _validate_replication_enabled(self):
        if not self._replica_enabled:
            msg = _("Replication is not properly configured on backend.")
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

    def _get_specs_replicated_type(self, volume_type):
        replication_type = None
        extra_specs = volume_type.get("extra_specs", {})
        rep_val = extra_specs.get('replication_enabled')
        if rep_val == "<is> True":
            replication_type = extra_specs.get('replication_type',
                                               instorage_const.ASYNC)
            # The format for replication_type in extra spec is in
            # "<in> async". Otherwise, the code will
            # not reach here.
            if replication_type != instorage_const.ASYNC:
                # Pick up the replication type specified in the
                # extra spec from the format like "<in> async".
                replication_type = replication_type.split()[1]
            if replication_type not in instorage_const.VALID_REP_TYPES:
                msg = (_("Invalid replication type %s.") % replication_type)
                LOG.error(msg)
                raise exception.InvalidInput(reason=msg)
        return replication_type

    def _get_volume_replicated_type(self, ctxt, volume, vol_type_id=None):
        replication_type = None
        volume_type = None
        volume_type_id = volume.get("volume_type_id") if volume else vol_type_id
        if volume_type_id:
            volume_type = volume_types.get_volume_type(
                ctxt, volume_type_id)
        if volume_type:
            replication_type = self._get_specs_replicated_type(volume_type)
        return replication_type

    def _get_instorage_config(self):
        self._do_replication_setup()

        if self._active_backend_id and self._replica_target:
            self._assistant = self._aux_backend_assistant

        self._replica_enabled = (True if (self._assistant.
                                          replication_licensed() and
                                          self._replica_target) else False)
        if self._replica_enabled:
            self._supported_replica_types = instorage_const.VALID_REP_TYPES

    def _do_replication_setup(self):
        rep_devs = self.configuration.safe_get('replication_device')
        if not rep_devs:
            return

        if len(rep_devs) > 1:
            raise exception.InvalidInput(
                reason='Multiple replication devices are configured. '
                       'Now only one replication_device is supported.')

        required_flags = ['san_ip', 'backend_id', 'san_login',
                          'san_password', 'pool_name']
        for flag in required_flags:
            if flag not in rep_devs[0]:
                raise exception.InvalidInput(
                    reason=_('%s is not set.') % flag)

        rep_target = {}
        rep_target['san_ip'] = rep_devs[0].get('san_ip')
        rep_target['backend_id'] = rep_devs[0].get('backend_id')
        rep_target['san_login'] = rep_devs[0].get('san_login')
        rep_target['san_password'] = rep_devs[0].get('san_password')
        rep_target['pool_name'] = rep_devs[0].get('pool_name')
        if ((self.configuration.instorage_rsa_private_key is not None) and
                (self.configuration.instorage_rsa_private_key.strip() != '')):
            if not self.rsa_util:
                self.rsa_util = RsaUtil(self.configuration.instorage_rsa_private_key)
            rep_target['san_password'] = self.rsa_util.decrypt_by_private_key(
                rep_devs[0].get('san_password'))

        # Each replication target will have a corresponding replication.
        self._replication_initialize(rep_target)

    def _replication_initialize(self, target):
        rep_manager = instorage_rep.InStorageMCSReplicationManager(
            self, target, InStorageAssistant)

        if self._active_backend_id:
            if self._active_backend_id != target['backend_id']:
                msg = (_("Invalid secondary id %s.") % self._active_backend_id)
                LOG.error(msg)
                raise exception.InvalidInput(reason=msg)
        # Setup partnership only in non-failover state
        else:
            try:
                rep_manager.establish_target_partnership()
            except exception.VolumeDriverException:
                LOG.error(_('The replication src %(src)s has not '
                            'successfully established partnership with the '
                            'replica target %(tgt)s.'),
                          {'src': self.configuration.san_ip,
                           'tgt': target['backend_id']})

        self._aux_backend_assistant = rep_manager.get_target_assistant()
        self.replica_manager[target['backend_id']] = rep_manager
        self._replica_target = target


class InStorageAssistant(object):
    """inspur InStorage MCS CLI assistant for cinder driver."""
    # All the supported QoS key are saved in this dict. When a new
    # key is going to add, three values MUST be set:
    # 'default': to indicate the value, when the parameter is disabled.
    # 'param': to indicate the corresponding parameter in the command.
    # 'type': to indicate the type of this value.
    WAIT_TIME = 5
    mcs_qos_keys = {'IOThrottling': {'default': '0',
                                     'param': 'rate',
                                     'type': int},
                    'Bandwidth': {'default': '0',
                                  'param': 'rate',
                                  'type': int}}

    def __init__(self, run_cmd,
                 pool_map=None, iogrp_map=None,
                 loop_check_interval=10):

        self.executor = InStorageExector(run_cmd)
        self.loop_check_interval = loop_check_interval
        self.pool_map = pool_map
        self.iogrp_map = iogrp_map

    @staticmethod
    def handle_keyerror(cmd, out):
        msg = (_('Could not find key in output of command %(cmd)s: %(out)s.')
               % {'out': out, 'cmd': cmd})
        raise exception.VolumeBackendAPIException(data=msg)

    def compression_enabled(self):
        """Return whether or not compression is enabled for this system."""
        resp = self.executor.lslicense()
        keys = ['license_compression_enclosures',
                'license_compression_capacity']
        for key in keys:
            if resp.get(key, '0') != '0':
                return True
        try:
            resp = self.executor.lsguicapabilities()
            if resp.get('compression', '0') == 'yes':
                return True
        except exception.VolumeBackendAPIException:
            LOG.exception(_("Failed to fetch licensing scheme."))
        return False

    def deduped_enabled(self):
        """Return whether or not deduplication is enabled for this system."""
        try:
            resp = self.executor.lsguicapabilities()
            if resp.get('deduplication', '0') == 'yes':
                return True
        except exception.VolumeBackendAPIException:
            LOG.exception(_("Failed to fetch deduped licensing scheme."))
        return False

    def extend_cloning_support(self, capabilities_value):
        if int(capabilities_value) & 0b00000001 == 1:
            LOG.info("extend a cloning volume is supported")
            return True
        LOG.info("extend a cloning volume is not supported")
        return False

    def cascade_cloned_autodelete_support(self, capabilities_value):
        if int(capabilities_value) & 0b00000010 == 1:
            LOG.info("cascade cloned autodelete is supported")
            return True
        LOG.info("cascade cloned autodelete is not supported")
        return False

    def get_capabilities_value(self):
        capabilities_value = 0
        try:
            resp = self.executor.lsguicapabilities()
            capabilities_value = resp.get('license_type_localcopy', '0')
        except exception.VolumeBackendAPIException:
            LOG.exception(_("Failed to fetch license_type_localcopy."))
        return capabilities_value

    def replication_licensed(self):
        """Return whether or not replication is enabled for this system."""
        return True

    def get_system_info(self):
        """Return system's name, ID, and code level."""
        resp = self.executor.lssystem()
        level = resp['code_level']
        match_obj = re.search('([0-9].){3}[0-9]', level)
        if match_obj is None:
            msg = _('Failed to get code level (%s).') % level
            raise exception.VolumeBackendAPIException(data=msg)
        code_level = match_obj.group().split('.')
        return {'code_level': tuple([int(x) for x in code_level]),
                'system_name': resp['name'],
                'system_id': resp['id'],
                'topology': resp['topology'],
                'subnqn': resp.get('subnqn', None)}

    def get_cluster_capacity_info(self):
        """Return system's total and allocated_extent capacity."""
        resp = self.executor.lssystemcapcities()
        return {'total_mdisk_capacity': resp['total_mdisk_capacity'],
                'total_allocated_extent_capacity': resp['total_allocated_extent_capacity']}

    def get_node_info(self):
        """Return dictionary containing information on system's nodes."""
        nodes = {}
        resp = self.executor.lsnode()
        for node_data in resp:
            try:
                if node_data['status'] != 'online':
                    continue
                node = {}
                node['id'] = node_data['id']
                node['name'] = node_data['name']
                node['IO_group'] = node_data['IO_group_id']
                node['iscsi_name'] = node_data['iscsi_name']
                node['site_name'] = node_data['site_name']
                node['WWNN'] = node_data['WWNN']
                node['status'] = node_data['status']
                node['WWPN'] = []
                node['ipv4'] = []
                node['ipv6'] = []
                node['enabled_protocols'] = []
                nodes[node['id']] = node
            except KeyError:
                self.handle_keyerror('lsnode', node_data)
        return nodes

    def get_storage_nqn(self):
        return self.get_system_info()['subnqn']

    def get_pool_attrs(self, pool):
        """Return attributes for the specified pool."""
        return self.executor.lsmdiskgrp(pool)

    def is_ssa_pool(self, pool):
        attrs = self.get_pool_attrs(pool)
        try:
            return True if attrs['category'] == 'ssa' else False
        except (AttributeError,KeyError):
            return False

    def get_available_io_groups(self):
        """Return list of available IO groups."""
        iogrps = []
        resp = self.executor.lsiogrp()
        for iogrp in resp:
            try:
                if int(iogrp['node_count']) > 0:
                    iogrps.append(int(iogrp['id']))
            except KeyError:
                self.handle_keyerror('lsiogrp', iogrp)
            except ValueError:
                msg = (_('Expected integer for node_count, '
                         'mcsinq lsiogrp returned: %(node)s.') %
                       {'node': iogrp['node_count']})
                raise exception.VolumeBackendAPIException(data=msg)
        return iogrps

    def get_vdisk_count_by_io_group(self):
        res = {}
        resp = self.executor.lsiogrp()
        for iogrp in resp:
            try:
                if int(iogrp['node_count']) > 0:
                    res[int(iogrp['id'])] = int(iogrp['vdisk_count'])
            except KeyError:
                self.handle_keyerror('lsiogrp', iogrp)
            except ValueError:
                msg = (_('Expected integer for node_count, '
                         'mcsinq lsiogrp returned: %(node)s') %
                       {'node': iogrp['node_count']})
                raise exception.VolumeBackendAPIException(data=msg)
        return res

    def select_io_group(self, state, opts):
        selected_iog = 0
        iog_list = InStorageAssistant.get_valid_requested_io_groups(
            state, opts)
        if len(iog_list) == 0:
            raise exception.InvalidInput(
                reason=_('Given I/O group(s) %(iogrp)s not valid; available '
                         'I/O groups are %(avail)s.')
                % {'iogrp': opts['iogrp'],
                   'avail': state['available_iogrps']})
        iog_vdc = self.get_vdisk_count_by_io_group()
        LOG.debug("IO group current balance %s", iog_vdc)
        min_vdisk_count = iog_vdc[iog_list[0]]
        selected_iog = iog_list[0]
        for iog in iog_list:
            if iog_vdc[iog] < min_vdisk_count:
                min_vdisk_count = iog_vdc[iog]
                selected_iog = iog
        LOG.debug("Selected io_group is %d", selected_iog)
        return selected_iog

    def get_volume_io_group(self, vol_name):
        vdisk = self.executor.lsvdisk(vol_name)
        if vdisk:
            resp = self.executor.lsiogrp()
            for iogrp in resp:
                if iogrp['name'] == vdisk['IO_group_name']:
                    return int(iogrp['id'])
        return None

    def add_ip_addrs(self, storage_nodes):
        """Add IP addresses to system node information."""
        resp = self.executor.lsportip()
        # if NAS engine started. Storage will set up some ips for NAS cluster.
        # And this ips can not be used for iscsi or rdma. We will filter it
        filter_ips_v4 = ['172.16.2.33','172.16.2.43','172.16.2.53',
                         '172.16.2.32','172.16.2.42','172.16.2.52']
        for ip_data in resp:
            try:
                state = ip_data['state']
                if ip_data['node_id'] in storage_nodes and (
                        state == 'configured' or state == 'online'):
                    node = storage_nodes[ip_data['node_id']]
                    if len(ip_data['IP_address']) and ip_data['IP_address'] not in filter_ips_v4:
                        node['ipv4'].append(ip_data['IP_address'])
                    if len(ip_data['IP_address_6']):
                        node['ipv6'].append(ip_data['IP_address_6'])
            except KeyError:
                self.handle_keyerror('lsportip', ip_data)

    def add_fc_wwpns(self, storage_nodes):
        """Add FC WWPNs to system node information."""
        for key in storage_nodes:
            node = storage_nodes[key]
            wwpns = set(node['WWPN'])
            resp = self.executor.lsportfc(node_id=node['id'])
            for port_info in resp:
                if (port_info['type'] == 'fc' and
                        port_info['status'] == 'active'):
                    wwpns.add(port_info['WWPN'])
            node['WWPN'] = list(wwpns)
            LOG.info(_('WWPN on node %(node)s: %(wwpn)s.'),
                     {'node': node['id'], 'wwpn': node['WWPN']})

    def get_conn_fc_wwpns(self, host):
        wwpns = set()
        resp = self.executor.lsfabric(host=host)
        for wwpn in resp.select('local_wwpn'):
            if wwpn is not None:
                wwpns.add(wwpn)
        return list(wwpns)

    def add_chap_secret_to_host(self, host_name):
        """Generate and store a randomly-generated CHAP secret for the host."""
        chap_secret = volume_utils.generate_password()
        self.executor.add_chap_secret(chap_secret, host_name)
        return chap_secret

    def get_chap_secret_for_host(self, host_name):
        """Generate and store a randomly-generated CHAP secret for the host."""
        resp = self.executor.lsiscsiauth()
        host_found = False
        for host_data in resp:
            try:
                if host_data['name'] == host_name:
                    host_found = True
                    if host_data['iscsi_auth_method'] == 'chap':
                        return host_data['iscsi_chap_secret']
            except KeyError:
                self.handle_keyerror('lsiscsiauth', host_data)
        if not host_found:
            msg = _('Failed to find host %s.') % host_name
            raise exception.VolumeBackendAPIException(data=msg)
        return None

    def get_host_from_connector(self, connector, volume_name=None):
        """Return the InStorage host described by the connector."""
        LOG.debug('Enter: get_host_from_connector: %s.', connector)

        # If we have FC information, we have a faster lookup option
        host_name = None
        if 'wwpns' in connector:
            for wwpn in connector['wwpns']:
                resp = self.executor.lsfabric(wwpn=wwpn)
                for wwpn_info in resp:
                    try:
                        if (wwpn_info['remote_wwpn'] and
                                wwpn_info['name'] and
                                wwpn_info['remote_wwpn'].lower() ==
                                wwpn.lower()):
                            host_name = wwpn_info['name']
                            break
                    except KeyError:
                        self.handle_keyerror('lsfabric', wwpn_info)
                if host_name:
                    break
        if host_name:
            LOG.debug('Leave: get_host_from_connector: host %s.', host_name)
            return host_name

        def update_host_list(host, host_list):
            idx = host_list.index(host)
            del host_list[idx]
            host_list.insert(0, host)

        # That didn't work, so try exhaustive search
        hosts_info = self.executor.lshost()
        host_list = list(hosts_info.select('name'))
        # If we have a "real" connector, we might be able to find the
        # host entry with fewer queries if we move the host entries
        # that contain the connector's host property value to the front
        # of the list
        if 'host' in connector:
            # order host_list such that the host entries that
            # contain the connector's host name are at the
            # beginning of the list
            for host in host_list:
                if re.search(connector['host'], host):
                    update_host_list(host, host_list)
        # If we have a volume name we have a potential fast path
        # for finding the matching host for that volume.
        # Add the host_names that have mappings for our volume to the
        # head of the list of host names to search them first
        if volume_name:
            hosts_map_info = self.executor.lsvdiskhostmap(volume_name)
            hosts_map_info_list = list(hosts_map_info.select('host_name'))
            # remove the fast path host names from the end of the list
            # and move to the front so they are only searched for once.
            for host in hosts_map_info_list:
                update_host_list(host, host_list)
        found = False
        for name in host_list:
            try:
                resp = self.executor.lshost(host=name)
            except exception.VolumeBackendAPIException as ex:
                LOG.debug("Exception message: %s" % ex.msg)
                if 'CMMVC5754E' in ex.msg:
                    LOG.debug("CMMVC5754E found in CLI exception.")
                    # CMMVC5754E: The specified object does not exist
                    # The host has been deleted while walking the list.
                    # This is a result of a host change on the MCS that
                    # is out of band to this request.
                    continue
                # unexpected error so reraise it
                with excutils.save_and_reraise_exception():
                    pass
            if 'initiator' in connector:
                for iscsi in resp.select('iscsi_name'):
                    if iscsi == connector['initiator']:
                        host_name = name
                        found = True
                        break
            elif 'wwpns' in connector and len(connector['wwpns']):
                connector_wwpns = [str(x).lower() for x in connector['wwpns']]
                for wwpn in resp.select('WWPN'):
                    if wwpn and wwpn.lower() in connector_wwpns:
                        host_name = name
                        found = True
                        break
            elif 'nqn' in connector:
                for nqn in resp.select('nqn'):
                    if nqn == connector['nqn']:
                        host_name = name
                        found = True
                        break
            if found:
                break

        LOG.debug('Leave: get_host_from_connector: host %s.', host_name)
        return host_name

    def create_host(self, connector, site_name=''):
        """Create a new host on the storage system.

        We create a host name and associate it with the given connection
        information.  The host name will be a cleaned up version of the given
        host name (at most 55 characters), plus a random 8-character suffix to
        avoid collisions. The total length should be at most 63 characters.
        """
        LOG.debug('Enter: create_host: host %s.', connector['host'])

        # Before we start, make sure host name is a string and that we have
        # one port at least .
        host_name = connector['host']
        if not isinstance(host_name, six.string_types):
            msg = _('create_host: Host name is not unicode or string.')
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)

        ports = []
        if 'initiator' in connector:
            ports.append(['initiator', '%s' % connector['initiator']])
            protocol = 'scsi'
        if 'wwpns' in connector:
            for wwpn in connector['wwpns']:
                ports.append(['wwpn', '%s' % wwpn])
            protocol = 'scsi'
        if 'nqn' in connector:
            ports.append(['nqn', '%s' % connector['nqn']])
            protocol = 'nvme'
        if not ports:
            msg = _('create_host: No initiators or wwpns supplied.')
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)

        # Build a host name for the InStorage host - first clean up the name
        if isinstance(host_name, six.text_type):
            host_name = unicodedata.normalize('NFKD', host_name).encode(
                'ascii', 'replace').decode('ascii')

        for num in range(0, 128):
            char = str(chr(num))
            if not char.isalnum() and char not in [' ', '.', '-', '_']:
                host_name = host_name.replace(char, '-')

        # InStorage doesn't expect hostname that doesn't starts with letter or
        # _.
        if not re.match('^[A-Za-z]', host_name):
            host_name = '_' + host_name

        # Add a random 8-character suffix to avoid collisions
        rand_id = str(random.randint(0, 99999999)).zfill(8)
        host_name = '%s-%s' % (host_name[:55], rand_id)

        # Create a host with one port
        port = ports.pop(0)
        self.executor.mkhost(host_name, port[0], port[1], site_name, protocol)

        # Add any additional ports to the host
        for port in ports:
            self.executor.addhostport(host_name, port[0], port[1])

        LOG.debug('Leave: create_host: host %(host)s - %(host_name)s.',
                  {'host': connector['host'], 'host_name': host_name})
        return host_name

    def delete_host(self, host_name):
        self.executor.rmhost(host_name)

    def check_host_mapped_vols(self, host_name):
        return self.executor.lshostvdiskmap(host_name)

    def map_vol_to_host(self, volume_name, host_name, multihostmap, protocol='scsi'):
        """Create a mapping between a volume to a host."""

        LOG.debug('Enter: map_vol_to_host: volume %(volume_name)s to '
                  'host %(host_name)s.protocol: %(protocol)s',
                  {'volume_name': volume_name,
                   'host_name': host_name,
                   'protocol' : protocol}
                 )
        if protocol == 'nvme':
             # Check if this volume is already mapped to this host
            is_mapped = self.executor.check_volume_mapped(volume_name, host_name)
            if not is_mapped:
                self.executor.mkvdiskhostmap(host_name, volume_name, None,
                                                    multihostmap, protocol)
            LOG.debug('Leave: map_vol_to_host: volume '
                    '%(volume_name)s, host %(host_name)s, protocol: %(protocol)s.',
                    {'volume_name': volume_name,
                    'host_name': host_name,
                    'protocol': protocol})
            return True
        else:
            # Check if this volume is already mapped to this host
            iogrp_lun_id_map = self.executor.get_vdiskhostmapid(
                volume_name, host_name)
            if len(iogrp_lun_id_map) == 0:
                iogrp_lun_id_map = self.executor.mkvdiskhostmap(
                    host_name, volume_name, None, multihostmap)

            LOG.debug('Leave: map_vol_to_host: iogrp_lun_id_map %(iogrp_lun_id_map)s, volume '
                    '%(volume_name)s, host %(host_name)s, protocol: %(protocol)s',
                    {'iogrp_lun_id_map': iogrp_lun_id_map,
                    'volume_name': volume_name,
                    'host_name': host_name,
                    'protocol': protocol})
            return iogrp_lun_id_map

    def unmap_vol_from_host(self, volume_name, host_name):
        """Unmap the volume and delete the host if it has no more mappings."""

        LOG.debug('Enter: unmap_vol_from_host: volume %(volume_name)s from '
                  'host %(host_name)s.',
                  {'volume_name': volume_name, 'host_name': host_name})

        # Check if the mapping exists
        resp = self.executor.lsvdiskhostmap(volume_name)
        if not resp:
            LOG.warning(_('unmap_vol_from_host: No mapping of volume '
                          '%(vol_name)s to any host found.'),
                        {'vol_name': volume_name})
            return host_name
        if host_name is None:
            if len(resp) > 1:
                LOG.warning(_('unmap_vol_from_host: Multiple mappings of '
                              'volume %(vol_name)s found, no host '
                              'specified.'), {'vol_name': volume_name})
                return
            else:
                host_name = resp[0]['host_name']
        else:
            found = False
            for item in resp.select('host_name'):
                if item == host_name:
                    found = True
            if not found:
                LOG.warning(_('unmap_vol_from_host: No mapping of volume '
                              '%(vol_name)s to host %(host)s found.'),
                            {'vol_name': volume_name, 'host': host_name})
                return host_name
        # We now know that the mapping exists
        self.executor.rmvdiskhostmap(host_name, volume_name)

        LOG.debug('Leave: unmap_vol_from_host: volume %(volume_name)s from '
                  'host %(host_name)s.',
                  {'volume_name': volume_name, 'host_name': host_name})
        return host_name

    @staticmethod
    def build_default_opts(config):
        # Ignore capitalization
        opt = {
            'rsize': config.instorage_mcs_vol_rsize,
            'warning': config.instorage_mcs_vol_warning,
            'autoexpand': config.instorage_mcs_vol_autoexpand,
            'grainsize': config.instorage_mcs_vol_grainsize,
            'compressed': config.instorage_mcs_vol_compression,
            'deduped': config.instorage_mcs_vol_deduplicate,
            'intier': config.instorage_mcs_vol_intier,
            'iogrp': config.instorage_mcs_vol_iogrp,
            'access_iogrp':config.instorage_mcs_vol_access_iogrp,
            'syncrate': config.instorage_mcs_syncrate,
            'localcopyrate': config.instorage_mcs_localcopy_rate,
            'cleanrate':config.instorage_mcs_localcopy_clean_rate,
            'aa': config.instorage_mcs_enable_aa,
            'qos': None,
            'replication': False,
            'node': None
        }
        return opt

    @staticmethod
    def check_vdisk_opts(state, opts):
        # Check that grainsize is 32/64/128/256
        if opts['grainsize'] not in [32, 64, 128, 256]:
            raise exception.InvalidInput(
                reason=_('Illegal value specified for '
                         'instorage_mcs_vol_grainsize: set to either '
                         '32, 64, 128, or 256.'))

        # Check that compression is supported
        if opts['compressed'] and not state['compression_enabled']:
            raise exception.InvalidInput(
                reason=_('System does not support compression.'))

        # Check that deduped is supported
        if opts['deduped'] and not state['deduped_enabled']:
            raise exception.InvalidInput(
                reason=_('System does not support deduped.'))

        if opts['compressed'] and (opts['rsize'] == -1 or opts['rsize'] == 100):
            opts['rsize'] = 2

        if opts['deduped'] and (opts['rsize'] == -1 or opts['rsize'] == 100):
            opts['rsize'] = 2

        iogs = InStorageAssistant.get_valid_requested_io_groups(state, opts)

        if len(iogs) == 0:
            raise exception.InvalidInput(
                reason=_('Given I/O group(s) %(iogrp)s not valid; available '
                         'I/O groups are %(avail)s.')
                % {'iogrp': opts['iogrp'],
                   'avail': state['available_iogrps']})

    @staticmethod
    def get_valid_requested_io_groups(state, opts):
        given_iogs = str(opts['iogrp'])
        iog_list = given_iogs.split(',')
        # convert to int
        iog_list = list(map(int, iog_list))
        LOG.debug("Requested iogroups %s", iog_list)
        LOG.debug("Available iogroups %s", state['available_iogrps'])
        filtiog = set(iog_list).intersection(state['available_iogrps'])
        iog_list = list(filtiog)
        LOG.debug("Filtered (valid) requested iogroups %s", iog_list)
        return iog_list

    def _get_opts_from_specs(self, opts, specs):
        qos = {}
        for k, value in specs.items():
            # Get the scope, if using scope format
            key_split = k.split(':')
            if len(key_split) == 1:
                scope = None
                key = key_split[0]
            else:
                scope = key_split[0]
                key = key_split[1]

            # We generally do not look at capabilities in the driver, but
            # replication is a special case where the user asks for
            # a volume to be replicated, and we want both the scheduler and
            # the driver to act on the value.
            if not scope or scope == 'capabilities':
                if  key == 'replication' or key == 'compressed' or key == 'deduped':
                    # scope = None
                    # key = 'replication'
                    words = value.split()
                    if not (words and len(words) == 2 and words[0] == '<is>'):
                        LOG.error(_('%(key)s must be specified as '
                                    '\'<is> True\' or \'<is> False\'.') %
                                     {'key': key})
                    del words[0]
                    value = words[0]

            # Add the QoS.
            if scope and scope == 'qos':
                if key in self.mcs_qos_keys:
                    try:
                        type_fn = self.mcs_qos_keys[key]['type']
                        value = type_fn(value)
                        qos[key] = value
                    except ValueError:
                        continue

            # Any keys that the driver should look at should have the
            # 'drivers' scope.
            if key in opts:
                this_type = type(opts[key]).__name__
                if this_type == 'int':
                    value = int(value)
                elif this_type == 'bool':
                    value = strutils.bool_from_string(value)
                opts[key] = value
        if len(qos) != 0:
            opts['qos'] = qos
        return opts

    def _get_qos_from_volume_metadata(self, volume_metadata):
        """Return the QoS information from the volume metadata."""
        qos = {}
        for i in volume_metadata:
            k = i.get('key', None)
            value = i.get('value', None)
            key_split = k.split(':')
            if len(key_split) == 1:
                scope = None
                key = key_split[0]
            else:
                scope = key_split[0]
                key = key_split[1]
            # Add the QoS.
            if scope and scope == 'qos':
                if key in self.mcs_qos_keys:
                    try:
                        type_fn = self.mcs_qos_keys[key]['type']
                        value = type_fn(value)
                        qos[key] = value
                    except ValueError:
                        continue
        return qos

    def wait_for_a_condition(self, testmethod, timeout=None,
                              interval=INTERVAL_1_SEC,
                              raise_exception=False):
        start_time = time.time()
        if timeout is None:
            timeout = DEFAULT_TIMEOUT

        def _inner():
            try:
                test_value = testmethod()
            except Exception as ex:
                if raise_exception:
                    LOG.exception(_("wait_for_a_condition: %s"
                                    " execution failed."),
                                  testmethod.__name__)
                    raise exception.VolumeBackendAPIException(data=ex)
                else:
                    test_value = False
                    LOG.debug('Assistant.'
                              '_wait_for_condition: %(method_name)s '
                              'execution failed for %(exception)s.',
                              {'method_name': testmethod.__name__,
                               'exception': ex.message})
            if test_value:
                raise loopingcall.LoopingCallDone()

            if int(time.time()) - start_time > timeout:
                msg = (
                    _('CommandLineAssistant._wait_for_condition: '
                      '%s timeout.') % testmethod.__name__)
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

        timer = loopingcall.FixedIntervalLoopingCall(_inner)
        timer.start(interval=interval).wait()

    def get_vdisk_params(self, config, state, type_id,
                         volume_type=None, volume_metadata=None):
        """Return the parameters for creating the vdisk.

        Get volume type and defaults from config options
        and take them into account.
        """
        opts = self.build_default_opts(config)
        ctxt = context.get_admin_context()
        if volume_type is None and type_id is not None:
            volume_type = volume_types.get_volume_type(ctxt, type_id)
        if volume_type:
            qos_specs_id = volume_type.get('qos_specs_id')
            specs = dict(volume_type).get('extra_specs')

            # NOTE: We prefer the qos_specs association
            # and over-ride any existing
            # extra-specs settings if present
            if qos_specs_id is not None:
                kvs = qos_specs.get_qos_specs(ctxt, qos_specs_id)['specs']
                # Merge the qos_specs into extra_specs and qos_specs has higher
                # priority than extra_specs if they have different values for
                # the same key.
                specs.update(kvs)
            opts = self._get_opts_from_specs(opts, specs)
        if (opts['qos'] is None and config.instorage_mcs_allow_tenant_qos and
                volume_metadata):
            qos = self._get_qos_from_volume_metadata(volume_metadata)
            if len(qos) != 0:
                opts['qos'] = qos

        self.check_vdisk_opts(state, opts)
        return opts

    @staticmethod
    def _get_vdisk_create_params(opts, is_ssa_pool=False):
        params = []
        params.extend(['-syncrate', str(opts['syncrate'])])
        if is_ssa_pool or (opts['rsize'] != -1 and opts['rsize'] != 100):
            if opts['rsize'] == -1 or opts['rsize'] == 100:
                opts['rsize'] = 2
            params.extend(['-rsize', '%s%%' % str(opts['rsize'])])
            if opts['autoexpand']:
                params.append('-autoexpand')
            params.extend(['-warning', '%s%%' % str(opts['warning'])])

        if is_ssa_pool:
            if opts['deduped']:
                params.append('-deduped')
            if opts['compressed']:
                params.append('-compressed')
        elif opts['rsize'] != -1 and opts['rsize'] != 100:
            if opts['compressed']:
                params.append('-compressed')
            else:
                params.extend(['-grainsize', str(opts['grainsize'])])

        # intier option is ignore when pool is ssa
        if not is_ssa_pool:
            intier = 'on' if opts['intier'] else 'off'
            params.extend(['-intier', intier])

        return params

    @staticmethod
    def _get_volume_create_params(opts, is_ssa_pool=False):
        params = []
        if is_ssa_pool or (opts['rsize'] != -1 and opts['rsize'] != 100):
            if opts['rsize'] in [-1, 100]:
                #  if the configured pool is ssa and rsize is not configured
                #  we will set the rsize to 2
                opts['rsize'] = 2
            params.extend(['-buffersize', '%s%%' % str(opts['rsize']),
                           '-warning', '%s%%' % str(opts['warning'])])
            if not opts['autoexpand']:
                params.append('-noautoexpand')

        if is_ssa_pool:
            params.append('-thin')
            if opts['deduped']:
                params.append('-deduped')
            if opts['compressed']:
                params.append('-compressed')
        elif opts['rsize'] != -1 and opts['rsize'] != 100:
            if opts['compressed']:
                params.append('-compressed')
            else:
                params.append('-thin')
                params.extend(['-grainsize', str(opts['grainsize'])])

        # thick volume has no other params
        return params

    def check_ssa_pool_params(self, opts):
        """Check the configured parameters if vol in ssa pool."""
        if opts['rsize'] == -1 or opts['rsize'] == 100:
            opts['rsize'] = 2

    def check_normal_pool_params(self,opts):
        """Check the configured parameters if vol in normal pool."""
        if opts['deduped']:
            msg = (_('You cannot create deduped volume in normal pool'))
            raise exception.VolumeDriverException(message=msg)
        if opts['compressed'] and (opts['rsize'] == -1 or opts['rsize'] == 100):
            opts['rsize'] = 2

    def create_vdisk(self, name, size, units, pool, opts, nofmtdisk=False):
        LOG.debug('we will create a general volume %s' % name)

        name = '"%s"' % name

        is_ssa_pool = self.is_ssa_pool(pool)
        if is_ssa_pool:
            self.check_ssa_pool_params(opts)
        else:
            self.check_normal_pool_params(opts)
        params = self._get_vdisk_create_params(opts, is_ssa_pool)

        self.executor.mkvdisk(name, size, units, pool, opts, params, nofmtdisk)

    def create_volume(self, name, size, units, pool, opts):
        LOG.debug('we will create an active-active volume %s' % name)

        name = '"%s"' % name
        is_ssa_pool = self.is_ssa_pool(pool)
        if is_ssa_pool:
            self.check_ssa_pool_params(opts)
        else:
            self.check_normal_pool_params(opts)
        params = self._get_volume_create_params(opts, is_ssa_pool)

        poollist = '%s:%s' % (pool, self.pool_map[pool])
        iogrp = '%s' % opts['iogrp']
        iogrplist = '%s:%s' % (iogrp, self.iogrp_map[iogrp])
        self.executor.mkvolume(name, size, units, poollist, iogrplist, params)

    def transfer_vdisk_to_volume(self, name, pool, opts):
        LOG.debug('we will transfter vdisk %s to active-active volume.' % name)

        name = '"%s"' % name

        params = self._get_volume_create_params(opts, self.is_ssa_pool(pool))
        dst_pool = self.pool_map[pool]
        dst_iogrp = self.iogrp_map[opts['iogrp']]
        self.executor.addvolumecopy(name, dst_pool, dst_iogrp, params)

    def transfer_volume_to_vdisk(self, name, pool):
        LOG.debug('we will transfer volume %s to general vdisk by delete a copy.' % name)

        name = '"%s"' % name
        self.executor.rmvolumecopy(name, self.pool_map[pool])

    # maybe change a name is better, in this function itself will know
    # whether it is a aa volume or a general vdisk
    def delete_vdisk(self, vdisk, force):
        """Ensures that vdisk is not part of FC mapping and deletes it."""
        LOG.debug('Enter: delete_vdisk: vdisk %s.', vdisk)
        if not self.is_vdisk_defined(vdisk):
            LOG.info(_('Tried to delete non-existent vdisk %s.'), vdisk)
            return

        if self.is_aa_vdisk(vdisk):
            LOG.debug('Remove active-active volume')
            self.executor.rmvolume(vdisk, force=force)
        else:
            LOG.debug('Remove general vdisk')
            # however we always need to promise it has no local copy
            self.ensure_vdisk_no_lc_mappings(vdisk, allow_snaps=True, allow_lctgt=True)

            self.executor.rmvdisk(vdisk, force=force)
        LOG.debug('Leave: delete_vdisk: vdisk %s.', vdisk)

    def is_aa_vdisk(self, vdisk_name):
        attrs = self.executor.lsvdisks_from_filter('volume_name', vdisk_name)
        return True if len(attrs) == 4 else False

    def is_vdisk_defined(self, vdisk_name):
        """Check if vdisk is defined."""
        attrs = self.get_vdisk_attributes(vdisk_name)
        return attrs is not None

    def get_vdisk_attributes(self, vdisk):
        attrs = self.executor.lsvdisk(vdisk)
        return attrs
    
    def get_volumes(self, pool_name):
        """get all volumes in a specify pool"""
        return self.executor.lsvdisks_from_filter('mdisk_grp_name', pool_name)

    def get_volume_pool_iogrp_map(self, volume_name):
        attrs = self.executor.lsvdisks_from_filter('volume_name', volume_name)
        master = None
        auxiliary = None
        for attr in attrs:
            subattr = {
                'mdisk_grp_name': attr['mdisk_grp_name'],
                'IO_group_id': attr['IO_group_id']
            }
            if attr['function'] == 'master':
                master = subattr
            elif attr['function'] == 'aux':
                auxiliary = subattr

        return master, auxiliary

    def get_volume_iogrps(self, volume_name):
        attrs = self.executor.lsvdisks_from_filter('volume_name', volume_name)
        iogrps = set()
        for attr in attrs:
            iogrps.add(attr['IO_group_id'])
        return iogrps

    def get_volume_prefer_site_name(self, volume_name):
        site_name = ''
        prefer_iogrp_name = None

        attrs = self.executor.lsvdisks_from_filter('volume_name', volume_name)
        topology = self.get_system_info()['topology']
        for attr in attrs:
            if attr['function'] == 'master':
                prefer_iogrp_name = attr['IO_group_name']
                break
            if topology == 'inmetro':
                prefer_iogrp_name = attr['IO_group_name']
                break

        if prefer_iogrp_name is None:
            return site_name

        iogrps = self.executor.lsiogrp()
        for iogrp in iogrps:
            if iogrp['name'] == prefer_iogrp_name:
                site_name = iogrp['site_name']
                break
        return site_name

    def get_volumes_used_size(self, volume_name=None, pool=None):
        result = {}
        if volume_name:
            try:
                resp = self.executor.lsvdiskcopy(volume_name, None, None)[0]
                volume_id = volume_name[7:]
                result.update({volume_id: int(resp['used_capacity'])})
            except Exception as ex:
                if 'CMMVC5804E' in ex.msg:
                    LOG.warning('could not find the  %(vdisk)s.',
                            {'vdisk': volume_name})
                    return None
        elif pool:
            try:
                resp = self.executor.lsvdiskcopy(None, None, pool)
                for vdisk_name, used_capacity in resp.select(
                        'vdisk_name', 'used_capacity'):
                    if not vdisk_name.startswith('volume-'):
                        continue
                    volume_id = vdisk_name[7:]
                    result.update({volume_id: int(used_capacity)})
            except Exception:
                return None
        return result


    def find_vdisk_copy_id(self, vdisk, pool):
        resp = self.executor.lsvdiskcopy(vdisk)
        for copy_id, mdisk_grp in resp.select('copy_id', 'mdisk_grp_name'):
            if mdisk_grp == pool:
                return copy_id
        msg = _('Failed to find a vdisk copy in the expected pool.')
        LOG.error(msg)
        raise exception.VolumeDriverException(message=msg)

    def get_vdisk_copy_attrs(self, vdisk, copy_id):
        try:
            return self.executor.lsvdiskcopy(vdisk, copy_id=copy_id)[0]
        except Exception as ex:
            if 'CMMVC5804E' in ex.msg:
                LOG.warning('could not find the copy%(copyid)s of %(vdisk)s.',
                         {'copyid': copy_id, 'vdisk': vdisk})
                return None

    def get_vdisk_copies(self, vdisk):
        copies = {'primary': None,
                  'secondary': None}

        resp = self.executor.lsvdiskcopy(vdisk)
        for copy_id, status, sync, primary, mdisk_grp in (
            resp.select('copy_id', 'status', 'sync',
                        'primary', 'mdisk_grp_name')):
            copy = {'copy_id': copy_id,
                    'status': status,
                    'sync': sync,
                    'primary': primary,
                    'mdisk_grp_name': mdisk_grp,
                    'sync_progress': None}
            if copy['sync'] != 'yes':
                progress_info = self.executor.lsvdisksyncprogress(vdisk, copy_id)
                copy['sync_progress'] = progress_info['progress']
            if copy['primary'] == 'yes':
                copies['primary'] = copy
            else:
                copies['secondary'] = copy
        return copies

    def create_copy(self, src, tgt, src_id, config, opts,
                    full_copy, pool=None, nofmtdisk=False):
        """Create a new snapshot using LocalCopy."""
        LOG.debug('Enter: create_copy: snapshot %(src)s to %(tgt)s.',
                  {'tgt': tgt, 'src': src})

        src_attrs = self.get_vdisk_attributes(src)
        if src_attrs is None:
            msg = (_('create_copy: Source vdisk %(src)s (%(src_id)s) '
                     'does not exist.') % {'src': src, 'src_id': src_id})
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)

        src_size = src_attrs['capacity']
        src_node = src_attrs['preferred_node_id']
        if src_node and int(src_node) != 0:
            opts['node'] = src_node
        # In case we need to use a specific pool
        if not pool:
            pool = src_attrs['mdisk_grp_name']

        opts['iogrp'] = src_attrs['IO_group_id']
        self.create_vdisk(tgt, src_size, 'b', pool, opts, nofmtdisk)
        timeout = config.instorage_mcs_localcopy_timeout
        try:
            self.run_localcopy(src, tgt, timeout,
                               opts['localcopyrate'],
                               opts['cleanrate'],
                               full_copy=full_copy)
        except Exception:
            with excutils.save_and_reraise_exception():
                self.delete_vdisk(tgt, True)

        LOG.debug('Leave: _create_copy: snapshot %(tgt)s from '
                  'vdisk %(src)s.',
                  {'tgt': tgt, 'src': src})

    def extend_vdisk(self, vdisk, amount):
        self.executor.expandvdisksize(vdisk, amount)

    def add_vdisk_copy(self, vdisk, dest_pool, volume_type, state, config):
        """Add a vdisk copy in the given pool."""
        resp = self.executor.lsvdiskcopy(vdisk)
        if len(resp) > 1:
            msg = (_('add_vdisk_copy failed: A copy of volume %s exists. '
                     'Adding another copy would exceed the limit of '
                     '2 copies.') % vdisk)
            raise exception.VolumeDriverException(message=msg)
        orig_copy_id = resp[0].get("copy_id", None)

        if orig_copy_id is None:
            msg = (_('add_vdisk_copy started without a vdisk copy in the '
                     'expected pool.'))
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)

        if volume_type is None:
            opts = self.get_vdisk_params(config, state, None)
        else:
            opts = self.get_vdisk_params(config, state, volume_type['id'],
                                         volume_type=volume_type)
        params = self._get_vdisk_create_params(opts, self.is_ssa_pool(dest_pool))
        new_copy_id = self.executor.addvdiskcopy(vdisk, dest_pool, params)
        return (orig_copy_id, new_copy_id)

    def check_vdisk_copy_synced(self, vdisk, copy_id):
        sync = self.executor.lsvdiskcopy(vdisk, copy_id=copy_id)[0]['sync']
        if sync == 'yes':
            return True
        return False

    def rm_vdisk_copy(self, vdisk, copy_id):
        self.executor.rmvdiskcopy(vdisk, copy_id)

    def _prepare_lc_map(self, lc_map_id, timeout, restore=False):
        self.executor.prestartlcmap(lc_map_id, restore)
        mapping_ready = False
        max_retries = (timeout // self.WAIT_TIME) + 1
        for _ in range(1, max_retries):
            mapping_attrs = self.get_localcopy_mapping_attributes(lc_map_id)
            if (mapping_attrs is None or
                    'status' not in mapping_attrs):
                break
            if mapping_attrs['status'] == 'prepared':
                mapping_ready = True
                break
            elif mapping_attrs['status'] == 'stopped':
                self.executor.prestartlcmap(lc_map_id, restore)
            elif mapping_attrs['status'] != 'preparing':
                msg = (_('Unexecpted mapping status %(status)s for mapping '
                         '%(id)s. Attributes: %(attr)s.')
                       % {'status': mapping_attrs['status'],
                          'id': lc_map_id,
                          'attr': mapping_attrs})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            greenthread.sleep(self.WAIT_TIME)

        if not mapping_ready:
            msg = (_('Mapping %(id)s prepare failed to complete within the'
                     'allotted %(to)d seconds timeout. Terminating.')
                   % {'id': lc_map_id,
                      'to': timeout})
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)

    def stop_lcmap(self, lc_map_id, split=False):
        self.executor.stoplcmap(lc_map_id, split)

    def rm_lcmap(self, lc_map_id):
        self.executor.rmlcmap(lc_map_id)

    # Consistency Group
    def start_lc_consistgrp(self, lc_consistgrp):
        self.executor.startlcconsistgrp(lc_consistgrp)

    def create_lc_consistgrp(self, lc_consistgrp):
        self.executor.mklcconsistgrp(lc_consistgrp)

    def delete_lc_consistgrp(self, lc_consistgrp):
        self.executor.rmlcconsistgrp(lc_consistgrp)

    def stop_lc_consistgrp(self, lc_consistgrp):
        self.executor.stoplcconsistgrp(lc_consistgrp)

    def run_group_snapshots(self, lc_group, snapshots, state,
                            config, timeout, snap_name_generator, vol_name_generator):
        model_update = {'status': 'available'}
        snapshots_model_update = []
        try:
            for snapshot in snapshots:
                opts = self.get_vdisk_params(config, state,
                                             snapshot['volume_type_id'])

                snap_name = snap_name_generator(snapshot)
                snap_vol_name = vol_name_generator(snapshot['volume'])
                opts['rsize'] = 0
                self.create_localcopy_to_consistgrp(snap_vol_name,
                                                    snap_name,
                                                    lc_group,
                                                    opts)

            self.prepare_lc_consistgrp(lc_group, timeout)
            self.start_lc_consistgrp(lc_group)
            # There is CG limitation that could not create more than 128 CGs.
            # After start CG, we delete CG to avoid CG limitation.
            # Cinder general will maintain the group and snapshots
            # relationship.
            self.delete_lc_consistgrp(lc_group)
        except exception.VolumeBackendAPIException as err:
            model_update['status'] = 'error'
            # Release cg
            self.delete_lc_consistgrp(lc_group)
            LOG.error(_("Failed to create Group_Snapshot. "
                        "Exception: %s."), err)

        for snapshot in snapshots:
            snapshots_model_update.append(
                {'id': snapshot['id'],
                 'status': model_update['status'],
                 'replication_status': 'not-capable'})

        return model_update, snapshots_model_update

    def delete_group_snapshots(self, lc_group, snapshots, snap_name_generator):
        """Delete localcopy maps and group."""
        model_update = {'status': 'deleted'}
        snapshots_model_update = []

        try:
            for snapshot in snapshots:
                snap_name = snap_name_generator(snapshot)
                if self.get_vdisk_attributes(snap_name) is not None:
                    self.executor.rmvdisk(snap_name, True)
        except exception.VolumeBackendAPIException as err:
            model_update['status'] = 'error_deleting'
            LOG.error(_("Failed to delete the snapshot %(snap)s of "
                        "Group_Snapshot. Exception: %(exception)s."),
                      {'snap': snapshot['id'], 'exception': err})

        for snapshot in snapshots:
            snapshots_model_update.append(
                {'id': snapshot['id'],
                 'status': model_update['status']})

        return model_update, snapshots_model_update

    def prepare_lc_consistgrp(self, lc_consistgrp, timeout):
        """Prepare LC Consistency Group."""
        self.executor.prestartlcconsistgrp(lc_consistgrp)

        def prepare_lc_consistgrp_success():
            mapping_ready = False
            mapping_attrs = self._get_localcopy_consistgrp_attr(lc_consistgrp)
            if (mapping_attrs is None or
                    'status' not in mapping_attrs):
                return mapping_ready
            if mapping_attrs['status'] == 'prepared':
                mapping_ready = True
            elif mapping_attrs['status'] == 'stopped':
                self.executor.prestartlcconsistgrp(lc_consistgrp)
            elif mapping_attrs['status'] != 'preparing':
                msg = (_('Unexpected mapping status %(status)s for mapping'
                         '%(id)s. Attributes: %(attr)s.') %
                       {'status': mapping_attrs['status'],
                        'id': lc_consistgrp,
                        'attr': mapping_attrs})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            return mapping_ready
        self.wait_for_a_condition(prepare_lc_consistgrp_success, timeout)

    def create_group_from_source(self, group, lc_group,
                                 sources, targets, state,
                                 config, timeout, src_name_generator, tgt_name_generator):
        """Create group from source"""
        LOG.debug('Enter: create_group_from_source: group %(group)s'
                  ' source %(source)s, target %(target)s',
                  {'group': lc_group, 'source': sources, 'target': targets})
        model_update = {'status': 'available'}
        ctxt = context.get_admin_context()
        try:
            for source, target in zip(sources, targets):
                opts = self.get_vdisk_params(config, state,
                                             source['volume_type_id'])
                pool = volume_utils.extract_host(target['host'], 'pool')

                source_name = src_name_generator(source)
                target_name = tgt_name_generator(target)

                self.create_localcopy_to_consistgrp(source_name,
                                                    target_name,
                                                    lc_group,
                                                    opts,
                                                    True, pool=pool)
            self.prepare_lc_consistgrp(lc_group, timeout)
            self.start_lc_consistgrp(lc_group)
            self.delete_lc_consistgrp(lc_group)
            volumes_model_update = self._get_volume_model_updates(
                ctxt, targets, group['id'], model_update['status'])
        except exception.VolumeBackendAPIException as err:
            model_update['status'] = 'error'
            volumes_model_update = self._get_volume_model_updates(
                ctxt, targets, group['id'], model_update['status'])
            with excutils.save_and_reraise_exception():
                self.delete_lc_consistgrp(lc_group)
                LOG.error(_("Failed to create group from group_snapshot. "
                            "Exception: %s"), err)
            return model_update, volumes_model_update

        LOG.debug('Leave: create_cg_from_source.')
        return model_update, volumes_model_update

    def create_rc_consistgrp(self, group_name, system_id):
        self.executor.mkrcconsistgrp(group_name, system_id)

    def delete_rc_consistgrp(self, group_name):
        self.executor.rmrcconsistgrp(group_name)

    def chrcrelationship(self, key, rcrel, group_name=None):
        resp = self.executor.lsrcrelationship(key, rcrel)
        if resp:
            rcrel_dict = resp[0]
        else:
            with excutils.save_and_reraise_exception():
                LOG.error(_("The rcrelationship %s is not exist ,"
                            "chrcrelationship failed. "), rcrel)
        if rcrel_dict and rcrel_dict['consistency_group_name'] == group_name:
            LOG.warning('relationship %(rel)s is aleady added to group'
                        '%(grp)s.',{'rel':rcrel_dict['name'],'grp':group_name})
            return

        if not group_name and rcrel_dict['consistency_group_name'] == '':
            LOG.warning('relationship %(rel)s is aleady removed from group'
                        '%(grp)s.',{'rel':rcrel_dict['name'],'grp':group_name})
            return

        self.executor.chrcrelationship(rcrel, group_name)

    def start_rc_consistgrp(self, rccg, primary=None):
        self.executor.startrcconsistgrp(rccg, primary)

    def stop_rc_consistgrp(self, rccg, access=False):
        self.executor.stoprcconsistgrp(rccg, access)

    def _get_volume_model_updates(self, ctxt, volumes, cg_id,
                                  status='available'):
        """Update the volume model's status and return it."""
        volume_model_updates = []
        LOG.info(_(
            "Updating status for CG: %(id)s."),
            {'id': cg_id})
        if volumes:
            for volume in volumes:
                volume_model_updates.append({'id': volume['id'],
                                             'status': status})
        else:
            LOG.info(_("No volume found for CG: %(cg)s."),
                     {'cg': cg_id})
        return volume_model_updates

    def run_localcopy(self, source, target, timeout, copy_rate, clean_rate,
                      full_copy=True, restore=False):
        """Create a LocalCopy mapping from the source to the target."""
        LOG.debug('Enter: run_localcopy: execute LocalCopy from source '
                  '%(source)s to target %(target)s.',
                  {'source': source, 'target': target})

        lc_map_id = self.executor.mklcmap(source, target, full_copy, copy_rate,
                                          clean_rate)
        self._prepare_lc_map(lc_map_id, timeout, restore)
        self.executor.startlcmap(lc_map_id, restore)

        LOG.debug('Leave: run_localcopy: LocalCopy started from '
                  '%(source)s to %(target)s.',
                  {'source': source, 'target': target})

    def create_localcopy_to_consistgrp(self, source, target, consistgrp,
                                       opts, full_copy=False, pool=None):
        """Create a LocalCopy mapping and add to consistent group."""
        LOG.debug('Enter: create_localcopy_to_consistgrp: create LocalCopy'
                  ' from source %(source)s to target %(target)s'
                  'Then add the localcopy to %(cg)s.',
                  {'source': source, 'target': target, 'cg': consistgrp})

        src_attrs = self.get_vdisk_attributes(source)
        if src_attrs is None:
            msg = (_('create_copy: Source vdisk %(src)s '
                     'does not exist.') % {'src': source})
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)

        src_size = src_attrs['capacity']
        opts['iogrp'] = src_attrs['IO_group_id']
        pool = src_attrs['mdisk_grp_name']
        src_node = src_attrs['preferred_node_id']
        if src_node and int(src_node) != 0:
            opts['node'] = src_node

        self.create_vdisk(target, src_size, 'b', pool, opts)

        self.executor.mklcmap(source, target, full_copy,
                         opts['localcopyrate'],
                         opts['cleanrate'],
                         consistgrp=consistgrp)

        LOG.debug('Leave: create_localcopy_to_consistgrp: '
                  'LocalCopy started from  %(source)s to %(target)s.',
                  {'source': source, 'target': target})

    def get_vdisk_lc_mappings(self, vdisk, type='all'):
        """Return LocalCopy mappings that this vdisk is associated with."""
        src_mappings = []
        tgt_mappings = []
        lcmappings = self.executor.lsvdisklcmappings(vdisk)
        for lcmapping in lcmappings:
            attrs = self.get_localcopy_mapping_attributes(lcmapping['id'])
            if attrs is None:
                continue
            if attrs['source_vdisk_name'] == vdisk and type in ['all', 'source']:
                src_mappings.append(attrs)
            elif attrs['target_vdisk_name'] == vdisk and type in ['all', 'target']:
                tgt_mappings.append(attrs)

        return src_mappings, tgt_mappings

    def get_localcopy_mapping_attributes(self, lc_map_id):
        resp = self.executor.lslcmap(lc_map_id)
        if not resp:
            return None
        return resp[0]

    def _get_localcopy_consistgrp_attr(self, lc_consistgrp):
        resp = self.executor.lslcconsistgrp(lc_consistgrp)
        if not resp:
            return None
        return resp[0]

    def get_remotecopy_consistgrp_attr(self, rc_consistgrp):
        resp = self.executor.lsrcconsistgrp(rc_consistgrp)
        if not resp:
            return None
        return resp[0]

    def _check_vdisk_lc_mappings(self, name,
                                 allow_snaps=True, allow_lctgt=False, type='all'):
        """LocalCopy mapping check helper."""
        LOG.debug('Loopcall: _check_vdisk_lc_mappings(), vdisk %s.', name)
        wait_for_copy = False
        rmlcmap_failed_e = None
        src_mappings, tgt_mappings = self.get_vdisk_lc_mappings(name,type)
        for mappings in [src_mappings, tgt_mappings]:
            for attrs in mappings:
                map_id = attrs['id']
                source = attrs['source_vdisk_name']
                target = attrs['target_vdisk_name']
                copy_rate = attrs['copy_rate']
                status = attrs['status']
                progress = attrs['progress']

                if allow_lctgt and target == name and status == 'copying':
                    src_mappings, _ = self.get_vdisk_lc_mappings(target, type='source')
                    if src_mappings:
                        self.executor.stoplcmap(map_id, split=True)
                    else:
                        self.executor.stoplcmap(map_id, split=False)
                    attrs = self.get_localcopy_mapping_attributes(map_id)
                    if attrs:
                        status = attrs['status']
                    else:
                        return

                if copy_rate == '0':
                    if source == name:
                        # Vdisk with snapshots. Return False if snapshot
                        # not allowed.
                        if not allow_snaps:
                            raise loopingcall.LoopingCallDone(retvalue=False)
                        self.executor.chlcmap(map_id, copyrate='100', autodel='on')
                        wait_for_copy = True
                    else:
                        # A snapshot
                        if status in ['copying', 'prepared']:
                            src_mappings, _ = self.get_vdisk_lc_mappings(target, type='source')
                            if src_mappings:
                                self.executor.stoplcmap(map_id, split=True)
                            else:
                                self.executor.stoplcmap(map_id, split=False)
                            # Need to wait for the lcmap to change to
                            # stopped state before remove lcmap
                            wait_for_copy = True
                        elif status in ['stopping', 'preparing']:
                            wait_for_copy = True
                        else:
                            try:
                                self.executor.rmlcmap(map_id)
                            except exception.VolumeBackendAPIException as exp:
                                rmlcmap_failed_e = exp
                # Case 4: Copy in progress - wait and will autodelete
                else:
                    if status == 'prepared':
                        self.executor.stoplcmap(map_id)
                        self.executor.rmlcmap(map_id)
                    elif status in ['idle_or_copied', 'stopped']:
                        # Prepare failed or stopped
                        self.executor.rmlcmap(map_id)
                    elif progress == '100' and status == 'copying':
                        src_mappings, _ = self.get_vdisk_lc_mappings(target, type='source')
                        if src_mappings:
                            self.executor.stoplcmap(map_id, split=True)
                        else:
                            self.executor.stoplcmap(map_id, split=False)
                        wait_for_copy = True
                    else:
                        wait_for_copy = True
            if wait_for_copy:
                return

        if rmlcmap_failed_e is not None:
            raise rmlcmap_failed_e

        raise loopingcall.LoopingCallDone(retvalue=True)

    def ensure_vdisk_no_lc_mappings(self, name, allow_snaps=True,
                                    allow_lctgt=False, type='all'):
        """Ensure vdisk has no localcopy mappings."""
        timer = loopingcall.FixedIntervalLoopingCall(
            self._check_vdisk_lc_mappings, name,
            allow_snaps, allow_lctgt, type)
        # Create a timer greenthread. The default volume service heart
        # beat is every 10 seconds. The localcopy usually takes hours
        # before it finishes. Don't set the sleep interval shorter
        # than the heartbeat. Otherwise volume service heartbeat
        # will not be serviced.
        LOG.debug('Calling _ensure_vdisk_no_lc_mappings: vdisk %s.',
                  name)
        ret = timer.start(interval=self.loop_check_interval).wait()
        timer.stop()
        return ret

    def _check_vdisk_formatting(self, name):
        LOG.debug('Loopcall: _check_vdisk_formatting(), vdisk %s.', name)
        formatting_count = 0
        attrs = self.executor.lsvdisks_from_filter('volume_name', name)
        for attr in attrs:
            if attr['formatting'] == 'yes':
                formatting_count += 1

        if formatting_count == 0:
            raise loopingcall.LoopingCallDone(retvalue=True)
        else:
            return

    def ensure_vdisk_not_formatting(self, name):
        """Ensure vdisk is not formatting."""
        timer = loopingcall.FixedIntervalLoopingCall(
            self._check_vdisk_formatting, name)
        # Create a timer greenthread. The default volume service heart
        # beat is every 10 seconds. The localcopy usually takes hours
        # before it finishes. Don't set the sleep interval shorter
        # than the heartbeat. Otherwise volume service heartbeat
        # will not be serviced.
        LOG.debug('Calling _ensure_vdisk_not_formatting: vdisk %s.', name)
        ret = timer.start(interval=self.loop_check_interval).wait()
        timer.stop()
        return ret

    def _check_rollback_snapshot_mappings(self, vol_name, snap_name):
        """Rollback snapshot mapping check helper."""
        LOG.debug('Loopcall: _check_rollback_snapshot_mappings(), vdisk %s.',
                   snap_name)
        wait_for_copy = False

        src_mappings, _ = self.get_vdisk_lc_mappings(snap_name, type='source')
        for attrs in src_mappings:
            source = attrs['source_vdisk_name']
            target = attrs['target_vdisk_name']
            copy_rate = attrs['copy_rate']
            if source == snap_name and target == vol_name and copy_rate != '0':
                wait_for_copy = True
        if wait_for_copy:
            return
        raise loopingcall.LoopingCallDone(retvalue=True)

    def ensure_rollback_snapshot_ok(self, vol_name, snap_name):
        """Ensure rollback_snapshot has finished"""
        timer = loopingcall.FixedIntervalLoopingCall(
            self._check_rollback_snapshot_mappings, vol_name, snap_name
        )
        LOG.debug('Enter ensure_rollback_snapshot_ok: sourcename: %s', snap_name)
        timer.start(interval=5).wait()
        LOG.debug('Leave ensure_rollback_snapshot_ok: sourcename: %s', snap_name)

    def start_relationship(self, volume_name, primary=None):
        vol_attrs = self.get_vdisk_attributes(volume_name)
        if vol_attrs['RC_name']:
            self.executor.startrcrelationship(vol_attrs['RC_name'], primary)

    def start_rccg(self, rccg, primary=None):
        self.executor.startrcconsistgrp(rccg, primary)

    def stop_rccg(self, rccg, access=False):
        self.executor.stoprcconsistgrp(rccg, access)

    def stop_relationship(self, volume_name, access=False):
        vol_attrs = self.get_vdisk_attributes(volume_name)
        if vol_attrs['RC_name']:
            self.executor.stoprcrelationship(vol_attrs['RC_name'], access=access)

    def create_relationship(self, master, aux, system, asynccopy, rccg_name=None):
        try:
            rc_id = self.executor.mkrcrelationship(master, aux, system,
                                              asynccopy)
        except exception.VolumeBackendAPIException as exp:
            # CMMVC5959E is the code in InStorage, meaning that
            # there is a relationship that already has this name on the
            # master cluster.
            if 'CMMVC5959E' not in six.text_type(exp):
                # If there is no relation between the primary and the
                # secondary back-end storage, the exception is raised.
                raise
        if rc_id:
            self.start_relationship(master)
        if rc_id and rccg_name:
            self.chrcrelationship('id', rc_id, rccg_name)

    def delete_relationship(self, volume_name):
        vol_attrs = self.get_vdisk_attributes(volume_name)
        if vol_attrs['RC_name']:
            self.executor.rmrcrelationship(vol_attrs['RC_name'], True)

    def get_relationship_info(self, volume_name):
        vol_attrs = self.get_vdisk_attributes(volume_name)
        if not vol_attrs or not vol_attrs['RC_name']:
            LOG.debug(_("Unable to get remote copy information for "
                       "volume %s"), volume_name)
            return None

        relationship = self.executor.lsrcrelationship('name', vol_attrs['RC_name'])
        return relationship[0] if len(relationship) > 0 else None

    def delete_rc_volume(self, volume_name, target_vol=False):
        vol_name = volume_name
        if target_vol:
            vol_name = instorage_const.REPLICA_AUX_VOL_PREFIX + volume_name

        try:
            rel_info = self.get_relationship_info(vol_name)
            if rel_info:
                self.delete_relationship(vol_name)
            self.delete_vdisk(vol_name, False)
        except Exception as exp:
            msg = (_('Unable to delete the volume for '
                     'volume %(vol)s. Exception: %(err)s.'),
                   {'vol': vol_name, 'err': six.text_type(exp)})
            LOG.exception(msg)
            raise exception.VolumeDriverException(message=msg)

    def switch_relationship(self, relationship, aux=True):
        self.executor.switchrelationship(relationship, aux)

    def get_partnership_info(self, system_name):
        partnership = self.executor.lspartnership(system_name)
        return partnership[0] if len(partnership) > 0 else None

    def get_partnershipcandidate_info(self, system_name):
        candidates = self.executor.lspartnershipcandidate()
        for candidate in candidates:
            if system_name == candidate['name']:
                return candidate
        return None

    def mkippartnership(self, ip_v4, bandwith=1000, copyrate=50):
        self.executor.mkippartnership(ip_v4, bandwith, copyrate)

    def mkfcpartnership(self, system_name, bandwith=1000, copyrate=50):
        self.executor.mkfcpartnership(system_name, bandwith, copyrate)

    def chpartnership(self, partnership_id):
        self.executor.chpartnership(partnership_id)

    @staticmethod
    def can_migrate_to_host(host, state):
        if 'location_info' not in host['capabilities']:
            return None
        info = host['capabilities']['location_info']
        try:
            (dest_type, dest_id, dest_pool) = info.split(':')
        except ValueError:
            return None
        if (dest_type != 'InStorageMCSDriver' or dest_id !=
                state['system_id']):
            return None
        return dest_pool

    def add_vdisk_qos(self, vdisk, qos):
        """Add the QoS configuration to the volume."""
        for key, value in qos.items():
            param = self.mcs_qos_keys[key]['param']
            if key in self.mcs_qos_keys:
                if key == 'IOThrottling':
                    self.executor.chvdisk(vdisk, ['-' + param, str(value)])
                if key == 'Bandwidth':
                    self.executor.chvdisk(vdisk, ['-' + param, str(value), '-unitmb'])

    def update_vdisk_qos(self, vdisk, qos):
        """Update all the QoS in terms of a key and value.

        mcs_qos_keys saves all the supported QoS parameters. Going through
        this dict, we set the new values to all the parameters. If QoS is
        available in the QoS configuration, the value is taken from it;
        if not, the value will be set to default.
        """
        for key, value in self.mcs_qos_keys.items():
            param = value['param']
            if key in qos.keys():
                # If the value is set in QoS, take the value from
                # the QoS configuration.
                qos_value = qos[key]
            else:
                # If not, set the value to default.
                qos_value = value['default']
            if key == 'IOThrottling':
                self.executor.chvdisk(vdisk, ['-' + param, str(qos_value)])
            if key == 'Bandwidth':
                self.executor.chvdisk(vdisk, ['-' + param, str(qos_value), '-unitmb'])

    def disable_vdisk_qos(self, vdisk, qos):
        """Disable the QoS."""
        for key, value in qos.items():
            if key in self.mcs_qos_keys:
                param = self.mcs_qos_keys[key]['param']
                # Take the default value.
                value = self.mcs_qos_keys[key]['default']
                if key == 'IOThrottling':
                    self.executor.chvdisk(vdisk, ['-' + param, value])
                if key == 'Bandwidth':
                    self.executor.chvdisk(vdisk, ['-' + param, value, '-unitmb'])

    def change_vdisk_options(self, vdisk, changes, opts):
        if 'warning' in opts:
            opts['warning'] = '%s%%' % str(opts['warning'])
        if 'intier' in opts:
            opts['intier'] = 'on' if opts['intier'] else 'off'
        if 'autoexpand' in opts:
            opts['autoexpand'] = 'on' if opts['autoexpand'] else 'off'
        # can be changed only when pool is SSA
        if 'deduped' in opts:
            opts['deduped'] = 'on' if opts['deduped'] else 'off'
        # can be changed only when pool is SSA
        if 'compressed' in opts:
            opts['compressed'] = 'on' if opts['compressed'] else 'off'

        for key in changes:
            self.executor.chvdisk(vdisk, ['-' + key, opts[key]])

    def change_vdisk_iogrp(self, vdisk, state, iogrp):
        if state['code_level'] < (3, 0, 0, 0):
            LOG.debug('Ignore change IO group as storage code level is '
                      '%(code_level)s, below the required 3, 0, 0, 0.',
                      {'code_level': state['code_level']})
        else:
            self.executor.movevdisk(vdisk, str(iogrp[0]))
            self.executor.addvdiskaccess(vdisk, str(iogrp[0]))
            self.executor.rmvdiskaccess(vdisk, str(iogrp[1]))

    def vdisk_by_uid(self, vdisk_uid):
        """Returns the properties of the vdisk with the specified UID.

        Returns None if no such disk exists.
        """

        vdisks = self.executor.lsvdisks_from_filter('vdisk_UID', vdisk_uid)

        if len(vdisks) == 0:
            return None

        if len(vdisks) != 1:
            msg = (_('Expected single vdisk returned from lsvdisk when '
                     'filtering on vdisk_UID.  %(count)s were returned.') %
                   {'count': len(vdisks)})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        vdisk = vdisks.result[0]

        return self.executor.lsvdisk(vdisk['name'])

    def is_vdisk_in_use(self, vdisk):
        """Returns True if the specified vdisk is mapped to at least 1 host."""
        resp = self.executor.lsvdiskhostmap(vdisk)
        return len(resp) != 0

    def rename_vdisk(self, vdisk, new_name):
        self.executor.chvdisk(vdisk, ['-name', new_name])

    def change_vdisk_primary_copy(self, vdisk, copy_id):
        self.executor.chvdisk(vdisk, ['-primary', copy_id])

    def get_host_info(self, hostname):
        return self.executor.lshost(hostname)

    def migratevdisk(self, vdisk, dest_pool, copy_id='0'):
        self.executor.migratevdisk(vdisk, dest_pool, copy_id)

    def get_vdisk_dependent(self, vdisk):
        return self.executor.lsrmvdiskdependentmaps(vdisk)


class InStorageExector(object):
    """Exector interface to inspur InStorage systems."""

    def __init__(self, run_ssh):
        self._ssh = run_ssh

    def _run_ssh(self, ssh_cmd):
        try:
            return self._ssh(ssh_cmd)
        except processutils.ProcessExecutionError as exp:
            msg = (_('CLI Exception output:\n command: %(cmd)s\n '
                     'stdout: %(out)s\n stderr: %(err)s.') %
                   {'cmd': ssh_cmd,
                    'out': exp.stdout,
                    'err': exp.stderr})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

    @cinder_utils.retry(exception.VolumeBackendAPIException, interval=5, retries=3, backoff_rate=1)
    def run_ssh_inq(self, ssh_cmd, delim='!', with_header=False):
        """Run an SSH command and return parsed output."""
        raw = self._run_ssh(ssh_cmd)
        return CLIParser(raw, ssh_cmd=ssh_cmd, delim=delim,
                         with_header=with_header)

    def run_ssh_assert_no_output(self, ssh_cmd):
        """Run an SSH command and assert no output returned."""
        out, _ = self._run_ssh(ssh_cmd)
        if len(out.strip()) != 0:
            msg = (_('Expected no output from CLI command %(cmd)s, '
                     'got %(out)s.') % {'cmd': ' '.join(ssh_cmd), 'out': out})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

    def run_ssh_check_created(self, ssh_cmd):
        """Run an SSH command and return the ID of the created object."""
        out, err = self._run_ssh(ssh_cmd)
        try:
            match_obj = re.search(r'\[([0-9]+)\],? successfully created', out)
            return match_obj.group(1)
        except (AttributeError, IndexError):
            msg = (_('Failed to parse CLI output:\n command: %(cmd)s\n '
                     'stdout: %(out)s\n stderr: %(err)s.') %
                   {'cmd': ssh_cmd,
                    'out': out,
                    'err': err})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

    def lsnode(self, node_id=None):
        with_header = True
        ssh_cmd = ['mcsinq', 'lsnode', '-delim', '!']
        if node_id:
            with_header = False
            ssh_cmd.append(node_id)
        return self.run_ssh_inq(ssh_cmd, with_header=with_header)

    def retry_func(func, attempts=3):
        def wrapper(self, *args, **kw):
            for retry_num in range(attempts):
                try:
                    return func(self, *args, **kw)
                except IndexError:
                    if retry_num == attempts - 1:
                        LOG.error('%(func)s, Retry was exceeded. Unable '
                                'to execute relevant command',
                                {'func': func.__name__})
                        raise
                    LOG.warning('%(func)s, retry, '
                                'TryNum: %(rn)s.',
                                {'func': func.__name__,
                                'rn': retry_num + 1})
                    greenthread.sleep(5)
        return wrapper

    @retry_func
    def lslicense(self):
        ssh_cmd = ['mcsinq', 'lslicense', '-delim', '!']
        return self.run_ssh_inq(ssh_cmd)[0]

    @retry_func
    def lsguicapabilities(self):
        ssh_cmd = ['mcsinq', 'lsguicapabilities', '-delim', '!']
        return self.run_ssh_inq(ssh_cmd)[0]

    @retry_func
    def lssystem(self):
        ssh_cmd = ['mcsinq', 'lssystem', '-delim', '!']
        return self.run_ssh_inq(ssh_cmd)[0]

    @retry_func
    def lssystemcapcities(self):
        ssh_cmd = ['mcsinq', 'lssystem', '-bytes', '-delim', '!']
        return self.run_ssh_inq(ssh_cmd)[0]

    @retry_func
    def lsmdiskgrp(self, pool):
        ssh_cmd = ['mcsinq', 'lsmdiskgrp', '-bytes', '-delim', '!',
                   '"%s"' % pool]
        return self.run_ssh_inq(ssh_cmd)[0]

    def lsiogrp(self):
        ssh_cmd = ['mcsinq', 'lsiogrp', '-delim', '!']
        return self.run_ssh_inq(ssh_cmd, with_header=True)

    def lsportip(self):
        ssh_cmd = ['mcsinq', 'lsportip', '-delim', '!']
        return self.run_ssh_inq(ssh_cmd, with_header=True)

    def lshost(self, host=None):
        with_header = True
        ssh_cmd = ['mcsinq', 'lshost', '-delim', '!']
        if host:
            with_header = False
            ssh_cmd.append('"%s"' % host)
        return self.run_ssh_inq(ssh_cmd, with_header=with_header)

    def lsiscsiauth(self):
        ssh_cmd = ['mcsinq', 'lsiscsiauth', '-delim', '!']
        return self.run_ssh_inq(ssh_cmd, with_header=True)

    def lsfabric(self, wwpn=None, host=None):
        ssh_cmd = ['mcsinq', 'lsfabric', '-delim', '!']
        if wwpn:
            ssh_cmd.extend(['-wwpn', wwpn])
        elif host:
            ssh_cmd.extend(['-host', '"%s"' % host])
        else:
            msg = (_('Must pass wwpn or host to lsfabric.'))
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)
        return self.run_ssh_inq(ssh_cmd, with_header=True)

    def lsrcrelationship(self, key, rc_rel):
        key_value = '%s=%s' % (key, rc_rel)
        ssh_cmd = ['mcsinq', 'lsrcrelationship', '-filtervalue',
                   key_value, '-delim', '!']
        return self.run_ssh_inq(ssh_cmd, with_header=True)

    def lspartnership(self, system_name):
        key_value = 'name=%s' % system_name
        ssh_cmd = ['mcsinq', 'lspartnership', '-filtervalue',
                   key_value, '-delim', '!']
        return self.run_ssh_inq(ssh_cmd, with_header=True)

    def lspartnershipcandidate(self):
        ssh_cmd = ['mcsinq', 'lspartnershipcandidate', '-delim', '!']
        return self.run_ssh_inq(ssh_cmd, with_header=True)

    def lsvdiskhostmap(self, vdisk):
        ssh_cmd = ['mcsinq', 'lsvdiskhostmap', '-delim', '!', '"%s"' % vdisk]
        out, err = self._ssh(ssh_cmd, check_exit_code=False)
        if not err:
            return CLIParser((out, err), ssh_cmd=ssh_cmd, delim='!',
                             with_header=True)
        if 'CMMVC5754E' in err:
            return []
        msg = (_('CLI Exception output:\n command: %(cmd)s\n '
                 'stdout: %(out)s\n stderr: %(err)s.') %
               {'cmd': ssh_cmd,
                'out': out,
                'err': err})
        LOG.error(msg)
        raise exception.VolumeBackendAPIException(data=msg)

    def lshostvdiskmap(self, host):
        ssh_cmd = ['mcsinq', 'lshostvdiskmap', '-delim', '!', '"%s"' % host]
        return self.run_ssh_inq(ssh_cmd, with_header=True)

    @retry_func
    def lsvdisk(self, vdisk):
        """Return vdisk attributes or None if it doesn't exist."""
        ssh_cmd = ['mcsinq', 'lsvdisk', '-bytes', '-delim', '!',
                   '"%s"' % vdisk]
        out, err = self._ssh(ssh_cmd, check_exit_code=False)
        if not err:
            return CLIParser((out, err), ssh_cmd=ssh_cmd, delim='!',
                             with_header=False)[0]
        if 'CMMVC5754E' in err:
            return None
        msg = (_('CLI Exception output:\n command: %(cmd)s\n '
                 'stdout: %(out)s\n stderr: %(err)s.') %
               {'cmd': ssh_cmd,
                'out': out,
                'err': err})
        LOG.error(msg)
        raise exception.VolumeBackendAPIException(data=msg)

    def lsvdisks_from_filter(self, filter_name, value):
        """Performs an lsvdisk command, filtering the results as specified.

        Returns an iterable for all matching vdisks.
        """
        ssh_cmd = ['mcsinq', 'lsvdisk', '-bytes', '-delim', '!',
                   '-filtervalue', '%s=%s' % (filter_name, value)]
        return self.run_ssh_inq(ssh_cmd, with_header=True)

    def lsvdisklcmappings(self, vdisk):
        ssh_cmd = ['mcsinq', 'lsvdisklcmappings', '-delim', '!',
                   '"%s"' % vdisk]
        return self.run_ssh_inq(ssh_cmd, with_header=True)

    def lslcmap(self, lc_map_id):
        ssh_cmd = ['mcsinq', 'lslcmap', '-gui', '-filtervalue',
                   'id=%s' % lc_map_id, '-delim', '!']
        return self.run_ssh_inq(ssh_cmd, with_header=True)

    def lslcconsistgrp(self, lc_consistgrp):
        ssh_cmd = ['mcsinq', 'lslcconsistgrp', '-delim', '!', lc_consistgrp]
        out, err = self._ssh(ssh_cmd)
        return CLIParser((out, err), ssh_cmd=ssh_cmd, delim='!',
                         with_header=False)

    def lsrcconsistgrp(self, rc_consistgrp):
        ssh_cmd = ['mcsinq', 'lsrcconsistgrp', '-delim', '!', rc_consistgrp]
        out, err = self._ssh(ssh_cmd)
        return CLIParser((out, err), ssh_cmd=ssh_cmd, delim='!',
                         with_header=False)

    def lsvdiskcopy(self, vdisk=None, copy_id=None, pool=None):
        ssh_cmd = ['mcsinq', 'lsvdiskcopy', '-gui', '-bytes', '-delim', '!']
        with_header = True
        if copy_id:
            ssh_cmd += ['-copy', copy_id]
            with_header = False
        if vdisk:
            ssh_cmd += ['"%s"' % vdisk]
        if pool:
            ssh_cmd += ['-filtervalue', 'mdisk_grp_name=%s' % pool]
        return self.run_ssh_inq(ssh_cmd, with_header=with_header)

    @retry_func
    def lsvdisksyncprogress(self, vdisk, copy_id):
        ssh_cmd = ['mcsinq', 'lsvdisksyncprogress', '-delim', '!',
                   '-copy', copy_id, '"%s"' % vdisk]
        return self.run_ssh_inq(ssh_cmd, with_header=True)[0]

    def lsportfc(self, node_id):
        ssh_cmd = ['mcsinq', 'lsportfc', '-delim', '!',
                   '-filtervalue', 'node_id=%s' % node_id]
        return self.run_ssh_inq(ssh_cmd, with_header=True)

    def lsrmvdiskdependentmaps(self, vdisk):
        ssh_cmd = ['mcsinq', 'lsrmvdiskdependentmaps', '-delim', '!', '"%s"' % vdisk]
        return self.run_ssh_inq(ssh_cmd, with_header=True)

    @staticmethod
    def _create_port_arg(port_type, port_name):
        if port_type == 'initiator':
            port = ['-iscsiname']
        elif port_type == 'wwpn':
            port = ['-hbawwpn']
        else:
            port = ['-nqn']
        port.append(port_name)
        return port

    def mkhost(self, host_name, port_type, port_name, site='', protocol = 'scsi'):
        port = self._create_port_arg(port_type, port_name)
        ssh_cmd = ['mcsop', 'mkhost', '-force'] + port
        ssh_cmd += ['-name', '"%s"' % host_name]
        if site:
            ssh_cmd += ['-site', '"%s"' % site]
        if protocol == 'nvme':
            ssh_cmd += ['-protocol', 'nvme']
            ssh_cmd += ['-nvmetransport', 'roce']
        return self.run_ssh_check_created(ssh_cmd)

    def addhostport(self, host, port_type, port_name):
        port = self._create_port_arg(port_type, port_name)
        ssh_cmd = ['mcsop', 'addhostport', '-force'] + port + ['"%s"' % host]
        self.run_ssh_assert_no_output(ssh_cmd)

    def add_chap_secret(self, secret, host):
        ssh_cmd = ['mcsop', 'chhost', '-chapsecret', secret, '"%s"' % host]
        self.run_ssh_assert_no_output(ssh_cmd)

    def mkvdiskhostmap(self, host, vdisk, lun, multihostmap, protocol = 'scsi'):
        """Map vdisk to host.

        If vdisk already mapped and multihostmap is True, use the force flag.
        """
        ssh_cmd = ['mcsop', 'mkvdiskhostmap', '-host', '"%s"' % host, vdisk]

        if lun:
            ssh_cmd.insert(ssh_cmd.index(vdisk), '-scsi')
            ssh_cmd.insert(ssh_cmd.index(vdisk), lun)

        if multihostmap:
            ssh_cmd.insert(ssh_cmd.index('mkvdiskhostmap') + 1, '-force')
        try:
            self.run_ssh_check_created(ssh_cmd)
            if protocol == 'scsi':
                iogrp_lun_id_map = self.get_vdiskhostmapid(vdisk, host)
                if len(iogrp_lun_id_map) == 0:
                    msg = (_('mkvdiskhostmap error:\n command: %(cmd)s\n '
                            'lun: %(lun)s\n iogrp_lun_id_map: %(iogrp_lun_id_map)s') %
                        {'cmd': ssh_cmd,
                            'lun': lun,
                            'iogrp_lun_id_map': iogrp_lun_id_map})
                    LOG.error(msg)
                    raise exception.VolumeDriverException(message=msg)
                return iogrp_lun_id_map
            else:
                is_mapped = self.check_volume_mapped(vdisk, host)
                if not is_mapped:
                    msg = (_('mkvdiskhostmap error:\n command: %(cmd)s\n '
                            'lun: %(lun)s\n ') %
                        {'cmd': ssh_cmd,
                            'lun': lun})
                    LOG.error(msg)
                    raise exception.VolumeDriverException(message=msg)
                return True
        except Exception as ex:
            if (not multihostmap and hasattr(ex, 'message') and
                    'CMMVC6071E' in ex.message):
                LOG.error(_('instorage_mcs_multihostmap_enabled is set '
                            'to False, not allowing multi host mapping.'))
                raise exception.VolumeDriverException(
                    message=_('CMMVC6071E The VDisk-to-host mapping was not '
                              'created because the VDisk is already mapped '
                              'to a host.\n"'))
            with excutils.save_and_reraise_exception():
                LOG.error(_('Error mapping VDisk-to-host'))

    def mkrcrelationship(self, master, aux, system, asynccopy):
        ssh_cmd = ['mcsop', 'mkrcrelationship', '-master', master,
                   '-aux', aux, '-cluster', system]
        if asynccopy:
            ssh_cmd.append('-async')
        return self.run_ssh_check_created(ssh_cmd)

    def rmrcrelationship(self, relationship, force=False):
        ssh_cmd = ['mcsop', 'rmrcrelationship']
        if force:
            ssh_cmd += ['-force']
        ssh_cmd += [relationship]
        self.run_ssh_assert_no_output(ssh_cmd)

    def switchrelationship(self, relationship, aux=True):
        primary = 'aux' if aux else 'master'
        ssh_cmd = ['mcsop', 'switchrcrelationship', '-primary',
                   primary, relationship]
        self.run_ssh_assert_no_output(ssh_cmd)

    def startrcrelationship(self, rc_rel, primary=None):
        ssh_cmd = ['mcsop', 'startrcrelationship', '-force']
        if primary:
            ssh_cmd.extend(['-primary', primary])
        ssh_cmd.append(rc_rel)
        self.run_ssh_assert_no_output(ssh_cmd)

    def startrcconsistgrp(self, rc_consistgrp, primary=None):
        ssh_cmd = ['mcsop', 'startrcconsistgrp', '-force']
        if primary:
            ssh_cmd.extend(['-primary', primary])
        ssh_cmd.append(rc_consistgrp)
        self.run_ssh_assert_no_output(ssh_cmd)

    def stoprcconsistgrp(self, rccg, access=False):
        ssh_cmd = ['mcsop', 'stoprcconsistgrp']
        if access:
            ssh_cmd.append('-access')
        ssh_cmd.append(rccg)
        self.run_ssh_assert_no_output(ssh_cmd)

    def stoprcrelationship(self, relationship, access=False):
        ssh_cmd = ['mcsop', 'stoprcrelationship']
        if access:
            ssh_cmd.append('-access')
        ssh_cmd.append(relationship)
        self.run_ssh_assert_no_output(ssh_cmd)

    def mkippartnership(self, ip_v4, bandwith=1000, backgroundcopyrate=50):
        ssh_cmd = ['mcsop', 'mkippartnership', '-type', 'ipv4',
                   '-clusterip', ip_v4, '-linkbandwidthmbits',
                   six.text_type(bandwith),
                   '-backgroundcopyrate', six.text_type(backgroundcopyrate)]
        return self.run_ssh_assert_no_output(ssh_cmd)

    def mkfcpartnership(self, system_name, bandwith=1000,
                        backgroundcopyrate=50):
        ssh_cmd = ['mcsop', 'mkfcpartnership', '-linkbandwidthmbits',
                   six.text_type(bandwith),
                   '-backgroundcopyrate', six.text_type(backgroundcopyrate),
                   system_name]
        return self.run_ssh_assert_no_output(ssh_cmd)

    def chpartnership(self, partnership_id, start=True):
        action = '-start' if start else '-stop'
        ssh_cmd = ['mcsop', 'chpartnership', action, partnership_id]
        return self.run_ssh_assert_no_output(ssh_cmd)

    def rmvdiskhostmap(self, host, vdisk):
        ssh_cmd = ['mcsop', 'rmvdiskhostmap', '-host', '"%s"' % host,
                   '"%s"' % vdisk]
        self.run_ssh_assert_no_output(ssh_cmd)

    def get_vdiskhostmapid(self, vdisk, host):
        # return the map that the key is io_group_id and the value is lun_id
        result_map = {}
        resp = self.lsvdiskhostmap(vdisk)
        for mapping_info in resp:
            if mapping_info['host_name'] == host:
                result_map[mapping_info['IO_group_id']] = int(mapping_info['SCSI_id'])
        return result_map

    def check_volume_mapped(self, vdisk, host):
        resp = self.lsvdiskhostmap(vdisk)
        for mapping_info in resp:
            if mapping_info['host_name'] == host:
                return True
        return False

    def rmhost(self, host):
        ssh_cmd = ['mcsop', 'rmhost', '"%s"' % host]
        self.run_ssh_assert_no_output(ssh_cmd)

    def mkvolume(self, name, size, units, poollist, iogrplist, params):
        ssh_cmd = ['mcsop', 'mkvolume', '-name', name,
                   '-size', size, '-unit', units,
                   '-pool', poollist] + params

        try:
            return self.run_ssh_check_created(ssh_cmd)
        except Exception as ex:
            if hasattr(ex, 'msg') and 'CMMVC6035E' in ex.msg:
                LOG.warning(_('CMMVC6035E The volume already exists '))
                return
            with excutils.save_and_reraise_exception():
                LOG.exception(_('Failed to create vdisk %(vol)s.'),
                              {'vol': name})

    def mkvdisk(self, name, size, units, pool, opts, params, nofmtdisk=False):
        ssh_cmd = ['mcsop', 'mkvdisk', '-name', name, '-mdiskgrp',
                   '"%s"' % pool, '-iogrp', six.text_type(opts['iogrp']),
                   '-size', size, '-unit', units] + params
        if nofmtdisk:
            ssh_cmd.extend(['-nofmtdisk'])
        if opts['node']:
            ssh_cmd.extend(['-node','%s' % opts['node']])
        if opts['access_iogrp']:
            ssh_cmd.extend(['-accessiogrp', '%s' % opts['access_iogrp']])
        try:
            return self.run_ssh_check_created(ssh_cmd)
        except Exception as ex:
            if hasattr(ex, 'msg') and 'CMMVC6372W' in ex.msg:
                vdisk = self.lsvdisk(name)
                if vdisk:
                    LOG.warning(_('CMMVC6372W The virtualized storage '
                                  'capacity that the cluster is using is '
                                  'approaching the virtualized storage '
                                  'capacity that is licensed.'))
                    return vdisk['id']
            elif hasattr(ex, 'msg') and 'CMMVC6035E' in ex.msg:
                LOG.warning(_('CMMVC6035E The volume already exists '))
                return
            with excutils.save_and_reraise_exception():
                LOG.exception(_('Failed to create vdisk %(vol)s.'),
                              {'vol': name})

    def rmvolume(self, vdisk, force=True):
        ssh_cmd = ['mcsop', 'rmvolume']
        if force:
            ssh_cmd += ['-removehostmappings', '-removercrelationships',
                        '-removelcmaps', '-discardimage', '-cancelbackup']
        ssh_cmd += ['"%s"' % vdisk]
        self.run_ssh_assert_no_output(ssh_cmd)

    def rmvdisk(self, vdisk, force=True):
        ssh_cmd = ['mcsop', 'rmvdisk']
        if force:
            ssh_cmd += ['-force']
        ssh_cmd += ['"%s"' % vdisk]
        self.run_ssh_assert_no_output(ssh_cmd)

    def chvdisk(self, vdisk, params):
        ssh_cmd = ['mcsop', 'chvdisk'] + params + ['"%s"' % vdisk]
        self.run_ssh_assert_no_output(ssh_cmd)

    def movevdisk(self, vdisk, iogrp):
        ssh_cmd = ['mcsop', 'movevdisk', '-iogrp', iogrp, '"%s"' % vdisk]
        self.run_ssh_assert_no_output(ssh_cmd)

    def expandvdisksize(self, vdisk, amount):
        ssh_cmd = (
            ['mcsop', 'expandvdisksize', '-size', six.text_type(amount),
             '-unit', 'gb', '"%s"' % vdisk])
        self.run_ssh_assert_no_output(ssh_cmd)

    def mklcmap(self, source, target, full_copy, copy_rate, clean_rate, consistgrp=None):
        ssh_cmd = ['mcsop', 'mklcmap', '-source', '"%s"' % source, '-target',
                   '"%s"' % target, '-autodelete']
        if not full_copy:
            ssh_cmd.extend(['-copyrate', '0'])
        else:
            ssh_cmd.extend(['-copyrate', six.text_type(copy_rate)])
            ssh_cmd.extend(['-cleanrate', six.text_type(clean_rate)])
        if consistgrp:
            ssh_cmd.extend(['-consistgrp', consistgrp])
        out, err = self._ssh(ssh_cmd, check_exit_code=False)
        if 'successfully created' not in out:
            msg = (_('CLI Exception output:\n command: %(cmd)s\n '
                     'stdout: %(out)s\n stderr: %(err)s.') %
                   {'cmd': ssh_cmd,
                    'out': out,
                    'err': err})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        try:
            match_obj = re.search(r'LocalCopy Mapping, id \[([0-9]+)\], '
                                  'successfully created', out)
            lc_map_id = match_obj.group(1)
        except (AttributeError, IndexError):
            msg = (_('Failed to parse CLI output:\n command: %(cmd)s\n '
                     'stdout: %(out)s\n stderr: %(err)s.') %
                   {'cmd': ssh_cmd,
                    'out': out,
                    'err': err})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        return lc_map_id

    def prestartlcmap(self, lc_map_id, restore=False):
        ssh_cmd = ['mcsop', 'prestartlcmap']
        if restore:
            ssh_cmd.append('-restore')
        ssh_cmd.append(lc_map_id)
        self.run_ssh_assert_no_output(ssh_cmd)

    def startlcmap(self, lc_map_id, restore=False):
        ssh_cmd = ['mcsop', 'startlcmap']
        if restore:
            ssh_cmd.append('-restore')
        ssh_cmd.append(lc_map_id)
        self.run_ssh_assert_no_output(ssh_cmd)

    def prestartlcconsistgrp(self, lc_consist_group):
        ssh_cmd = ['mcsop', 'prestartlcconsistgrp', lc_consist_group]
        self.run_ssh_assert_no_output(ssh_cmd)

    def startlcconsistgrp(self, lc_consist_group):
        ssh_cmd = ['mcsop', 'startlcconsistgrp', lc_consist_group]
        self.run_ssh_assert_no_output(ssh_cmd)

    def stoplcconsistgrp(self, lc_consist_group):
        ssh_cmd = ['mcsop', 'stoplcconsistgrp', lc_consist_group]
        self.run_ssh_assert_no_output(ssh_cmd)

    def chlcmap(self, lc_map_id, copyrate='50', autodel='on'):
        ssh_cmd = ['mcsop', 'chlcmap', '-copyrate', copyrate,
                   '-autodelete', autodel, lc_map_id]
        self.run_ssh_assert_no_output(ssh_cmd)

    def stoplcmap(self, lc_map_id, split=False):
        ssh_cmd = ['mcsop', 'stoplcmap']
        if split:
            ssh_cmd += ['-split']
        ssh_cmd += ['"%s"' % lc_map_id]
        out, err = self._ssh(ssh_cmd, check_exit_code=False)
        if not err and not out:
            return
        LOG.warning('Operation stoplcmap exec failed for %(out)s %(err)s' %
                    {'out': out, 'err': err})
        if 'CMMVC5753E' in err:
            # CMMVC5753E lcmap not exist.
            return None
        msg = (_('CLI Exception output:\n command: %(cmd)s\n '
                 'stdout: %(out)s\n stderr: %(err)s.') %
               {'cmd': ssh_cmd,
                'out': out,
                'err': err})
        LOG.error(msg)
        raise exception.VolumeBackendAPIException(data=msg)

    @cinder_utils.retry(exception.VolumeBackendAPIException, interval=5, retries=60, backoff_rate=1)
    def rmlcmap(self, lc_map_id):
        ssh_cmd = ['mcsop', 'rmlcmap', '-force', lc_map_id]
        out, err = self._ssh(ssh_cmd, check_exit_code=False)
        if not err and not out:
            return
        LOG.warning('Operation rmlcmap exec failed for %(out)s %(err)s' %
                    {'out': out, 'err': err})
        if 'CMMVC5753E' in err:
            # CMMVC5753E lcmap not exist.
            return None
        msg = (_('CLI Exception output:\n command: %(cmd)s\n '
                 'stdout: %(out)s\n stderr: %(err)s.') %
               {'cmd': ssh_cmd,
                'out': out,
                'err': err})
        LOG.error(msg)
        raise exception.VolumeBackendAPIException(data=msg)

    def mklcconsistgrp(self, lc_consist_group):
        ssh_cmd = ['mcsop', 'mklcconsistgrp', '-name', lc_consist_group]
        return self.run_ssh_check_created(ssh_cmd)

    def rmlcconsistgrp(self, lc_consist_group):
        ssh_cmd = ['mcsop', 'rmlcconsistgrp', '-force', lc_consist_group]
        return self.run_ssh_assert_no_output(ssh_cmd)

    def mkrcconsistgrp(self, group_name, system_id):
        ssh_cmd = ['mcsop', 'mkrcconsistgrp', '-name', group_name,
                   '-cluster', system_id]
        return self.run_ssh_check_created(ssh_cmd)

    def rmrcconsistgrp(self, group_name):
        ssh_cmd = ['mcsop', 'rmrcconsistgrp', '-force', group_name]
        out, err = self._ssh(ssh_cmd, check_exit_code=False)
        if not err and not out:
            return
        LOG.warning('Operation rmrcconsistgrp exec failed for %(out)s %(err)s' %
                    {'out': out, 'err': err})
        if 'CMMVC5753E' in err:
            # CMMVC5753E lcmap not exist.
            return None
        msg = (_('CLI Exception output:\n command: %(cmd)s\n '
                 'stdout: %(out)s\n stderr: %(err)s.') %
               {'cmd': ssh_cmd,
                'out': out,
                'err': err})
        LOG.error(msg)
        raise exception.VolumeBackendAPIException(data=msg)

    def chrcrelationship(self, rcrel_name, rccg_name):
        ssh_cmd = ['mcsop', 'chrcrelationship']
        if rccg_name:
            ssh_cmd.extend(['-consistgrp', rccg_name])
        else:
            ssh_cmd.extend(['-noconsistgrp'])
        ssh_cmd.append(rcrel_name)
        return self.run_ssh_assert_no_output(ssh_cmd)

    def addvdiskcopy(self, vdisk, dest_pool, params):
        ssh_cmd = (['mcsop', 'addvdiskcopy'] +
                   params +
                   ['-mdiskgrp', '"%s"' %
                    dest_pool, '"%s"' %
                    vdisk])
        return self.run_ssh_check_created(ssh_cmd)

    def rmvdiskcopy(self, vdisk, copy_id):
        ssh_cmd = ['mcsop', 'rmvdiskcopy', '-copy', copy_id, '"%s"' % vdisk]
        self.run_ssh_assert_no_output(ssh_cmd)

    def addvolumecopy(self, vdisk_name, dest_pool, dest_iogrp, params):
        ssh_cmd = (
            ['mcsop', 'addvolumecopy'] +
            params +
            ['-iogrp', '"%s"' % dest_iogrp] +
            ['-pool', '"%s"' % dest_pool] +
            [vdisk_name]
        )
        self.run_ssh_assert_no_output(ssh_cmd)

    def rmvolumecopy(self, name, pool):
        ssh_cmd = ['mcsop', 'rmvolumecopy', '-pool', '"%s"' % pool, name]
        self.run_ssh_assert_no_output(ssh_cmd)

    def addvdiskaccess(self, vdisk, iogrp):
        ssh_cmd = ['mcsop', 'addvdiskaccess', '-iogrp', iogrp,
                   '"%s"' % vdisk]
        self.run_ssh_assert_no_output(ssh_cmd)

    def rmvdiskaccess(self, vdisk, iogrp):
        ssh_cmd = ['mcsop', 'rmvdiskaccess', '-iogrp', iogrp, '"%s"' % vdisk]
        self.run_ssh_assert_no_output(ssh_cmd)

    def migratevdisk(self, vdisk, dest_pool, copy_id='0'):
        ssh_cmd = ['mcsop', 'migratevdisk', '-mdiskgrp', dest_pool, '-copy',
                   copy_id, '-vdisk', vdisk]
        self.run_ssh_assert_no_output(ssh_cmd)


class CLIParser(object):
    """Parse MCS CLI output and generate iterable."""

    def __init__(self, raw, ssh_cmd=None, delim='!', with_header=True):
        super(CLIParser, self).__init__()
        if ssh_cmd:
            self.ssh_cmd = ' '.join(ssh_cmd)
        else:
            self.ssh_cmd = 'None'
        self.raw = raw
        self.delim = delim
        self.with_header = with_header
        self.result = self._parse()

    def select(self, *keys):
        for d_item in self.result:
            vs_list = []
            for k in keys:
                value = d_item.get(k, None)
                if isinstance(value, six.string_types) or value is None:
                    value = [value]
                if isinstance(value, list):
                    vs_list.append(value)
            for item in zip(*vs_list):
                if len(item) == 1:
                    yield item[0]
                else:
                    yield item

    def __getitem__(self, key):
        try:
            return self.result[key]
        except KeyError:
            msg = (_('Did not find the expected key %(key)s in %(fun)s: '
                     '%(raw)s.') % {'key': key, 'fun': self.ssh_cmd,
                                    'raw': self.raw})
            raise exception.VolumeBackendAPIException(data=msg)

    def __iter__(self):
        for item in self.result:
            yield item

    def __len__(self):
        return len(self.result)

    def _parse(self):
        def get_reader(content, delim):
            for line in content.lstrip().splitlines():
                line = line.strip()
                if line:
                    yield line.split(delim)
                else:
                    yield []

        if isinstance(self.raw, six.string_types):
            stdout, _ = self.raw, ''
        else:
            stdout, _ = self.raw

        reader = get_reader(stdout, self.delim)
        result = []

        if self.with_header:
            hds = tuple()
            # add temp by zby at 12/31
            remove_flag = False
            for row in reader:
                # add temp by zby  at 12/31
                if 'host_cluster_name' in row:
                    row.remove('host_cluster_name')
                    remove_flag = True
                hds = row
                break
            for row in reader:
                cur = dict()
                # add temp by zby  at 12/31
                if remove_flag:
                    row = row[:len(hds)]
                if len(hds) != len(row):
                    err_msg = (_('stdout: %(stdout)s')
                           % {'stdout': stdout})
                    LOG.error(err_msg)
                    msg = (_('Unexpected CLI response: header/row mismatch. '
                             'header: %(header)s, row: %(row)s.')
                           % {'header': hds,
                              'row': row})
                    raise exception.VolumeBackendAPIException(data=msg)
                for key, value in zip(hds, row):
                    CLIParser.append_dict(cur, key, value)
                result.append(cur)
        else:
            cur = dict()
            for row in reader:
                if row:
                    CLIParser.append_dict(cur, row[0], ' '.join(row[1:]))
                elif cur:  # start new section
                    result.append(cur)
                    cur = dict()
            if cur:
                result.append(cur)
        return result

    @staticmethod
    def append_dict(dict_, key, value):
        key, value = key.strip(), value.strip()
        obj = dict_.get(key, None)
        if obj is None:
            dict_[key] = value
        elif isinstance(obj, list):
            obj.append(value)
            dict_[key] = obj
        else:
            dict_[key] = [obj, value]
        return dict_
